#include "STLutils.h"
#include "Ast.h"
#include "ParserUtil.h"

int paramStrictCheck=WARN_LEVEL;
int safeCoercions=0;

AstNode::AstNode(AstNodeType nt, int fl, int line, int column, string file):
  ProgramElem(nullptr, fl, line, column, file) {
  nodeType_ = nt;
}

/****************************************************************/

ExprNode::ExprNode(ExprNodeType et, const Value* val, int fl, int line, int col,
				   string file):
  AstNode(EXPR_NODE, fl, line, col, file) {
  exprType_ = et;
  val_ = val;
  coercedType_ = nullptr;
}

void ExprNode::
coerce(const Type* tp) {
   // Since type() returns coercedType_ (if not null), it is necessary to
   // clear it first before calling isEquiv.
   coercedType_ = nullptr;
   if (tp && !tp->isEquiv(*type())) {
      if (Type::isNumeric(type()->tag()) && Type::isNumeric(tp->tag()) &&
          (type()->isSubType(*tp, true))) // Use relaxed version of check
         coercedType_ = tp; 
      else
         errMsg(type()->fullCodeName() + " cannot be converted to " 
                + tp->fullCodeName(), this);
   }
}

bool 
ExprNode::coercedTypeCodeGen(ostream& os, int indent, CGFlags& flags) {
   if (coercedType_) {
      if (coercedType_->isNumeric() && coercedType_->isSubType(*type())) {
         os << '(' << coercedType_->fullCodeName() << ')'; 
      }
      else os << '(' << coercedType_->fullCodeName() << ')'; 
      return true;
   }
   return false;
}

/****************************************************************/

RefExprNode::RefExprNode(RefExprNode* base, string ext, SymTabEntry* sym,
						 int fl, int ln,int col, string file):
  ExprNode(REF_EXPR_NODE, nullptr, fl, ln, col,file) {
  base_ = base;
  ext_ = ext;
  sym_ = sym;
  subscript_ = nullptr;
  constraint_ = nullptr;
}

AstNode* 
RefExprNode::clone(CGFlags* f) const {
   RefExprNode* rv;
   assert(base_ == nullptr); // Not supported yet.
   assert(symTabEntry()->kind() == VARIABLE_KIND); // Only vars can be assigned

   if (f != nullptr) {
      auto ve = (VariableEntry*)symTabEntry();
      if (f->subst_.count(ve) == 1) {
         assert(ve->varKind() != EVENT_PARAM_VAR);
         assert(base_ == nullptr); // We can substitute x, but not a.x
         ExprNode* e = f->subst_[(VariableEntry*)symTabEntry()];
         if (e->exprNodeType() != REF_EXPR_NODE) {
            assert(constraint_ == nullptr); // To replace a var with expr, there
            assert(subscript_ == nullptr);  // shd be no refexpr operations on it
            return e;
         }
         rv = (RefExprNode*)e->clone();
      }
      else if (ve->varKind() == GLOBAL_VAR && f->context() != "") {
         rv = new RefExprNode(*this);
         rv->ext_ = f->context() + "." + ext_;
      }
      else if (ve->varKind() == EVENT_PARAM_VAR) {
         assert(base_ == nullptr); // We can substitute x, but not a.x
         ExprNode* e = f->evargsubst_[ext_];
         if (e->exprNodeType() != REF_EXPR_NODE) {
            assert(constraint_ == nullptr); // To replace a var with expr, there
            assert(subscript_ == nullptr);  // shd be no refexpr operations on it
            return e;
         }
         rv = (RefExprNode*)e->clone();         
      }
      else {         
         rv = new RefExprNode(*this);
         if (f->prefix_ != "")
            rv->ext_ = f->prefix_ + ext_;
      }
   }
   else rv = new RefExprNode(*this);

   if (subscript_ != nullptr) {
      rv->subscript_ = new vector<ExprNode*>();
      for (auto r: *subscript_)
         rv->subscript_->push_back((ExprNode*)r->clone(f));
   }
   if (constraint_ != nullptr)
      rv->constraint_ = (ExprNode*)constraint_->clone(f);

   rv->tupleSubs_ = tupleSubs_;

   return rv;
}

const VariableEntry*
RefExprNode::root() const {
  if (base_ != nullptr) 
	return base_->root();
  else if ((sym_ != nullptr) && (sym_->kind() == VARIABLE_KIND))
	return (VariableEntry*)sym_;
  else return nullptr;
}

void 
RefExprNode::addSubs(ExprNode* e) { 
  if (subscript_ == nullptr)
    subscript_ = new vector<ExprNode*>();
  subscript_->push_back(e); 
}

void RefExprNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (base_)
      base_->usedSyms(use, assign);
   if (subscript_)
      for (unsigned j=0; j < subscript_->size(); j++)
         (*subscript_)[j]->usedSyms(use, assign);
   if (sym_ != nullptr) 
      use.insert(sym_);
}

const Type* RefExprNode::
typeCheck(set<const SymTabEntry*>& avlParams, 
          map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc) {
  if (coercedType()) 
     warnMsg("More than one typecheck", this);
  if (sym_ == nullptr) {
     if (base_ == nullptr)
        errM2("Identifier " << ext_ << " undeclared, or used outside its "
              << "scope", this);
        // For previously declared vars, parser would've initialized sym_
     else internalErr("Unresolved variable name"); // @@@@ for now, until 
     type(new Type(Type::errorType)); // @@@@ struct types are supported.
     return type();
  }
  if (base_ == nullptr) {
    if (sym_->kind() != VARIABLE_KIND) {
       errMsg("Expected a variable name, but found " + ext_, this); // CHECK
       //type(new Type(Type::errorType));
       type(new Type(*sym_->type()));
    }
    else {
       VariableEntry* ve = (VariableEntry*)sym_;
       if (tcc == COND_RHS && ve->varKind() == EVENT_PARAM_VAR)
          errMsg("Event parameters permitted only on left-hand sides of "
                 "comparisons in this context", this);
       else if (tcc == COND_RHS && ve->varKind() == MODULE_PARAM_VAR && 
                ve->type() && !ve->type()->isConst())
          errMsg("Right-hand side of conditions must be constants in this "
                 " context", this);
       else if (valVtypes.find(ve->varKind()) == valVtypes.end())
          errMsg(ve->name() + " cannot be used in this context", this);
       else if (ve->varKind() == EVENT_PARAM_VAR) {
          // Due to event patterns, the point where an event parameter becomes
          // invisible is determined by the rest of the pattern, so we need to
          // track it in avlParams.
          if (avlParams.find(sym_) == avlParams.end())
             errMsg("Use of " + ve->name() + " outside its scope", this);
       }
       type(new Type(*sym_->type()));
    }
  }
  else { // base_ != nullptr
    assert(0); // @@@@ Not handled yet.
    type(new Type(Type::errorType));
    const Type* baseType = base_->typeCheck(avlParams, valVtypes, tcc);
    if (baseType != nullptr && baseType->tag() == Type::STRUCT) {
       VariableEntry* baseEntry = ((VariableEntry*)base_->symTabEntry());
       if (baseEntry->type()->typeDesc()->kind() == STRUCT_KIND) {
         StructEntry* se = (StructEntry*)(baseEntry->type()->typeDesc());
         if (base_->base_ == nullptr) {
            // All the fancy lookup is applicable only for the second level name
            // resolution, i.e., for resolving b in a.b. For other levels,
            // e.g., resolving c in a.b.c, we should first completely resolve
            // a.b, and in the resulting struct or class, lookup c.
            list<StructEntry*> l;
            l.push_front(se);
            //sym_ = (SymTabEntry*)
               // (se->lookUpStruct(ext_, this->base_, constraint_, l, this));
         } 
         else sym_ = se->symTab()->lookUp(ext_);

         if (sym_ == nullptr) {
            errMsg(ext_ + " is not a field within " + base_->ext_, this);
            type(new Type(Type::errorType));
         } 
         else type(new Type(*sym_->type()));
       }
       else {
          errMsg("Invalid base for a field name access", this);
          type(new Type(Type::errorType));
          return type();
       }
    }
    else {
       type(new Type(Type::errorType));
       return &Type::errorType;
    }
  }

  assert(sym_ && sym_->kind() == VARIABLE_KIND);
  VariableEntry* ve = (VariableEntry*)sym_;
  if ((ve->nSubscripts() > 0 || nsubs() > 0) && type()->tag() != Type::ERROR) {
     // Subscript notation can be used for tuple/vector/dict, but in either case
     // # of supplied subscripts can't be less than what ve_ requires. Moreover,
     // since we don't permit arrays of containers, if ve_ expects any subscripts
     // at all, then all of them should be for arrays access.
     if (ve->nSubscripts() > nsubs() || 
         (ve->nSubscripts() > 0 && ve->nSubscripts() != nsubs())) {
        errMsg("Expected " + itoa(ve->nSubscripts()) + " subscripts", this);
        type(new Type(Type::errorType));
        return &Type::errorType;
     }

     const Type* mytype = sym_->type();
     if (mytype == nullptr) {
        type(new Type(Type::errorType));
        return &Type::errorType;
     }

     for (unsigned int i=0; i < nsubs(); i++) {
        const Type* t = (*subscript_)[i]->typeCheck(avlParams, valVtypes, tcc);
        if (t == nullptr || (*t == Type::errorType)) {
           errMsg("Invalid type for " + itoa(i+1) + "th subscript", this);
           type(new Type(Type::errorType));
           return &Type::errorType;
        }
        else if (mytype->tag() == Type::TUPLE) {
           tupleSubs_[i] = 1;
           auto argt = mytype->argTypes();
           const Value* v = (*subscript_)[i]->value();
           if (v == nullptr || !Type::isIntegral(v->type()->tag())) {
              errMsg("Tuple subscripts must be integer literals", this);
              type(new Type(Type::errorType));
              return &Type::errorType;
           }
           unsigned long sub = v->lval();
           if (sub >= argt->size()) {
              errMsg("Tuple subscript larger than tuple size", this);
              type(new Type(Type::errorType));
              return &Type::errorType;
           }
           mytype = (*argt)[sub];
        }
        else if (mytype->tag() == Type::DICT) {
           const Type* kt = mytype->keyType();
           if (Type::isIntegral(kt->tag())) {
              if (!Type::isIntegral(t->tag())) {
                 errMsg(itoa(i+1)+ string("th subscript must be integral"),this);
                 type(new Type(Type::errorType));
                 return &Type::errorType;
              }
           }
           else if (!kt->isEquiv(*t)) {
              if (t->isSubType(*kt))
                 (*subscript_)[i]->coerce(kt);
              else {
                 errM2("Expected type " << *kt<< " for " << i+1 << "th subscript"
                       << ", got type " << *t, this);
                 type(new Type(Type::errorType));
                 return &Type::errorType;
              }
           }
           mytype = mytype->elemType();     
        }
        else if (mytype->tag() == Type::VECTOR) {
           // @@@@ MISSING range check!!!!
           if (!Type::isIntegral(t->tag())) {
              errMsg(itoa(i+1) + string("th subscript must be integral"), this);
              type(new Type(Type::errorType));
              return &Type::errorType;
           }
           mytype = mytype->elemType();
        }
        else if (ve->nSubscripts() > 0) { // array
           if (!Type::isIntegral(t->tag())) {
              errMsg(itoa(i+1) + string("th subscript must be integral"), this);
              type(new Type(Type::errorType));
              return &Type::errorType;
           }
           ExprNode* e1 = (ExprNode*)ve->subscript(i);
           ExprNode* e2 = (*subscript_)[i];
           OpNode* opn1 = new OpNode(OpNode::LT, e2, e1, 0, 
                                     line(), column(), file());
           OpNode* opn2 = 
              new OpNode(OpNode::GE, e2, new ValueNode(new Value(Type::
                 BaseTypeTag(Type::BOOL), false)), 0, line(), column(), file());
           OpNode* opn = new OpNode(OpNode::AND, opn1, opn2,
                                    0, line(), column(), file());
           if (constraint_ == nullptr)
              constraint_ = opn;
           else 
              constraint_ = new OpNode(OpNode::AND, constraint_, opn, 
                                       0, line(), column(), file());
        }
        else {
           errMsg("Variable " + ve->name() + " of type " + ve->type()->fullName()
                  + " cannot be subscripted", this);
           type(new Type(Type::errorType));
           return &Type::errorType;
        }
     }
     type(new Type(*mytype));
  }
  return type();
}

void evalConstUpdate(ExprNode*& e) {
   const Value *v = e->evalConst();
   if (v) {
      auto n = new ValueNode(v, e);
      n->ExprNode::coerce(e->coercedType());
      e = n;
   }
}

const Value* RefExprNode::
evalConst() {
   if (base_)
      base_->evalConst();
   if (subscript_)
      for (unsigned j=0; j < subscript_->size(); j++) 
         evalConstUpdate((*subscript_)[j]);
   if (constraint_) {
      evalConstUpdate(constraint_);
   }
   // @@@@ Consider evalConst on constant tuple with subscript

   return nullptr;
}

void RefExprNode::
codeGen(ostream& os, int indent, CGFlags& flags) {
  bool coercion = coercedTypeCodeGen(os, indent, flags);
  if (coercion) os << '(';
  if (subscript_) {
     for (unsigned j=subscript_->size(); j != 0;) {
        --j;
        if (tupleSubs_[j])
           os << "get<" <<  (*subscript_)[j]->value()->lval() << ">(";
     }
  }

  if (base_ != nullptr) {
    base_->codeGen(os, indent, flags);
    if (ext_ != "")
       os << ".";
  }
  os << ext_;

  if (subscript_) {
     for (unsigned j=0; j < subscript_->size(); j++) {
        if (tupleSubs_[j])
           os << ")";
        else {
           os << "[";
           (*subscript_)[j]->codeGen(os, indent, flags);
           os << "]";
        }
     }
  }
  if (coercion) os << ')';
}
 
void 
RefExprNode::print(ostream& os, int indent) const {
  if (constraint_) 
    constraint_->print(os, indent);
  if (base_) {
    base_->print(os, indent);
    os << ".";
  }
  os << ext_;
  if (subscript_) {
    unsigned int i = 0, n = subscript_->size();
    while (i < n) {
      os << "[";
      (*subscript_)[i++]->print(os,indent);
      os << "]";
    }
  }
}
 
/****************************************************************/
//@@@@ Not used -- must have been introduced for some purpose,
//@@@@ need to check carefully.
const int OpNode::InverseOpCode[] = {
  -1, PLUS, -1, MULT, -1, -1, NE, EQ,
  LE, GE, LT, GT, -1, -1, -1, -1, -1, 
  -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1
};

const OpNode::OpInfo opInfo[] = {
  // print name, arity, paren_flag, fixity, arg types, out type, constraints
  //
  // Paren_flag -- opnode->print() outputs is surrounded by parenthesis if 
  // this flag is set. As set below, the expression may not print correctly
  // in some rare cases, e.g., ~(b * c) will get printed as ~b * c,
  // which actually corresponds to (~b)*c. To ensure that things get printed
  // correctly all the time, more paren_flags should be set to 1, but this
  // will lead to more clutter in printed output. Basically, what we have done
  // here is to look are expressions by type -- arithmetic, relational, 
  // boolean, bit operations, etc. Within each type, the highest priority 
  // operator is printed without paren. This will work correctly, as long
  // as the language doesn't permit mixing of different types of expressions.
  // But this assumption doesn't always hold, as in the example above. Also,
  // there is an exception to this general approach in the case of unary minus
  // and * -- since (-a)*b and -(a*b) have the same meaning, we can exclude
  // paren for * without an error.
  //
  // Codes for constraints:
  // first character:
  //    a digit, indicating which of the first 3 arguments are modified
  //    by the operation. (All other arguments are assumed to be unmodified
  //    by the operation.) A value '0' indicates an operation that changes no
  //    no arguments, while '5' indicates that the first and third arguments
  //    are modified.
  //    A special value 's' indicates that the function has some global
  //    side-effect but does not modify any arguments.
  // second character:
  //    N: No additional constraint over what is given by argTypes
  //    S: one of the arguments must have a type that is a supertype of
  //        of all other arguments. All other arguments require a coercion 
  //        operation to be introduced so as to convert their type to S.
  //    A: (assignment). Type of second argument must be a subtype of
  //       the first argument
  //    E: All arguments should have exact same type
  //    k: If the first argument is a collection type, second argument should
  //       be key type (elemtype, in case of non-key based collections)
  //
  // third character:
  //    N: No additional checks
  //    O: output type is the same as out type. (In the following cases,
  //        the output type need not be equal to out type, but a subtype
  //        of it.) Since a TypeTag provides complete type information only
  //        for primitive types, `O' is applicable only in this case.
  //    digit: (1..9) output type is the same as that of the digit'th argument
  //
  //    S: The output type is the same as that of the argument with the
  //        most general type. (Typically used with first character 'S')
  //
  // fourth character: (Used to control top-down propagation of coercions)
  //    N: No coercions needed.
  //    T: Coerce to the specified type of input
  //    M: Compute the most general type among all args, coerce all to this type
  //    C: Similar to M, but in addition, if argument is a constant and a 
  //       subtype of int, then coerce to int type.
  //    n: If output is double and input integral, propagate long or unsigned
  //       long, depending on whether an input arg. is integral and/or signed.
  //    g: All arguments are coerced to the most general type among the arguments.
  //       If the output type is more general than arguments, then arguments are
  //       instead coerced to this type. An exception occurs when output is
  //       double but inputs are integral: in this case, inputs are generalized
  //       to long (or unsigned long, depending on signedness)


  {OpNode::UMINUS, "-",  1, 0, OpNode::PREFIX, {Type::NUMERIC}, Type::SIGNED,                    "0N1g"},
  {OpNode::PLUS, "+",  2, 1, OpNode::INFIX, {Type::NUMERIC, Type::NUMERIC}, Type::NUMERIC,       "0SSg"},
  {OpNode::MINUS, "-",  2, 1, OpNode::INFIX, {Type::NUMERIC, Type::NUMERIC}, Type::NUMERIC,      "0SSg"},
  {OpNode::MULT, "*",  2, 0, OpNode::INFIX, {Type::NUMERIC, Type::NUMERIC}, Type::NUMERIC,       "0SSg"},
  {OpNode::DIV, "/",  2, 1, OpNode::INFIX, {Type::NUMERIC, Type::NUMERIC}, Type::NUMERIC,        "0SSg"},

  {OpNode::MOD, "%",  2, 1, OpNode::INFIX, {Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL,     "0S2n"},

  {OpNode::EQ, "==", 2, 0, OpNode::INFIX, {Type::NATIVE, Type::NATIVE}, Type::BOOL,              "0SOC"},
  {OpNode::NE, "!=", 2, 0, OpNode::INFIX, {Type::NATIVE, Type::NATIVE}, Type::BOOL,              "0SOC"},
  {OpNode::GT, ">",  2, 0, OpNode::INFIX, {Type::PRIMITIVE, Type::PRIMITIVE}, Type::BOOL,        "0SOC"},
  {OpNode::LT, "<",  2, 0, OpNode::INFIX, {Type::PRIMITIVE, Type::PRIMITIVE}, Type::BOOL,        "0SOC"},
  {OpNode::GE, ">=", 2, 0, OpNode::INFIX, {Type::PRIMITIVE, Type::PRIMITIVE}, Type::BOOL,        "0SOC"},
  {OpNode::LE, "<=", 2, 0, OpNode::INFIX, {Type::PRIMITIVE, Type::PRIMITIVE}, Type::BOOL,        "0SOC"},

  {OpNode::EQREG, "=~", 2, 0, OpNode::INFIX, {Type::STRING, Type::REGEXPR}, Type::BOOL,          "0NON"},
  {OpNode::NEQREG,"!~", 2, 0, OpNode::INFIX, {Type::STRING, Type::REGEXPR}, Type::BOOL,          "0NON"},

  {OpNode::AND, "&&",  2, 1, OpNode::INFIX, {Type::BOOL, Type::BOOL}, Type::BOOL,                "0NOO"},
  {OpNode::OR, "||",  2, 1, OpNode::INFIX, {Type::BOOL, Type::BOOL}, Type::BOOL,                 "0NOO"},
  {OpNode::NOT, "!",  1, 0, OpNode::PREFIX, {Type::BOOL}, Type::BOOL,                            "0NOO"}, 
  {OpNode::SEQ_AND, "&&",  2, 1, OpNode::INFIX, {Type::BOOL, Type::BOOL}, Type::BOOL,            "0NOO"},
  {OpNode::SEQ_OR, "||",  2, 1, OpNode::INFIX, {Type::BOOL, Type::BOOL}, Type::BOOL,             "0NOO"},

  {OpNode::BITNOT, "~",  1, 0, OpNode::PREFIX, {Type::INTEGRAL}, Type::INTEGRAL,                 "0N1n"},
  {OpNode::BITAND, "&",  2, 1, OpNode::INFIX, {Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL,  "0SSn"},
  {OpNode::BITOR, "|",  2, 1, OpNode::INFIX, {Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL,   "0SSn"},
  {OpNode::BITXOR, "^",  2, 0, OpNode::INFIX, {Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL,  "0SSn"},

  {OpNode::SHL, "<<", 2, 1, OpNode::INFIX, {Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL,     "0N1g"},
  {OpNode::SHR, ">>", 2, 1, OpNode::INFIX, {Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL,     "0N1g"},

  {OpNode::ASSIGN, "=",  2, 0, OpNode::INFIX, {Type::NATIVE, Type::NATIVE}, Type::VOID,          "1ANN"}, 

  {OpNode::EMPTY, "empty", 1, 1, OpNode::PREFIX, {Type::CONTAINER}, Type::BOOL,                  "0NON"},
  {OpNode::CLEAR, "clear", 1, 1, OpNode::PREFIX, {Type::CONTAINER}, Type::VOID,                  "1NON"},
  {OpNode::LENGTH, "length", 1, 1, OpNode::PREFIX, {Type::CONTAINER_OR_STR}, Type::UINT,         "0NON"},
  {OpNode::APPEND, "append", 2, 1, OpNode::PREFIX, {Type::VECTOR, Type::VECTOR}, Type::VOID,     "1NON"},
  {OpNode::UNION, "union_", 2, 1, OpNode::PREFIX, {Type::SET, Type::SET}, Type::VOID,            "1NON"},
  {OpNode::INS_SET, "insert", 2, 1, OpNode::PREFIX, {Type::SET, Type::NATIVE}, Type::BOOL,       "1kON"},
  {OpNode::INS_VEC,"insert",3,1,OpNode::PREFIX,{Type::VECTOR,Type::NATIVE,Type::UINT},Type::VOID,"1kON"}, 
  {OpNode::ERASE, "erase", 2, 1, OpNode::PREFIX, {Type::VECTOR, Type::UINT}, Type::VOID,         "1NON"},
  {OpNode::CONTAINS, "contains", 2, 1,OpNode::PREFIX,{Type::DICTORSET,Type::NATIVE},Type::BOOL,  "0kON"},
  {OpNode::REMOVE, "remove", 2, 1, OpNode::PREFIX, {Type::CONTAINER, Type::NATIVE}, Type::UINT,  "1kON"},
  {OpNode::CAT, "+",  2, 1, OpNode::INFIX, {Type::STRING, Type::STRING}, Type::STRING,           "1NON"},

  {OpNode::TUPLE, "make_tuple", OpNode::VARIABLE, 1, OpNode::PREFIX, {Type::NATIVE}, Type::TUPLE,"0NNN"},
  {OpNode::VECTOR, "list", OpNode::VARIABLE, 1, OpNode::PREFIX, {Type::NATIVE}, Type::VECTOR,    "0NNN"},
  {OpNode::SET, "set", OpNode::VARIABLE, 1, OpNode::PREFIX, {Type::NATIVE}, Type::SET,           "0NNN"},
  {OpNode::DICT, "dict", OpNode::VARIABLE, 1, OpNode::PREFIX, {Type::NATIVE}, Type::VECTOR,      "0NNN"},

  {OpNode::PRINT, "print", OpNode::VARIABLE, 0, OpNode::PREFIX, {Type::NATIVE}, Type::VOID,      "sNON"},
  {OpNode::PRINT_RAW, "print", OpNode::VARIABLE, 0, OpNode::PREFIX, {Type::NATIVE}, Type::VOID,      "sNON"},
  {OpNode::DUMMY_PROJECT, ".", 2, 1, OpNode::INFIX, {Type::TUPLE, Type::INTEGRAL}, Type::NATIVE, "0NNN"},
  {OpNode::EQBIT, "eqbit", 3, 1, OpNode::PREFIX, {Type::INTEGRAL, Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL, "0NNN"},
  {OpNode::NEQBIT, "neqbit", 3, 1, OpNode::PREFIX, {Type::INTEGRAL, Type::INTEGRAL, Type::INTEGRAL}, Type::INTEGRAL, "0NNN"},
  {OpNode::INTERVAL, "interval", 3, 1, OpNode::PREFIX, {Type::NUMERIC, Type::NUMERIC, Type::NUMERIC}, Type::NUMERIC, "0NNN"},
  {OpNode::INVALID, "invalid",            0, 0, OpNode::PREFIX, {}, Type::ERROR,                 "nNON"}
};

OpNode::OpNode(OpCode op, ExprNode* a1, ExprNode* a2, 
			   int fl, int ln, int col, string file):
  ExprNode(OP_NODE, nullptr, fl, ln,col,file) {
  opCode_ = op;
  if (a1 != nullptr) {
	arg_.push_back(a1);
	if (a2 != nullptr) {
	  arg_.push_back(a2);
	}
  }
  arity_ = arg_.size();
}

OpNode::OpNode(OpCode op, vector<ExprNode*>* args, int fl, int line, 
			   int column, string file):
  ExprNode(OP_NODE, nullptr, fl, line, column, file) { 
  opCode_ = op;
  if (args != nullptr) {
    int n = args->size(), i = 0;
    while (i < n)
      arg_.push_back((*args)[i++]);
    arity_ = n;
  } 
  arity_ = arg_.size();
}

/*
OpNode::OpNode(const OpNode &other):
  ExprNode(other) {
  arity_ = other.arity();
  opCode_ = other.opCode();
  for (unsigned int i=0; i < other.arity_; i++) {
    if (other.arg_[i])
       arg_.push_back((other.arg_[i])->clone());
	else arg_.push_back(nullptr);
  }
}
*/

AstNode* 
OpNode::clone(CGFlags* fl) const {
   OpNode* rv = new OpNode(*this);
   rv->arg_.clear();
   for (auto a: arg_)
      rv->arg_.push_back((ExprNode*)a->clone(fl));
   return rv;
}

bool OpNode::
checkRuleVarAsg() const {
   if (opCode_ != ASSIGN || arg_.size() < 2 || 
       arg_[0]->exprNodeType() != REF_EXPR_NODE)
      return false;
   auto lhs = (RefExprNode*)arg_[0];
   return (lhs->symTabEntry() && lhs->symTabEntry()->kind() == VARIABLE_KIND &&
           ((VariableEntry*)lhs->symTabEntry())->varKind() == RULE_VAR);
}

void OpNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   for (unsigned i=0; i < arity_; i++)
      if (arg_[i] != nullptr) {
         if (opInfo[opCode_].typeConstraints[1] == 'A' && i == 0)
            arg_[0]->usedSyms(assign, assign); // NOT a typo.
         else arg_[i]->usedSyms(use, assign);
      }

   for (unsigned i=0; i < arity_; i++)
      if (arg_[i] != nullptr && i < 3 && 
          opInfo[opCode_].typeConstraints[0] >= '0' && 
          opInfo[opCode_].typeConstraints[0] < '8')
         if ((opInfo[opCode_].typeConstraints[0] - '0') & (1<<i)) {
            if (arg_[i]->exprNodeType() == REF_EXPR_NODE)
               assign.insert(((RefExprNode*)arg_[i])->symTabEntry());
         }
}

const Type* OpNode::
typeCheck(set<const SymTabEntry*>& avlParams, 
          map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc) {
  if (coercedType()) 
     warnMsg("More than one typecheck", this);

/* Better to check this condition in normalize(), because we do permit
   other operations on the lhs, as long as they can be statically evaluated
  if (tcc == COND_LHS && opCode_ != BITAND) {
     errMsg(string("Operator ") + opInfo[opCode_].name_ + 
            " is invalid on left-hand sides of conditions", this);
     type(new Type(Type::voidType));
     return type();
  }
*/
  if (tcc == COND_RHS && (opCode_ == AND || opCode_ == OR || opCode_ == SEQ_AND
                          || opCode_ == SEQ_OR || opCode_ == ASSIGN ||
                          (opCode_ >= EMPTY && opCode_ <= REMOVE) ||
                          opCode_ == TUPLE || opCode_ == VECTOR ||
                          opCode_ >= DICT)) {
     errMsg(string("Operator ") + opInfo[opCode_].name_ + 
            " is invalid on right-hand sides of conditions", this);
     type(new Type(Type::voidType));
     return type();
  }
  switch (opInfo[opCode_].code_) {
     case TUPLE: {
        vector<Type*>* at = new vector<Type*>;
        for (unsigned i=0; i < arg_.size(); i++) {
           const Type* argtype = arg_[i]->typeCheck(avlParams, valVtypes, tcc);
           at->push_back(new Type(*argtype));
        }
        type(new Type(at, Type::TUPLE));
        return type();
     }

     case VECTOR:
     case SET: {
        auto tag = (opInfo[opCode_].code_ == VECTOR)? Type::VECTOR : Type::SET;
        if (arg_.size() == 0)
           type(new Type(new Type(Type::NATIVE), tag));
        else {
          Type* argtype = new Type(*arg_[0]->typeCheck(avlParams,valVtypes,tcc));
          for (unsigned j=1; j < arg_.size(); j++) {
            arg_[j]->typeCheck(avlParams, valVtypes);
            argtype->unify(arg_[j]->type(), this);
            if (!argtype->isValid()) {
              errM2("Dissimilar types not permitted within container", this);
              type(new Type(Type::errorType));
              return type();
            }
          }
          type(new Type(argtype, tag));
        }
        return type();
     }

     case DICT:
        if (arg_.size() == 0)
           type(new Type(new Type(Type::NATIVE), new Type(Type::NATIVE)));
        else {
          Type* argtype = new Type(*arg_[0]->typeCheck(avlParams,valVtypes,tcc));
          for (unsigned j=1; j < arg_.size(); j++) {
             arg_[j]->typeCheck(avlParams, valVtypes, tcc);
             argtype->unify(arg_[j]->type(), this);
             if (!argtype->isValid()) {
                errMsg("Dissimilar types not permitted within container", this);
                type(new Type(Type::errorType));
                return type();
             }
           }
           Type* t1 = (*(argtype->argTypes()))[0];
           Type* t2 = (*(argtype->argTypes()))[1];
           type(new Type(t1, t2));
         }
        return type();

     default:
       break;
  }

  // First, compute argument types
  if (opInfo[opCode_].typeConstraints[1] != 'A') {
     for (unsigned i=0; i < arity_; i++) {
        ExprNode *en = arg_[i];
        if (en != nullptr) {
           TypeCheckCtxt tcc1 = tcc;
           if (tcc == COND && opCode_ >= EQ && opCode_ <= EQREG) {
              if (i==0) tcc1 = COND_LHS;
              else if (i==1) tcc1 = COND_RHS;
           }
           Type* t1 = new Type(*(en->typeCheck(avlParams, valVtypes, tcc1)));
           en->type(t1);
           // If a subexpression leads to a type error, don't proceed further.
           // Otherwise, one error can generate a million more on the same line.
           if (!t1->isValid()) {
              type(t1);
              return t1;
           }
        }
        else errMsg("Unexpected null value", this);
     }
  }

  // Resolve overloading
  if (opCode_ == INS_VEC && arity_ == 2)
     opCode_ = INS_SET;
  else if (opCode_ == PLUS && 
           (Type::isString(arg_[0]->type()->tag()) 
            || Type::isString(arg_[1]->type()->tag())))
     opCode_ = CAT;

  if (opInfo[opCode_].typeConstraints[1] != 'A') { 
     for (unsigned i=0; i < arity_; i++) {
        const Type* t1 = arg_[i]->type();
        if (!t1->isSubType(opInfo[opCode_].argType(i))) {
           errM2("Expected type " << Type::name(opInfo[opCode_].argType(i)) <<
                 " for argument " << i+1 << " of operator '" <<
                 opInfo[opCode_].name_ << "' but got " << *t1, this);
           type(new Type(Type::errorType));
           return type();
        }
     }
  }

  Type *suptype = nullptr;

  switch (opInfo[opCode_].typeConstraints[1]) {
     case 'N': 
        break;

     case 'S': 
        suptype = new Type(*arg_[0]->type());
        for (unsigned j=1; j < arg_.size(); j++) 
           suptype->unify(arg_[j]->type(), this);
        if (!suptype->isValid()) {
           errMsg(string("All arguments of '") + opInfo[opCode_].name_ + 
                     "' must have compatible types", this);
           type(new Type(Type::errorType));
           return type();
        }
        for (unsigned j=0; j < arg_.size(); j++)
           arg_[j]->coerce(suptype);
        break;

     case 'A': {
       // Recall that for assignments, we skipped typecheck on args above.
        RefExprNode *lhs;
        ExprNode* rhs = arg_[1];
        VariableEntry *ve;
        if (arg_[0]->exprNodeType() != REF_EXPR_NODE || 
            ((lhs = (RefExprNode*)arg_[0])->base()) ||
            ((ve = (VariableEntry*)lhs->symTabEntry())->kind() != VARIABLE_KIND))
           internalErr("Unexpected RefExprNode");
        else {
           rhs->typeCheck(avlParams, valVtypes, tcc);
           const Type* rhst = rhs->type();
           bool isRuleAsg = false;
           if (rhst && rhst->isValid()) {
              if ((isRuleAsg = (ve->varKind() == RULE_VAR)))
                 // A rule variable's type to be inferred from first assignment.
                 ve->type(new Type(*rhst));

              lhs->typeCheck(avlParams, valVtypes, tcc);
              const Type* lhst = lhs->type();
              if (lhst && lhst->isValid()) {
                 if (!rhst->isSubType(*lhst)) {
                    if (!rhst->isSubType(*lhst, true)) // relaxed ver of check
                      errM2("Assigned type '"<< *rhst<< "' not a subtype of "
                        << "target type '" << *lhst << "'", this);
                    else {
                       warnMsg("Signedness or possible integer overflow "
                                 "in assignment", this);
                       rhs->coerce(new Type(*lhst));
                    }
                 }
                 else rhs->coerce(new Type(*lhst));

                 if (isRuleAsg)
                    type(new Type(Type::boolType));
                 else type(new Type(Type::voidType));
              }
              else type(new Type(Type::errorType));
           }
           else type(new Type(Type::errorType));
           return type();
        }
        break;
     }

     case 'E': {
        for (unsigned i=1; i < arity_; i++) {
           if (!arg_[0]->type()->isEquiv(*arg_[i]->type()))
              errMsg(string("All arguments of '") + opInfo[opCode_].name_ + 
                     "' must have identical types", this);
        }
        break;
     }

     case 'k': { 
        assert(arity_ >= 2);
        const Type* keytype = arg_[0]->type()->keyType();
        const Type* at = arg_[1]->type();
        /*if (Type::isNumeric(at->tag())) {*/
           if (at->isSubType(*keytype)) {
              if (!keytype->isEquiv(*at))
                 arg_[1]->coerce(keytype);
           }
           else {
              errMsg("Expected type " + keytype->fullCodeName() + 
                     " but got " + at->fullCodeName(), this);
           }

        /*}
        else if (!keytype->isEquiv(*at))
           errMsg(string("Incompatible arguments to '")
                       + opInfo[opCode_].name_ + "'", this);*/
        break; 
     }

     default:
        internalErr("Unhandled case");
        break;
  }

  char cc = opInfo[opCode_].typeConstraints[0];
  if (cc > '0' && cc < '8') {
     int modmask = cc - '0';
     for (unsigned j=0; j < 3; j++) {
        int m = (1<<j);
        if (m & modmask) {
           if (arg_[j]->exprNodeType() == REF_EXPR_NODE) {
             auto *ve = (VariableEntry*)(((RefExprNode*)arg_[j])->symTabEntry());
             if (valVtypes.find(ve->varKind()) == valVtypes.end() || 
                  !valVtypes[ve->varKind()]) {
                errM2("Variable " << ve->name() << " cannot be updated in " 
                                  << "this context", arg_[j]);
                //cerr << "varkind=" << ve->varKind() << endl;
             }

             if (ve->type()->isConst())
                errM2("const variable "<< ve->name() << " cannot be "
                         << "updated", arg_[j]);
           }
         }
     }
  }
            
  switch (opInfo[opCode_].typeConstraints[2]) {
     case 'N': 
        break;

     case 'O': 
        type(new Type(*Type::type[opInfo[opCode_].outType_]));
        break;

     case '1':
        if (arg_[0]->coercedType() != nullptr)
           type(new Type(*arg_[0]->coercedType()));
        else type(new Type(*arg_[0]->type()));
        break;

     case '2':
        if (arg_[1]->coercedType() != nullptr)
           type(new Type(*arg_[1]->coercedType()));
        else type(new Type(*arg_[1]->type()));
        break;

     case '3':
        if (arg_[2]->coercedType() != nullptr)
           type(new Type(*arg_[2]->coercedType()));
        else type(new Type(*arg_[2]->type()));
        break;

     case 'S':
        type(suptype);
        break;

     default:
        internalErr("Unhandled case");
        break;
  }

  if (!type() || !type()->isValid()) {
     errMsg(string("Invalid argument types for ") + 
            opInfo[opCode_].name_, this);
     type(new Type(Type::errorType));
  }

  if (type() && type()->isValid() &&
      !(type()->isSubType(opInfo[opCode_].outType_))) {
    errMsg("Output type does match expected type", this);
	if (Type::isIntegral(opInfo[opCode_].outType_))
	  type((Type*)&(Type::intType));
	else if (Type::isNumeric(opInfo[opCode_].outType_))
	  type((Type*)&(Type::doubleType));
  }
  return type();
}

void
OpNode::coerce(const Type* t) {
   switch (opCode_) {
   case APPEND:
   case UNION:
      arg_[1]->coerce(arg_[0]->type());
      return;

   case INS_SET:
   case INS_VEC:
   case CONTAINS:
   case REMOVE:
      arg_[1]->coerce(arg_[0]->type()->keyType());
      return;

   case TUPLE:
      for (unsigned j=0; j < arg_.size(); j++) 
         arg_[j]->coerce((*t->argTypes())[j]);
      type(new Type(*t));
      return;

   case SET:
   case VECTOR:
      for (unsigned j=0; j < arg_.size(); j++) 
         arg_[j]->coerce(t->elemType());
      type(new Type(*t));
      return;

   case DICT:
      for (unsigned j=0; j < arg_.size(); j++) {
         assert(arg_[j]->exprNodeType() == OP_NODE);
         OpNode* on = (OpNode*)arg_[j];
         on->arg_[0]->coerce(t->keyType());
         on->arg_[1]->coerce(t->elemType());
      }
      type(new Type(*t));
      return;      

   case ASSIGN:
      break;

   case PRINT:
   case PRINT_RAW:
      for (unsigned j=0; j < arg_.size(); j++)
         if (Type::isIntegral(arg_[j]->type()->tag())) {
            if (Type::isSigned(arg_[j]->type()->tag())) 
               arg_[j]->coerce(&Type::longType);
            else arg_[j]->coerce(&Type::ulongType);
         }
      // No need to propagate higher precision for other arg types

   default:
      break;
   }

   if (opInfo[opCode_].typeConstraints[3] == 'N')
      return;

   if (Type::isNumeric(t->tag()) || Type::isBool(t->tag()))
      ExprNode::coerce(t);

   // Meaningful to propagate coercion only if the output is needed at higher
   // precision than the default precision (given by this node's type); or if
   // there are non-zero arguments to propagate to.
   if (!type()->isSubType(*t) || arg_.size() == 0)
      return;

   Type* argtype = nullptr;
   switch (opInfo[opCode_].typeConstraints[3]) {
   case 'C':
   case 'g': 
      argtype = new Type(*arg_[0]->type());
      for (unsigned j=0; j < arg_.size(); j++) 
         argtype->unify(arg_[j]->type(), this);
      if (Type::isDouble(t->tag())) {
         if (Type::isIntegral(argtype->tag())) {
            if (Type::isSigned(type()->tag())) 
               *argtype = Type::longType;

            else *argtype = Type::ulongType;
         }
         else *argtype = *t;
      }
      else if (Type::isIntegral(t->tag()))
         argtype->unify(t, this);

      for (unsigned j=0; j < arg_.size(); j++) {
         if (/*opInfo[opCode_].typeConstraints[3] == 'C' &&*/
             arg_[j]->value() && 
             argtype->isSubType(Type::intType))
            arg_[j]->coerce(new Type(Type::intType));
         else arg_[j]->coerce(argtype);
      }
      break;

   case 'O':
      for (unsigned j=0; j < arg_.size(); j++) 
         arg_[j]->coerce(new Type(opInfo[opCode_].argType(j)));
      break;
      
   case 'n': 
      for (unsigned j=0; j < arg_.size(); j++) {
        if (Type::isDouble(t->tag())) {
           if (Type::isIntegral(arg_[j]->type()->tag())) {
              if (Type::isSigned(type()->tag())) 
                 arg_[j]->coerce(&Type::longType);
              else arg_[j]->coerce(&Type::ulongType);
           }
           else arg_[j]->coerce(t);
        }
        else arg_[j]->coerce(t); 
        // If we reach here, t must be integral. Plus, it should be more
        // more general than the type of thie node, so just propagate t.
      }
      break;

   default: assert(false);
   }
}

const Value* OpNode::
evalConst() {
   // Should be called after typecheck, but before normalize()
   if (opCode_ == EQREG || opCode_ == NEQREG) return nullptr;
   bool allConst = true;
   bool isBool = arg_[0]->type()->isBool();
   bool isString = arg_[0]->type()->isString();
   bool isIntegral = arg_[0]->type()->isIntegral();
   bool isNumeric = arg_[0]->type()->isNumeric();
   long lv[4]; double dv[4]; string sv[4];
 
   for (unsigned j=0; j < arg_.size(); j++) {
      evalConstUpdate(arg_[j]);
      const Value* v = arg_[j]->value();
      if (v) {
         if (j < 4) {
            if (isIntegral)
               lv[j] = v->lval();
            else if (isBool)
               lv[j] = v->bval();
            else if (isNumeric)
               dv[j] = v->dval();
            else if (isString)
               sv[j] = v->sval();
            else allConst = false;
         }
         else allConst = false;
      }
      else allConst = false;
   }
   if (!allConst) return nullptr;

#define retv(x) do {value(x); return x;} while (0)
#define mkbv(e) retv(new Value(Type::BOOL, e))
#define mkv(e)  retv(new Value(e))
#define mkldv(op) if (isIntegral) mkv(lv[0] op lv[1]); else mkv(dv[0] op dv[1])
#define anybv(op) if (isBool) mkbv(lv[0] op lv[1]); \
                  else if (isString) mkbv(sv[0] op sv[1]); \
                  else if (isIntegral) mkbv(lv[0] op lv[1]); \
                  else     mkbv(dv[0] op dv[1]);

   switch (opCode_) {
   case UMINUS: if (isIntegral) mkv(-lv[0]); else mkv(-dv[0]);
   case MOD:    mkv(lv[0] % lv[1]);

   case PLUS:   mkldv(+);
   case MINUS:  mkldv(-);
   case MULT:   mkldv(*);
   case DIV:    mkldv(/);

   case EQ: anybv(==);
   case NE: anybv(!=);
   case GT: anybv(>);
   case GE: anybv(>=);
   case LT: anybv(<);
   case LE: anybv(<=);

   case SEQ_AND:
   case AND: 
      mkbv(lv[0] && lv[1]);
   case SEQ_OR:
   case OR: 
      mkbv(lv[0] || lv[1]);
   case NOT:
      mkbv(!lv[0]);

   case BITNOT: mkv(~lv[0]);
   case BITAND: mkv(lv[0] & lv[1]);
   case BITOR:  mkv(lv[0] | lv[1]);
   case BITXOR: mkv(lv[0] ^ lv[1]);
   case SHL:    mkv(lv[0] << lv[1]);
   case SHR:    mkv(lv[0] >> lv[1]);

   case CAT: mkv(sv[0] + sv[1]);

   // @@@@ Consider: CONTAINS, EMPTY
   default:
      return nullptr;
   }

#undef mkbv
#undef mkv
#undef mkldv
#undef anybv
}

void 
OpNode::codeGen(ostream& os, int indent, CGFlags& flags) {
  switch (opCode_) {
     case PRINT:
     case PRINT_RAW: {
        for (unsigned j=0; j < arity_; j++) {
           if (opCode_ == PRINT && j != 0) os << "*esout << ' ';";
           os << "print(";
           arg_[j]->codeGen(os, indent, flags);
           os << ", *esout);";
        }
        if (opCode_ == PRINT) os << "*esout << endl";
        return;
     }

     case TUPLE: {
        os << type()->fullCodeName();
        os << "{"; bool c = false;
        for (unsigned j=0; j < arity_; j++) {
           if (c) os << ", ";
           else c = true;
           arg_[j]->codeGen(os, indent, flags);
        }
        os << "}";
        return;
     }
     case SET:
     case VECTOR: {
        os << type()->fullCodeName();
        os << "{"; bool c = false;
        for (unsigned j=0; j < arity_; j++) {
           if (c) os << ", ";
           else c = true;
           arg_[j]->codeGen(os, indent, flags);
        }
        os << "}";
        return;
     }

     case DICT: {
        os << type()->fullCodeName();
        os << "{"; bool c = false;
        for (unsigned j=0; j < arity_; j++) {
           if (c) os << ", ";
           else c = true;
           assert(arg_[j]->exprNodeType() == OP_NODE);
           OpNode* on = (OpNode*)arg_[j];
           ExprNode* en1 = on->arg_[0];
           ExprNode* en2 = on->arg_[1];
           os << "{";
           en1->codeGen(os, indent, flags);
           os << ", ";
           en2->codeGen(os, indent, flags);
           os << "}";
        }
        os << "}";
        return;
    }

    case INTERVAL:
       os << '(';
       arg_[0]->codeGen(os, indent, flags);
       if (arg_[1] != nullptr) {
          os << " >= ";
          arg_[1]->codeGen(os, indent, flags);
       }
       if (arg_[2] != nullptr) {
          if (arg_[1] != nullptr) {
             os << " && ";
             arg_[0]->codeGen(os, indent, flags);
          }
          os << " <= ";
          arg_[2]->codeGen(os, indent, flags);
       }
       os << ')';
       return;

    case EQBIT:
    case NEQBIT:
       os << '(';
       arg_[0]->codeGen(os, indent, flags);
       os << '&';
       arg_[1]->codeGen(os, indent, flags);
       if (opCode_ == EQBIT)
          os << " == ";
       else os << " != ";
       arg_[2]->codeGen(os, indent, flags);
       os << ')';
       return;
       
  default:
     coercedTypeCodeGen(os, indent, flags);
     os << "(";
     if (opInfo[opCode_].prtType_ == OpNode::PREFIX) {
        os << opInfo[opCode_].name_ << '(';
        for (unsigned i=0; i < arity_-1; i++) {
           arg_[i]->codeGen(os, indent, flags);
           os << ", ";
        }
        arg_[arity_-1]->codeGen(os, indent, flags);
        os << ")";
     }
     else if ((opInfo[opCode_].prtType_ == OpNode::INFIX) && (arity_ == 2)) {
        arg_[0]->codeGen(os, indent, flags);
        os << opInfo[opCode_].name_; 
        arg_[1]->codeGen(os, indent, flags);
     }
     else internalErr("Unhandled case in OpNode::codeGen");
     os << ")";
  }
}

void 
OpNode::print(ostream& os, int indent) const {
  if (opInfo[opCode_].prtType_ == OpNode::PREFIX) {
	os << opInfo[opCode_].name_;
	if (arity_ > 0) {
	  if (opInfo[opCode_].needParen_) 
		os << '(';
	  for (unsigned i=0; i < arity_-1; i++) {
		if (arg_[i])
		  arg_[i]->print(os, indent);
	    else os << "nullptr";
		os << ", ";
	  }
      if (arg_[arity_-1])
		arg_[arity_-1]->print(os, indent);
	  else os << "nullptr";
	  if (opInfo[opCode_].needParen_) 
		os << ")";
	}
  }
  else if ((opInfo[opCode_].prtType_ == OpNode::INFIX) && (arity_ == 2)) {
	if (opInfo[opCode_].needParen_)
	  os << "(";
	if(arg_[0])
	  arg_[0]->print(os, indent);
	else os << "nullptr";
	os << opInfo[opCode_].name_; 
	if(arg_[1])
	  arg_[1]->print(os, indent);
	else os << "nullptr";
	if (opInfo[opCode_].needParen_) 
	  os << ")";
  }
  else internalErr("Unhandled case in OpNode::print");
}

/****************************************************************/

const Type*
SizeOfNode::typeCheck(set<const SymTabEntry*>& avl, map<VarKind, bool>& val,
                      TypeCheckCtxt tcc) {
  if (ste_ == nullptr || ste_->kind() != STRUCT_KIND)
     errMsg("Invalid argument to sizeof()", this);
  type(new Type(Type::uintType));
  return type();
}

void 
SizeOfNode::codeGen(ostream& os, int indent, CGFlags& flags) {
  if(ste_ != nullptr) {
    os << "sizeof(" << ste_->name() << ")";
  }
}

void 
SizeOfNode::print(ostream& os, int indent) const {
  if(ste_ != nullptr){
    os << "sizeof(" << ste_->name() << ")";
  }
}

/****************************************************************/

const Type*
ValueNode::typeCheck(set<const SymTabEntry*>& avl, map<VarKind, bool>& val,
                     TypeCheckCtxt tcc) {
  if (coercedType()) 
     warnMsg("More than one typecheck", this);
  return ((const ValueNode*)this)->type();
}

void
ValueNode::codeGen(ostream& os, int indent, CGFlags& flags) {
  const Value* val = value();
  if (val != nullptr) {
    val->codeGen(os, coercedType(), flags.prefix());
  }
}

void
ValueNode::print(ostream& os, int indent) const {
  const Value* val = value();
  if (val != nullptr)
    val->print(os,indent);
  else os << "nullptr";
}

/****************************************************************/

InvocationNode::InvocationNode(SymTabEntry *ste, vector<ExprNode*>* par, 
							   int fl, int line, int column, string file):
  ExprNode(INV_NODE,0, fl, line,column,file) {
  params_ = (par == nullptr)? new vector<ExprNode*>() : par;
  (ste != nullptr) ? type(new Type(*ste->type())) : type(nullptr);
  ste_ = ste;
}

/*
InvocationNode::InvocationNode(const InvocationNode& other)
  : ExprNode(other) {
  if (other.params_ != nullptr) {
    params_ = new vector<ExprNode*>();
    for (unsigned int i=0; i< other.params_->size(); i++)
      params_->push_back( (*(other.params_))[i]->clone() ); 
  }
  ste_ = (SymTabEntry*)other.symTabEntry();
}
*/

AstNode* 
InvocationNode::clone(CGFlags* fl) const {
   InvocationNode* rv = new InvocationNode(*this);
   if (fl != nullptr && fl->subst_.count(ste_) == 1) {
      ExprNode* e = fl->subst_[ste_];
      assert(e->exprNodeType() == REF_EXPR_NODE);
      rv->ste_ = ((RefExprNode*)e)->symTabEntry();
   }
   if (params_ != nullptr) {
      rv->params_ = new vector<ExprNode*>;
      for (auto p: *params_)
         rv->params_->push_back((ExprNode*)p->clone(fl));
   }
   return rv;
}

void InvocationNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   SymTabEntry *ste = symTabEntry();
   if (ste != nullptr && ste->type() != nullptr) {
      use.insert(ste);
      for (unsigned i = 0; (i < params_->size() && ste->type()->arity()); i++) {
         (*params_)[i]->usedSyms(use, assign);
         const Type* argtype = (*ste->type()->argTypes())[i];
         if (!Type::isPrimitive(argtype->tag()) && !argtype->isConst()) {
            if ((*params_)[i]->exprNodeType() == REF_EXPR_NODE)
               assign.insert(((RefExprNode*)(*params_)[i])->symTabEntry());
         }
      }
   }
};

const Type* InvocationNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
   if (coercedType()) 
     warnMsg("More than one typecheck", this);

   const SymTabEntry *ste = symTabEntry();
   if (ste != nullptr && ste->type() != nullptr) {
      unsigned expArgs = ste->type()->arity();
      if (params_->size() != expArgs)
         errMsg(itoa(expArgs) + " arguments expected for " + ste->name(), this);
      else {
        //if (tcc != NONCOND) {
        //  errMsg("Function calls not permitted within event conditions", this);
        //  type(new Type(Type::errorType));
        //  return type();
        //}
        map<VarKind, bool> vv;
        switch (ste->kind()) {
          case FUNCTION_KIND: // Function invocation
          case VARIABLE_KIND: // Must be a module parameter of function type
            if ((valVtypes.find(GLOBAL_VAR) == valVtypes.end() || 
              !valVtypes[GLOBAL_VAR]) && !(ste->type()->isConst()))
               errMsg("Non-const functions prohibited in this context", this);
            vv = valVtypes;
            break;
          case MODULE_KIND: // Module instantiation. In this case the parameters
            vv[MODULE_PARAM_VAR] = false; // can contain only constants and
            vv[GLOBAL_VAR] = true; // other module parameters, and state vars.
            break; 

          default:
            errMsg(ste->name() + "can't be used as a function", this);
            break;
        }
        for (unsigned i = 0; (i < params_->size()); i++) {
           ExprNode* actparam = (*params_)[i];
           actparam->typeCheck(avlParams, vv, tcc);
           const Type &formaltp = *((*type()->argTypes())[i]);
           if (!(actparam->type()->isSubType(formaltp, true))) { // like asg chk
                 errM2(ste->name() << ": Type mismatch for argument " << i+1 
                       << " (expected: " << formaltp << ", got " 
                       << *actparam->type() << ')', this);
           }
           else {
              if (paramStrictCheck) {
                 if (!(actparam->type()->isSubType(formaltp))) 
                    prtM2(paramStrictCheck, ste->name() << ": Type mismatch for"
                          << " argument " << i+1 << " (expected: " << formaltp 
                          << ", got " << *actparam->type() << ')', this);
              }
              if (!formaltp.isEquiv(*actparam->type()))
                 actparam->coerce((Type*)&formaltp);
              // Const check: For function invocation, primitive types are always
              // passed by value, so check is needed only for non-primitive 
              // types. For module invocations, this check is always needed.
              if (ste->kind() != FUNCTION_KIND ||
                  !Type::isPrimitive(actparam->type()->tag())) {
                 if (actparam->type()->isConst()) {
                    if (!formaltp.isConst())
                       errMsg(ste->name()+ " expects argument " + itoa(i+1)
                              + " to be non-const", this);
                 }
                 else if (actparam->exprNodeType() != REF_EXPR_NODE
                          && !formaltp.isConst()) {
                    errMsg("Either " + ste->name() + " must take a constant " 
                           "argument at position " + to_string(i+1) + " or the "
                           "argument must be a variable", this);
                 }
              }
           }
        }
        if (ste->type()->retType() == nullptr)
           type(new Type(Type::voidType));
        else type(new Type(*ste->type()->retType()));
      }
   }
   else internalErr("Unhandled case");

   if (!type())
      type(new Type(Type::errorType));
   return type();
}

const Value*
InvocationNode::evalConst() {
   if (params_)
      for (auto param: *params_)
         param->evalConst();

   return nullptr; 
   // No support for compile-time evaluation of functions. Use macros if needed.
}

void 
InvocationNode::codeGen(ostream& os, int indent, CGFlags& flags) {
  bool coercion = coercedTypeCodeGen(os, indent, flags);
  if (coercion) os << '(';
  if(ste_ != nullptr) {
    os << flags.prefix() << ste_->name() << "(";
  }
  if(params_ != nullptr) {
    unsigned int i = 0, n = params_->size();
    for(; i < n; i++) {
      if ((*params_)[i] != nullptr) 
		(*params_)[i]->codeGen(os, indent, flags);
      if(i < n-1)
		os << ", ";
    }
  }
  os << ")";
  if (coercion) os << ')';
}

void 
InvocationNode::print(ostream& os, int indent) const {
  if(ste_ != nullptr) {
    os << ste_->name() << "(";
  }
  if(params_ != nullptr) {
    unsigned int i = 0, n = params_->size();
    for(; i < n; i++) {
      if ((*params_)[i] != nullptr) 
		(*params_)[i]->print(os, indent);
      if(i < n-1)
		os << ", ";
    }
  }
  os << ")";
}

/*
bool
InvocationNode::isConst() const {
   if (ste_ == nullptr || ste_->kind() != FUNCTION_KIND) return false;
   if (params_ != nullptr)
      for (unsigned i=0; i < params_->size(); i++)
         if (!(*params_)[i]->isConst()) return false;
   return true;
};
*/

/****************************************************************/
/*                  PRIMITIVE PAT NODE                          */
/****************************************************************/

PrimitivePatNode::PrimitivePatNode(EventEntry* ee,vector<VariableEntry*>* params,
        					   ExprNode* cond, int fl, int l,int c,string file):
  BasePatNode(BasePatNode::PRIMITIVE, fl, l,c,file) {
  ee_ = ee;
  params_ = (params==nullptr)? new vector<VariableEntry*>() : params;
  cond_ = cond;
  if (!cond_) cond_ = new ValueNode(trueExpr);
}

AstNode* 
PrimitivePatNode::clone(CGFlags* f) const {
   PrimitivePatNode* rv = new PrimitivePatNode(*this);
   if (f != nullptr && f->subst_.count(ee_) == 1)
      rv->ee_ = (EventEntry*)f->subst_[ee_];
   rv->params_ = new vector<VariableEntry*>;
   for (unsigned i=0; i < params_->size(); i++) {
      VariableEntry* oldv = (*params_)[i];
      VariableEntry* newv = new VariableEntry(*oldv);
      string newnm = "__p" + to_string(i);
      newv->name(newnm);
      rv->params_->push_back(newv);
      if (f != nullptr) {
         auto newnd = new RefExprNode(nullptr, newnm, oldv, -1, oldv->line(),
                                      oldv->column(), oldv->file());
         newnd->type(new Type(*oldv->type()));
         f->evargsubst_[oldv->name()] = newnd;
      }
   }
   if (cond_ != nullptr)
      rv->cond_ = (ExprNode*)cond_->clone(f);
   return rv;
}

vector<PrimitivePatNode*>
PrimitivePatNode::split() const {
   vector<PrimitivePatNode*> rv;
   rv.push_back(new PrimitivePatNode(*this));
   return rv;
}

bool 
PrimitivePatNode::hasSeqOps() const {
  return false;
}

bool 
PrimitivePatNode::hasAnyOrOther() const {
  return false;
}

bool PrimitivePatNode::
hasNeg() const {
  return false;
}

void PrimitivePatNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (ee_ != nullptr)
      use.insert(ee_);
   // %%%% We are not using the following check: rules are small enough that def
   // %%%% w/o use is likely obvious, plus (some) unused event args are common.
   if (params_ != nullptr)
      for (auto& param: *params_)
         assign.insert(param); // Matching binds parameters, so treat as asg.
   if (cond_ != nullptr)
      cond_->usedSyms(use, assign);
}

const Type* PrimitivePatNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
   if (ee_ == nullptr || ee_->symTab() == nullptr) 
	  internalErr("Unhandled case");
   else {
      unsigned i=0;
      for (auto ste: *(ee_->symTab())) {
         assert(ste->kind() == VARIABLE_KIND && i < params_->size());
         (*params_)[i]->type(ste->type());
         avlParams.insert((*params_)[i]);
         i++;
      }
      if (cond_ != nullptr) {
         cond_->typeCheck(avlParams, valVtypes, COND);
         if (!cond_->type() || cond_->type()->tag() != Type::BOOL)
            errMsg("Condition should have boolean type", cond_);
      }
   }
   type(new Type(Type::voidType));
   return type();
}

ExprNode* PrimitivePatNode::
getSubst(EventEntry* ee) {
   if (ee == ee_) 
      return cond_;
   else return new ValueNode(new Value(Type::BaseTypeTag(Type::BOOL), false));
}

void PrimitivePatNode::
codeGen(ostream &os, int indent, CGFlags& flags) {
/*
   if (flags.stateMachine()) {
      os << "ET_EVENT_RET " << name();
      emitST(os, indent, '(', ')', false);
      os << "{" << endl;
	  if (cond_ != nullptr)
		cond_->codeGen(os, indent+STEP_INDENT, flags);
      os << "{" << endl;
   }
*/
}

void
PrimitivePatNode::print(ostream &os, int indent) const {
	os << '(';
	if (ee_ == nullptr)
	  os << "nullptr()";
	else {
	  os << ee_->name();
	  if (params_ != nullptr) {
		unsigned int i = 0, n = params_->size();
		os << '(';
		for (;i < n; i++) {
		  if ((*params_)[i] != nullptr)
             os << (*params_)[i]->name();
		  if (i < n - 1) os << ", ";
		}
		os << ")";
	  }
	  if (cond_ != nullptr && !isTrue(cond_)) {
		os << '|';
		cond_->print(os, indent);
	  }
	  os << ')';
	}
}

/****************************************************************/
/*                          PAT NODE                            */
/****************************************************************/

PatNode::PatNode(int fl, int ln, int col, string f): 
  BasePatNode(BasePatNode::EMPTY, fl, ln, col, f) {
  pat1_ = pat2_ = nullptr;
}

PatNode::PatNode(PatNodeKind pk, BasePatNode *p1, BasePatNode* p2, 
				 int fl, int line, int column, string file):
  BasePatNode(pk, fl, line,column,file) {
  //
  // We can check for various degenerate cases like any||e1, which is
  // equivalent to any. But it is hard to do a good job, since that
  // would require consideration of the conditions associated with
  // primitive events. For this reason, it is best to perform such
  // checks after the EFSA is generated. 
  //
  pat1_ = p1;
  pat2_ = p2;
  if ((pk == BasePatNode::NEG) && (p1 != nullptr) && (!p1->isNegatable()))
		errMsgLn("Only simple patterns without `:', `*', and `!' operators"
                  "can be negated");
}

AstNode* 
PatNode::clone(CGFlags* f) const {
   assert(false);
   PatNode* rv = new PatNode(*this);
   if (pat1_ != nullptr)
       rv->pat1_ = (BasePatNode*)pat1_->clone(f);
   if (pat2_ != nullptr)
       rv->pat2_ = (BasePatNode*)pat2_->clone(f);
   return rv;
}

vector<PrimitivePatNode*>
PatNode::split() const {
   vector<PrimitivePatNode*> rv = pat1_->split();
   if (pat2_ != nullptr) {
      vector<PrimitivePatNode*> rv2 = pat2_->split();
      for (PrimitivePatNode* p: rv2)
         rv.push_back(p);
   }
   return rv;
}

bool
PatNode::hasNeg() const {
  if (kind() == NEG)
	return true;
  else if ((pat1_ != nullptr) && (pat1_->hasNeg()))
	return true;
  else if ((pat2_ != nullptr) && (pat2_->hasNeg()))
	return true;
  else return false;
}

bool
PatNode::hasSeqOps() const {
  if (kind() == BasePatNode::SEQ || kind() == BasePatNode::STAR ||
	  kind() == BasePatNode::PLUS)
	return true;
  else if ((pat1_ != nullptr) && (pat1_->hasSeqOps()))
	return true;
  else if ((pat2_ != nullptr) && (pat2_->hasSeqOps()))
	return true;
  else return false;
}

bool
PatNode::hasAnyOrOther() const {
  if ((pat1_ != nullptr) && (pat1_->hasAnyOrOther()))
	return true;
  else if ((pat2_ != nullptr) && (pat2_->hasAnyOrOther()))
	return true;
  else return false;
}

static const char* const patKindName[] = 
   {"", "empty", "!", "?", ":", " \\/ ", " /\\ ", "*", "+", "-", "=", "undef"};

void PatNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (pat1_ != nullptr)
      pat1_->usedSyms(use, assign);
   if (pat2_ != nullptr)
      pat2_->usedSyms(use, assign);
}

/*******************************************************************************
* A feature of the language is that it requires the programmer to be explicit
* when they want to remember an event argument across events. This is because it
* is necessary to store past event arguments in state variables, and we want the
* programmer to be aware of the introduction of state variables. Note that state
* variables not only increase the size of an individual match context (or state
* machine), but can also lead to an explosion in the number of match contexts
* that are active at any time. For this reason, only the most recent event's
* argument is available; access to previous events need to be remembered in rule
* variables if they are needed for access. This allows (the latest) event
* parameters to be accessed in the condition part of a primitive pattern, and
* also in the reaction part of a rule. Note that this requirement is ensured in
* the parser itself --- we eagerly exit the scope of an event as soon as we see
* a pattern operator, so we cannot resolve the names of event parameters past
* these operators. (For the prefix operator !, we exit the scope immediately
* after applying that operator.)
*
* For a disjunctive pattern, there can be two or more last events, so the
* arguments of any of these events can be accessed in the reaction part of a
* rule. Obviously, this won't make sense if the last events don't agree on
* argument bindings. In particular, only arguments that have the same name and
* type across these events can be accessed in the reaction. However, the
* scope-exiting trick can't ensure this. Instead, that trick simply ends up
* resolving identifiers in the reaction component in the scope of one of these
* last events that happens to occur textually at the end of the event pattern.
* So, to ensure correct semantics, additional work is needed during
* type-checking. In particular, we ensure: (a) only the variables that are bound
* in _all_ of the last events are available for use in the reaction, and (b)
* access is permitted only if the types match as well.
*******************************************************************************/

const Type* PatNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  set<const SymTabEntry*> avlParams1(avlParams);
  if (pat1_ != nullptr) {
     pat1_->typeCheck(avlParams1, valVtypes, tcc);
  }

  if (kind() == OR) { // Other binary ops unsupported (errmsg in constructor)
     set<const SymTabEntry*> avlParams2(avlParams);
     if (pat2_ != nullptr)
        pat2_->typeCheck(avlParams2, valVtypes, tcc);

     avlParams.clear();
     for (auto v1: avlParams1) {
        for (auto v2: avlParams2) {
           if (v1->name() != v2->name() 
               || v1->type() == nullptr || v2->type() == nullptr) continue;
           if (v1->type()->isEquiv(*v2->type())) {
              avlParams.insert(v1);
              avlParams.insert(v2);
           }
           else {
              warnMsg("Variable name \"" + v1->name() + "\" used to denote "
                        "variables of distinct types", v1);
            //warnMsg("Within a single event pattern, parameters with identical"
            //     " names must have identical types", v2);
           }
        }
     }
  }
  type(new Type(Type::voidType));
  return type();
}

ExprNode* PatNode::
getSubst(EventEntry* ee) {
   ExprNode* rv;
   if (kind() == NEG) {
      rv = pat1_->getSubst(ee);
      if (isFalse(rv)) // Nothing matches, so when we negate, it means
         rv = nullptr; // that there is no condition to check;
      else rv = new OpNode(OpNode::NOT, rv);
   }
   else if (kind() == OR) {
      auto rv1 = pat1_->getSubst(ee);
      auto rv2 = pat2_->getSubst(ee);
      if (isFalse(rv1))
         rv = rv2;
      else if (isFalse(rv2))
         rv = rv1;
      else {
      /* bool eqsubst = (rv1.first.size() == rv2.first.size());
         for (auto it: rv1.first)
            eqsubst = (eqsubst && rv2.first[it.first] == it.second);
         if (!eqsubst) {
            errMsg("Parameters must match across uses of event" + ee->name() +
                   " in a rule", this);
            rv.second = new ValueNode(new Value(Type::BOOL, false));
         }
         else {
            rv.first = rv1.first;
            rv.second = 
         } */
         rv = new OpNode(OpNode::OR, rv1, rv2);
      }
   }
   else {
      errMsg("Pattern type not (yet?) supported", this);
      rv = new ValueNode(new Value(Type::BaseTypeTag(Type::BOOL), false));
   }
   return rv;
}

void
PatNode::codeGen(ostream &os, int indent, CGFlags& f) {
   errMsg("Unexpected call of PatNode::codeGen", this);
}


void
PatNode::print(ostream &os, int indent) const {
  string pn = patKindName[kind()];
  os << '(';
  if (kind() == EMPTY)
	os << pn;
  else {
	if (kind() == NEG) os << pn;
	if (pat1_ != nullptr)
	  pat1_->print(os, indent);
	else os << "nullptr";
	if (kind() != NEG) {
	  os << patKindName[kind()];
	  if ((kind() != STAR) && (kind() != PLUS)) {
		  if (pat2_ != nullptr)
			pat2_->print(os, indent);
		  else os << "nullptr";
      }
	}
  }
  os << ')';
};

AstNode* 
ReturnStmtNode::clone(CGFlags* f) const {
   ReturnStmtNode* rv = new ReturnStmtNode(*this);
   if (expr_ != nullptr)
      rv->expr_ = (ExprNode*)expr_->clone(f);
   return rv;
}

void ReturnStmtNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (expr_ != nullptr) expr_->usedSyms(use, assign);
}

const Type* ReturnStmtNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  if (fun_ == nullptr) {
	errMsg("Return statement outside function", this);
  }
  else {
	const Type* rt = fun_->type()->retType();
    if (expr_ == nullptr) {
       if (rt->tag() != Type::VOID)
          errMsg("Return value expected for a non-void function", this);
    }
	else {
	  const Type *tp = expr_->typeCheck(avlParams, valVtypes, tcc);
      if (tp) {
         if (tp->isSubType(*rt, true)) // Like assignment, use relaxed check
            expr_->coerce(rt);
         else errMsg("Return value incompatible with current function's type",
                     this);
      }
    }
  }
  type(new Type(Type::voidType));
  return type();
}

void ReturnStmtNode::
codeGen(ostream& os, int indent, CGFlags& f) {
   os << "return "; 
   if (expr_ != nullptr) expr_->codeGen(os, indent, f);
   os << ';';
};

void ReturnStmtNode::
print(ostream& os, int indent) const {
   os << "return "; 
   if (expr_ != nullptr) expr_->print(os, indent); else os << "nullptr";
   os << ';';
};

/****************************************************************/

AstNode* 
ExprStmtNode::clone(CGFlags* f) const {
   ExprStmtNode* rv = new ExprStmtNode(*this);
   if (expr_ != nullptr)
      rv->expr_ = (ExprNode*)expr_->clone(f);
   return rv;
}

void ExprStmtNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (expr_ != nullptr) expr_->usedSyms(use, assign);
}

const Type* ExprStmtNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  if (expr_ != nullptr) {
	expr_->typeCheck(avlParams, valVtypes, tcc);
    if (expr_->type() && *expr_->type() != Type::errorType) {
       switch (expr_->exprNodeType()) {
       case REF_EXPR_NODE:
       case SIZEOF_NODE:
       case VALUE_NODE:
          errMsg("This expression is not a valid statement", this);
          break;

       case OP_NODE: {
          OpNode::OpCode oc = ((OpNode*)expr_)->opCode();
          expr_->coerce(type());
          if (opInfo[oc].typeConstraints[0] == '0')
             errMsg("You cannot ignore the value of this expression", this);
       }
       break;

       case INV_NODE: {
          const SymTabEntry *ste = ((InvocationNode*)expr_)->symTabEntry();
          if (ste != nullptr && ste->type() && ste->type()->isConst()) {
             bool c = true;
             for (unsigned j = 0; j < ste->type()->argTypes()->size(); j++)
                c = c && 
                   (Type::isPrimitive((*ste->type()->argTypes())[j]->tag()) ||
                   (*ste->type()->argTypes())[j]->isConst());
             if (c) 
               errMsg("Can't ignore return value from side-effect-free function",
                       this);
          }
       }

       default:
       break;
       }
    }
  }
  type(new Type(Type::voidType));
  return type();
}

void ExprStmtNode::
print(ostream& os, int indent) const { 
   if (expr_ != nullptr) 
      expr_->print(os, indent);
   endln(os, indent);
};

/****************************************************************/

AstNode* 
CompoundStmtNode::clone(CGFlags* f) const {
   CompoundStmtNode* rv = new CompoundStmtNode(*this);
   if (stmts_ != nullptr) {
      rv->stmts_ = new list<StmtNode*>;
      for (auto stmt: *stmts_)
         rv->stmts_->push_back((StmtNode*)stmt->clone(f));
   }
   return rv;
}

void CompoundStmtNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (decls_ != nullptr)
      decls_->usedSyms(use, assign);
   if (stmts_ != nullptr)
      for (auto& it: *stmts_)
         it->usedSyms(use, assign);
}

const Type* CompoundStmtNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  if (decls_)
     decls_->typeCheck(avlParams, valVtypes);
  if (stmts_)
    for (auto it = stmts_->begin(); it != stmts_->end(); ++it)
       (*it)->typeCheck(avlParams, valVtypes, tcc);
  type(new Type(Type::voidType));
  return type();
}

void
CompoundStmtNode::emitWithoutBraces(ostream &os, int ind, CGFlags& flags) const {
  if (decls_ && decls_->symTab() && decls_->symTab()->size() > 0)
     decls_->emitST(os, ind, '\0', ';', true, &flags);
  if (stmts_ != nullptr) {
    list<StmtNode*>::const_iterator iter = stmts_->begin();
    for(; iter != stmts_->end();) {
      if (flags.noneFlag())
         (*iter)->print(os, ind+STEP_INDENT);
      else (*iter)->codeGen(os, ind, flags);
/*
      if ( (*iter)->stmtNodeKind() != COMPOUND &&
		   (*iter)->stmtNodeKind() != IF &&
		   (*iter)->stmtNodeKind() != FORALL ) {
		if (++iter != stmts_->end())
		  endln(os, ind);
		else endln(os, ind);

	  }
	  else */ ++iter;
      //endln(os, ind);
    }
  }
}

void 
CompoundStmtNode::codeGen(ostream& os, int indent, CGFlags& flags) {
   os << " {";
   if (safeCoercions) {
      endln(os, indent); os << "tpvardecls"; 
   }
   if (decls_ && decls_->symTab() && decls_->symTab()->size() > 0)
     decls_->emitST(os, indent, '\0', ';', true, &flags);
   if ((stmts_ != nullptr) && (stmts_->size() != 0)) {
      indent += STEP_INDENT;
      prtln(os, indent);
      emitWithoutBraces(os, indent, flags);
      indent -= STEP_INDENT;
   }
   //prtRewind(os, STEP_INDENT, indent);
   //prtln(os, indent);
   os << "}";
   prtln(os, indent);
}

void
CompoundStmtNode::print(ostream &os, int indent) const {
  os << "{"; 
  if ((stmts_ != nullptr) && (stmts_->size() != 0)) {
	prtln(os, indent+STEP_INDENT);
    CGFlags fl("_NONE_");
    emitWithoutBraces(os, indent, fl);
    indent -= STEP_INDENT;
  }
  prtln(os, indent);
  os << "}";
  endln(os, indent);
}

/****************************************************************/

IfNode::IfNode(ExprNode* cond, StmtNode* thenStmt, 
	       StmtNode* elseStmt, int fl, int line, int column, string file)
  : StmtNode(StmtNode::IF, fl, line,column,file) {
  cond_ = cond; 
  then_ = thenStmt;
  else_ = elseStmt;
}


void IfNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (cond_) cond_->usedSyms(use, assign);
   if (then_) then_->usedSyms(use, assign);
   if (else_) else_->usedSyms(use, assign);
}

AstNode* 
IfNode::clone(CGFlags* f) const {
   IfNode* rv = new IfNode(*this);
   if (cond_ != nullptr)
      rv->cond_ = (ExprNode*)cond_->clone(f);
   if (then_ != nullptr)
      rv->then_ = (StmtNode*)then_->clone(f);
   if (else_ != nullptr)
      rv->else_ = (StmtNode*)else_->clone(f);
   return rv;
}

const Type* IfNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  if (cond_) {
	const Type *ct = cond_->typeCheck(avlParams, valVtypes, tcc);
	if (ct->tag() != Type::BOOL)
	  errMsg("Boolean argument expected", cond_);
  }

  if (then_)
	then_->typeCheck(avlParams, valVtypes, tcc);
  if (else_)
	else_->typeCheck(avlParams, valVtypes, tcc);

  type(new Type(Type::voidType));
  return type();
}

void
IfNode::codeGen(ostream &os, int indent, CGFlags& f) {
  os << "if (";
  if(cond_ != nullptr)
    cond_->codeGen(os, indent, f);
  os << ") ";

  if (then_ != nullptr) {
    indent += STEP_INDENT;
    os << "{"; prtln(os, indent);
	if (then_->stmtNodeKind() == StmtNode::COMPOUND)
       ((CompoundStmtNode*)then_)->emitWithoutBraces(os, indent, f);
	else then_->codeGen(os, indent, f);
    //prtRewind(os, STEP_INDENT, indent);
    indent -= STEP_INDENT;
    prtln(os, indent);
    os << "}";
    prtln(os, indent);
  }

  if (else_ != nullptr) {
    indent += STEP_INDENT;
    os << "else {"; prtln(os, indent);
	if (else_->stmtNodeKind() == StmtNode::COMPOUND)
       ((CompoundStmtNode*)else_)->emitWithoutBraces(os, indent, f);
	else else_->codeGen(os, indent, f);
    //prtRewind(os, STEP_INDENT, indent);
    indent -= STEP_INDENT;
    prtln(os, indent);
    os << "}";
  }
}

void
IfNode::print(ostream &os, int indent) const {
  os << "if (";
  if(cond_ != nullptr)
    cond_->print(os, indent);
  os << ") ";

  if (then_ != nullptr) {
	then_->print(os, indent);
	if (then_->stmtNodeKind() != StmtNode::COMPOUND) {
	  prtln(os, indent);
    } 
  }

  if (else_ != nullptr) {
    os << "else ";
	else_->print(os, indent);
	if (else_->stmtNodeKind() != StmtNode::COMPOUND) {
	  endln(os, indent);
	}
  }
}

/****************************************************************/

AstNode* 
ListForNode::clone(CGFlags* f) const {
   ListForNode* rv = new ListForNode(*this);
   if (indexVar_ != nullptr)
      rv->indexVar_ = (RefExprNode*)indexVar_->clone(f);
   if (list_ != nullptr)
      rv->list_ = (ExprNode*)list_->clone(f);
   if (body_ != nullptr)
      rv->body_ = (StmtNode*)body_->clone(f);
   return rv;
}

void ListForNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (indexVar_) indexVar_->usedSyms(assign, assign); // NOT a typo
   if (list_) list_->usedSyms(use, assign);
   if (body_) body_->usedSyms(use, assign);
}

const Type* ListForNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  const Type* t1 = list_->typeCheck(avlParams, valVtypes, tcc);
  Type* t1e = nullptr;
  if (t1)
     if (Type::isDict(t1->tag())) {
        auto t = new vector<Type*>();
        t->push_back(new Type(*t1->keyType()));
        t->push_back(new Type(*t1->elemType()));
        t1e = new Type(t, Type::TUPLE);
     }
     else t1e = new Type(*t1->elemType());

  if (!t1e)
    errMsg("Invalid type in foreach statement: List expected", this);
  else {
    indexVar_->type(t1e);
    indexVar_->symTabEntry()->type(t1e);
    body_->typeCheck(avlParams, valVtypes, tcc);
  }
  type(new Type(Type::voidType));
  return type();
}

void
ListForNode::codeGen(ostream &os, int indent, CGFlags& f) {
   if (indexVar_ != nullptr && list_ != nullptr) {
      //os << "auto __tp";
      //indexVar_->codeGen(os, indent, f);
      //os << " = ";
      //list_->codeGen(os, indent, f);
      //endln(os, indent);
      os << "for (auto ";
      indexVar_->codeGen(os, indent, f);
      os << ": "; 
      list_->codeGen(os, indent, f);
      //os << "__tp";
      //indexVar_->codeGen(os, indent, f);
      os << ")";
      indent += STEP_INDENT;
      os << "{"; prtln(os, indent);
      if (body_->stmtNodeKind() == StmtNode::COMPOUND)
         ((CompoundStmtNode*)body_)->emitWithoutBraces(os, indent, f);
      else body_->codeGen(os, indent, f);
      indent -= STEP_INDENT;
      prtln(os, indent);
      os << "}";
      prtln(os, indent);
   }
}

void
ListForNode::print(ostream &os, int indent) const {
  os << "foreach ";
  if(indexVar_ != nullptr)
    indexVar_->print(os, indent);
  else os << "nullptr";
  os << " in "; 
  if (list_ != nullptr)
    list_->print(os, indent);
  else os << "nullptr";
  os << " do ";
  if(body_ != nullptr)
    if (body_->stmtNodeKind() != StmtNode::COMPOUND) {
      endln(os, indent+STEP_INDENT);
      body_->print(os, indent+STEP_INDENT);
    } 
	else body_->print(os, indent);
  else os << "nullptr";
}

/****************************************************************/

AstNode* 
WhileNode::clone(CGFlags* f) const {
   WhileNode* rv = new WhileNode(*this);
   if (cond_ != nullptr)
      rv->cond_ = (ExprNode*)cond_->clone(f);
   if (body_ != nullptr)
      rv->body_ = (StmtNode*)body_->clone(f);
   return rv;
}

void WhileNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (cond_) cond_->usedSyms(use, assign);
   if (body_) body_->usedSyms(use, assign);
}

const Type* WhileNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypes,
          TypeCheckCtxt tcc) {
  if (cond_) {
	const Type *ct = cond_->typeCheck(avlParams, valVtypes, tcc);
	if (ct->tag() != Type::BOOL)
	  errMsg("Boolean argument expected", cond_);
  }

  if (body_)
     body_->typeCheck(avlParams, valVtypes, tcc);

  type(new Type(Type::voidType));
  return type();
}

void
WhileNode::codeGen(ostream &os, int indent, CGFlags& f) {
  os << "while (";
  if(cond_ != nullptr)
    cond_->codeGen(os, indent, f);
  os << ") ";

  if (body_ != nullptr) {
    indent += STEP_INDENT;
    os << "{"; prtln(os, indent);
	if (body_->stmtNodeKind() == StmtNode::COMPOUND)
       ((CompoundStmtNode*)body_)->emitWithoutBraces(os, indent, f);
	else body_->codeGen(os, indent, f);
    //prtRewind(os, STEP_INDENT, indent);
    indent -= STEP_INDENT;
    prtln(os, indent);
    os << "}";
    prtln(os, indent);
  }
}

void
WhileNode::print(ostream &os, int indent) const {
  os << "while (";
  if(cond_ != nullptr)
    cond_->print(os, indent);
  os << ") ";

  if (body_ != nullptr) {
	body_->print(os, indent);
	if (body_->stmtNodeKind() != StmtNode::COMPOUND) {
	  prtln(os, indent);
    } 
  }
}

/****************************************************************/

RuleNode::RuleNode(BlockEntry *re, BasePatNode* pat, 
		   StmtNode* reaction, int fl, int line, int column, string file)
  :AstNode(RULE_NODE, fl, line,column,file) {
  rste_ = re; 
  pat_ = pat; 
  reaction_ = reaction;
  label_ = nullptr;
}

AstNode* 
RuleNode::clone(CGFlags* f) const {
   RuleNode* rv = new RuleNode(*this);
   if (pat_ != nullptr)
      rv->pat_ = (PatNode*)pat_->clone(f);
   if (reaction_ != nullptr) {
      if (f->context() != "" && reaction_->stmtNodeKind() == StmtNode::EXPR) {
         ExprNode* e = ((ExprStmtNode*)reaction_)->expr();
         ExprNode* e1 = (ExprNode*)e->clone(f);
         ExprNode* e2 = new OpNode(OpNode::DUMMY_PROJECT, 
                                   new RefExprNode(nullptr, f->context()), e1);
         rv->reaction_ = new ExprStmtNode(e2);
      }
      else rv->reaction_ = (StmtNode*)reaction_->clone(f);
   }
   return rv;
}

vector<RuleNode*>
RuleNode::split() const {
   vector<PrimitivePatNode*> npats = pat_->split();
   vector<RuleNode*> rv;

   for (PrimitivePatNode* npat: npats) {
      RuleNode *rl = new RuleNode(*this);
      rl->pat_ = npat;
      rv.push_back(rl);
   }
   return rv;
}

void RuleNode::
usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (pat_) pat_->usedSyms(use, assign);
   if (reaction_) reaction_->usedSyms(use, assign);
}

// We prevent hasSeqOps patterns from using global variables. Otherwise, partial
// matches may need to be invalidated when these variables are updated. In
// addition, we enforce that rule variables cannot be updated.

const Type* RuleNode::
typeCheck(set<const SymTabEntry*>& avlParams, map<VarKind, bool>& valVtypesIgn,
          TypeCheckCtxt tcc) {
   set<const SymTabEntry*> avlPars;
   map<VarKind, bool> valVts;
/*
   if (pat_ != nullptr && pat_->hasSeqOps())
      valVts = {{FOREACH_VAR, true}, {FN_PARAM_VAR, false}, {LOCAL_VAR, true},
      {MODULE_PARAM_VAR, false}, {EVENT_PARAM_VAR, false}, {PATTERN_VAR, true},
      {RULE_VAR, true}};
   else
*/

  valVts = {/*{FOREACH_VAR, true}, {PARAM_VAR, false}, {LOCAL_VAR, true},
              {GLOBAL_VAR, false}, */ {MODULE_PARAM_VAR, false}, 
              {EVENT_PARAM_VAR, false}, 
              /*{PATTERN_VAR, true}, {RULE_VAR, true}*/};
  if (pat_ != nullptr)
     pat_->typeCheck(avlPars, valVts, tcc);

  valVts = {
      {FOREACH_VAR, false}, {FN_PARAM_VAR, false}, {LOCAL_VAR, true},
      {GLOBAL_VAR, true}, {MODULE_PARAM_VAR, true}, {EVENT_PARAM_VAR, false}, 
      {PATTERN_VAR, false}, {RULE_VAR, false}};
  if (reaction_ != nullptr)
	reaction_->typeCheck(avlPars, valVts, tcc);
  type(new Type(Type::voidType));
  return type();
}

void
RuleNode::codeGen(ostream &os, int indent, CGFlags& f)  {
   errMsg("RuleNode::codeGen should not be called", this);
}

void
RuleNode::print(ostream &os, int indent) const {
  if (label_ != nullptr) {
    if (*(label_->name().c_str()) != '_')
      os << label_->name() << ": ";
  }
  if (pat_ != nullptr) 
    pat_->print(os, indent);
  else os << "nullptr";
  os << " -->  ";
  if (reaction_ != nullptr)
    reaction_->print(os, indent);
  else os << "nullptr";

  //os << ";";
  //endln(os, indent);
}

/******************************************************************************/

