#ifndef AST_H
#define AST_H

/* The original version here has been moved to orig/. This version, from nBMSL,
   seems newer. It has (presumably) been rewritten to match revisions to the 
   language. There is also better refactoring of code. At the same time,
   type checking related functions have been commented out, which means that
   the type checker (or any of the subsequent compilation stages) haven't been 
   tested after these changes.
*/

#include "all.h"
#include "Value.h"
#include "ProgramElem.h"
#include "Error.h"
#include <set>
#include <map>
#include <bitset>
#include <assert.h>
#include <sstream>

using namespace std;
extern int paramStrictCheck; // Set to 0 to apply relaxed type compatibility for
// typechecking function calls, 1 to print a warning, 2 for printing an error.
extern int safeCoercions;    // Set to 0 to skip range checks during coercion

class BlockEntry;
class EFSA;
class EventEntry;
class FunctionEntry;
class LabelEntry;
class OpNode;
class PatNode;
class PrimitivePatNode;
class RulezNode;
class SymTabEntry;
class VariableEntry;
class ValueNode;
class MatchCond;

/*****************************************************************************
   Here is the class hierarchy:
                                               ProgramElem
											       |
                                                AstNode
     +--------------------------------------------+----------------+
     |		         |                 |                           | 
 BasePatNode      ExprNode          RuleNode                    StmtNode
     |               |                                             |
     |               |                                             |
     |      +---------+----------+-----------+------------+        |
     |      |         |          |           |            |        |
     |  RefExprNode  OpNode  SizeOfNode  ValueNode  InvocationNode |
     |                                                             |
     |                        +------------------------------------+------+
     |                        |                 |            |            |
     |                   ExprStmtNode   CompoundStmtNode   IfNode   ListForNode
     |
     |
   --+------------------+-----------------+---------------+
   |                    |                 |               |
   |                    |                 |               |
PrimitivePatNode    PatNode        PatRangeNode      PatAssignNode

******************************************************************************/

enum AstNodeType  {
   PAT_NODE,
   EXPR_NODE, 
   REACTION_NODE, 
   STMT_NODE, 
   SEQUENCE_NODE, 
   RULE_NODE,
   TRANSSPEC_NODE
};

enum TypeCheckCtxt {COND_LHS, COND_RHS, COND, NONCOND};

class AstNode: public ProgramElem {
 private: 
  AstNodeType nodeType_;
  const AstNode* operator=(const AstNode& other); /* disable asg */

 public: 
  AstNode(AstNodeType nt, int flags, int line=0, int column=0, string file="");
  virtual AstNode* clone(CGFlags* fl=0) const=0;
  virtual ~AstNode() {};


  AstNodeType nodeType() const { return nodeType_;}

  virtual void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign)=0;
  virtual const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
           map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND)=0;
  // avlParams is used to keep track of event parameters that can still
  // be accessed. Event parameters are removed from this list when event pattern
  // operators such as STAR and OR are processed. valVtypes is used to keep
  // track the VarType's of variables than can be accessed, and if the access is
  // limited to read-only. Finally, tcc is used to keep track of the context when
  // type-checking event pattern conditions.

  virtual void codeGen(ostream& os, int indent, CGFlags& flags)=0;
  virtual void print(ostream& os, int indent=0) const=0;
  string prtString() const {
    stringstream s;
    print(s);
    return s.str();
  }

  //virtual bool operator==(const AstNode&) const=0;
  //virtual bool operator!=(const AstNode& a) const 
  //{ return !operator==(a); };
};

inline ostream& operator<<(ostream& os, const AstNode& an) {
  an.print(os);
  return os;
};

enum ExprNodeType {
   REF_EXPR_NODE, 
   OP_NODE,
   SIZEOF_NODE,
   VALUE_NODE, 
   INV_NODE
};

/****************************************************************/

class ExprNode: public AstNode {
 public:
 private:
  ExprNodeType exprType_;
  const Value *val_; // reference semantics for val_ and coercedType_
  const Type* coercedType_; 

 public:
  ExprNode(ExprNodeType et, const Value* val=0, int fl=-1, int line=0, int col=0,
		   string file=""); // val is saved, but not copied
  //ExprNode(const ExprNode&);
  AstNode* clone(CGFlags* fl=0) const=0;
  virtual ~ExprNode() {};

  ExprNodeType exprNodeType() const { return exprType_;};
  void exprNodeType(ExprNodeType t) { exprType_ = t; };

  const Value* value() const { return val_; }
  void value(const Value* v) { val_ = v; };

  const Type* type() const { 
     return (coercedType_ == nullptr)? ProgramElem::type() : coercedType_;
  }
  void type(Type* t) { 
     if (coercedType_ != nullptr && t != nullptr && *t != *coercedType_)
        internalError("Attempt to modify type_ when coercedType_ is set", 
                      this->line(), this->file()); 
     ProgramElem::type(t);
  };

  const Type* coercedType() const { return coercedType_; }
  virtual void coerce(const Type* tp);
  virtual bool coercedTypeCodeGen(ostream& os, int indent, CGFlags& flags);
   // return value indicates if there was a coercion operation

  virtual bool isConst() const { return false;};
  virtual const Value* evalConst() {return nullptr;} // Eval constant subexprs
  // as much as possible. Returns a ValueNode if everything is eval'd, OW null
  virtual MatchCond* normalize(bool neg, ostream& os, int indent) {
     errMsg("Internal error: unexpected expression type in normalize()", this);
     return nullptr;
  };

  virtual bool checkRuleVarAsg() const { return false; };

  //-void buildAssignments(list<OpNode*>& localVarAsg, 
  //					list<OpNode*>& ruleVarAsg) const;
  //-ExprNode* buildCondition() const;
  //-void actualParamsUsed(set<int>&) const;

};

/****************************************************************/
class RefExprNode: public ExprNode {
 private:
  RefExprNode* base_;
  string ext_;
  vector<ExprNode*> *subscript_;
  bitset<32> tupleSubs_; // @@@@ Max 32 subscripts!
  ExprNode* constraint_;
  SymTabEntry* sym_;

 public:
  RefExprNode(RefExprNode* base, string ext, SymTabEntry* ste=nullptr, int f=-1,
	      int line=0, int column=0, string file="");
  //RefExprNode(const RefExprNode&);
  AstNode* clone(CGFlags* fl=0) const;
  ~RefExprNode() {};


  const RefExprNode *base() const { return base_;};
  void base(RefExprNode* base) { base_ = base; };

  string ext() const { return ext_;};
  void ext(string str) { ext_ = str;}; 
  string fullName() const {
    if (base_ == nullptr) return ext_;
    else return base_->fullName() + "." + ext_;
  };

  const SymTabEntry* symTabEntry() const { return sym_;};
  SymTabEntry* symTabEntry() { return sym_;};
  void symTabEntry(SymTabEntry *ste)  { sym_ = ste;};
  const VariableEntry* root() const;

  unsigned nsubs() const { return (subscript_? subscript_->size() : 0); };
  const ExprNode* subscript(unsigned int i) const 
    { return (subscript_ != nullptr && i < subscript_->size())? 
	(const ExprNode*)((*subscript_)[i]) : nullptr;  };
  ExprNode* subscript(unsigned int i)
    { return (subscript_ != nullptr && i < subscript_->size())? 
	     (*subscript_)[i] : nullptr;  };
  const vector<ExprNode*>* subscript() const { return subscript_; }
  vector<ExprNode*>* subscript() { return subscript_; };
  void addSubs(ExprNode* e); 

  const ExprNode* constraint() const { return constraint_; };
  ExprNode* constraint() { return constraint_; };
  void constraint(ExprNode* constraint) { constraint_ = constraint; };

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                    map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);
  
  const Value* evalConst();
  MatchCond* normalize(bool neg, ostream& os, int indent);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;
};

/****************************************************************/

class OpNode: public ExprNode {
 public:
  enum OpCode {
    UMINUS, PLUS, MINUS, MULT, DIV, MOD, 
    EQ, NE, GT, LT, GE, LE,
    EQREG, NEQREG,
    AND, OR, NOT, SEQ_AND, SEQ_OR, 
    BITNOT, BITAND, BITOR, BITXOR, 
    SHL, SHR,
    ASSIGN,
    EMPTY, CLEAR, LENGTH, APPEND, UNION, INS_SET, 
    INS_VEC, ERASE, CONTAINS, REMOVE, CAT,
    TUPLE, VECTOR, SET, DICT, 
    PRINT, PRINT_RAW, 
    DUMMY_PROJECT, // hack used to pass info from instantiation to codegen phase
    EQBIT,
    NEQBIT,
    INTERVAL,
    INVALID
  };


  static const int InverseOpCode[]; 
    
  enum OpPrintType {PREFIX, INFIX, POSTFIX};
  struct OpInfo {
    OpCode code_;
    const char* name_;
    int arity_;
	int needParen_;
    OpPrintType prtType_;
    Type::BaseTypeTag argType_[3]; // argType_[2] used for args 2..N
    Type::BaseTypeTag outType_;
    const char* typeConstraints;
    //@@@@ void* (*evalFn)(void*, void*);

	Type::BaseTypeTag argType(int i) const {
	  if (i >= arity_) return Type::ERROR;
	  if (i > 2) i = 2;

      for (; i >= 0; i--)
         if (argType_[i] > 0) return argType_[i];
      return Type::ERROR;
	}
  };

 private: 
  unsigned int arity_;
  OpCode   opCode_;
  vector<ExprNode*> arg_;

 public:
  static const int VARIABLE = 100;

 public:
  OpNode(OpCode op, ExprNode *l, ExprNode *r=nullptr, int fl=-1,
	 int line=0, int column=0, string file="");
  OpNode(OpCode op, vector<ExprNode*>* args, int fl=-1,
	 int line =0, int column=0, string file=""); // for list or tuple type
  //OpNode(const OpNode&);
  AstNode* clone(CGFlags* fl=0) const;
  ~OpNode() {};


  OpCode opCode() const { return opCode_;};
  const ExprNode* arg(unsigned int i) const 
    { return (i < arg_.size())? arg_[i] : nullptr; };
  unsigned int arity() const { return arity_; }

  void opCode(OpCode a) { opCode_ = a; };
  ExprNode* arg(unsigned int i) 
    { return (i < arg_.size())? arg_[i] : nullptr; };
  vector<ExprNode*>* args() 
    { return &arg_; }

  virtual bool checkRuleVarAsg() const;

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void coerce(const Type* type);

  bool isConst() const {
     for (unsigned i=0; i < arg_.size(); i++)
        if (!arg_[i]->isConst()) return false;
     return true;
  };
  const Value* evalConst();
  MatchCond* normalize(bool neg, ostream& os, int indent);
  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;  
};

extern const OpNode::OpInfo opInfo[];

/****************************************************************/

class SizeOfNode: public ExprNode {
 private:
  const SymTabEntry* ste_;

 public:
  SizeOfNode(const SymTabEntry* ste, int fl=-1, int ln=0, int cl=0, string f="")
    : ExprNode(SIZEOF_NODE, 0, fl, ln,cl, f) { ste_ = ste; }
  //SizeOfNode(const SizeOfNode& e):
  //  ExprNode(e) { ste_ = e.ste_; };
  AstNode* clone(CGFlags* fl=0) const { return new SizeOfNode(*this); }
  ~SizeOfNode() {};

  const SymTabEntry* symTabEntry() const { return ste_; };
  //void symTabEntry(const SymTabEntry* ste) { ste_ = ste; };

  bool isConst() const { return true;};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {};
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;
};

/****************************************************************/

class ValueNode: public ExprNode {
 private:
  /* val_ field is already included in ExprNode, so no new data members */

 public:
  ValueNode(const Value* val=0, int fl=-1, int line=0, int col=0, string file="")
    : ExprNode(VALUE_NODE, val, fl, line, col, file) {
    if (val != nullptr) type((Type*)(val->type()));
  }
  ValueNode(const Value* val, AstNode* t)
    : ExprNode(VALUE_NODE, val, -1, t->line(), t->column(), t->file()) {
    if (val != nullptr) type((Type*)(val->type()));
  }
  //ValueNode(const ValueNode& val): ExprNode(val) {};
  AstNode* clone(CGFlags* fl=0) const { return new ValueNode(*this); }
  ~ValueNode() {};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {};
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);
  
  bool isConst() const { return true;};
  const Value* evalConst() { return value(); };
  MatchCond* normalize(bool neg, ostream& os, int indent);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;

};

static const ValueNode falseExpr = 
   ValueNode(new Value(Type::BaseTypeTag(Type::BOOL), false));
static const ValueNode  trueExpr = 
   ValueNode(new Value(Type::BaseTypeTag(Type::BOOL), true));
inline static bool isFalse(const ExprNode* e) {
   return (e != nullptr && e->exprNodeType() == VALUE_NODE && 
           e->value() != nullptr && 
           *e->value() == Value(Type::BaseTypeTag(Type::BOOL), false));
}
inline static bool isTrue(const ExprNode* e) {
   return (e == nullptr || 
           (e->exprNodeType() == VALUE_NODE && e->value() != nullptr && 
            *e->value() == Value(Type::BaseTypeTag(Type::BOOL), true)));
}

/****************************************************************/

class InvocationNode: public ExprNode {
  // Used to represent function invocation, module instantiation and 
  // emitting of output events
 private:
  vector<ExprNode*>* params_;
  SymTabEntry *ste_; // reference semantics

 public:
  InvocationNode(SymTabEntry *ste, vector<ExprNode*>* param=0, int fl=-1, 
		 int line=0, int column=0, string file="");
  //InvocationNode(const InvocationNode&);
  AstNode* clone(CGFlags* fl=0) const;
  ~InvocationNode() {};

  const SymTabEntry* symTabEntry() const { return ste_; };
  SymTabEntry* symTabEntry() { return ste_; };

  const vector<ExprNode*>* params() const { return params_;};
  vector<ExprNode*>* params() { return params_;}
  void params(vector<ExprNode*>* args){ params_ = args;};
  const ExprNode* param(unsigned int i) const 
    { return (params_ != nullptr && i < params_->size())? (const ExprNode*)((*params_)[i]) : nullptr; };
  ExprNode* param(unsigned int i)
    { return (ExprNode*)(((const InvocationNode*)this)->param(i));}
  void param(ExprNode* arg, unsigned int i) 
    { if (params_ != nullptr && i < params_->size()) (*params_)[i] = arg;};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  const Value* evalConst();
  MatchCond* normalize(bool neg, ostream& os, int indent);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;
};

/****************************************************************/
class BasePatNode: public AstNode { // Abstract base class
 public:
  enum PatNodeKind {PRIMITIVE, EMPTY, NEG, QMARK, SEQ, OR, AND, STAR, PLUS, 
					RANGE, ASSIGN, UNDEFINED};

 private:
  PatNodeKind patKind_;
  BasePatNode* parent_;
  BasePatNode* root_;

 public:
  BasePatNode(PatNodeKind pk, int fl=-1, int ln=0, int col=0, string f=""):
	AstNode(PAT_NODE, fl, ln, col, f) {
       parent_ = nullptr; root_ = nullptr; patKind_ = pk;
  };
  //BasePatNode(const BasePatNode& bpn): AstNode(bpn) {
  //	patKind_ = bpn.patKind_; parent_ = bpn.parent_; root_ = bpn.root_;
  //}
  AstNode* clone(CGFlags* fl=0) const=0;
  virtual vector<PrimitivePatNode*> split() const=0;
  ~BasePatNode() {};

  PatNodeKind kind() const { return patKind_; };
  void kind(PatNodeKind p) {patKind_ = p;}

  const BasePatNode* parent() const { return parent_; } 
  BasePatNode* parent() { return parent_;} 

  virtual bool hasSeqOps() const {return false;};
  virtual bool hasNeg() const { return false;};
  virtual bool hasAnyOrOther() const {return false;};
  virtual bool isNegatable() const { return ((!hasSeqOps()) && (!hasNeg())); }

  virtual ExprNode* getSubst(EventEntry*)=0;

  const Value* evalConst() {return nullptr;};
};

/****************************************************************/
//   PrimitivePatNodes are of the form: event|cond
class PrimitivePatNode: public BasePatNode {
 private:
  EventEntry* ee_;
  vector<VariableEntry*>* params_;
  ExprNode* cond_;      

 public:
  PrimitivePatNode(EventEntry* e, vector<VariableEntry*>* params=nullptr, 
                   ExprNode* c=0, int fl=-1, int ln=0, int col=0, string f="");
  //PrimitivePatNode(const PrimitivePatNode& ppn);
  AstNode* clone(CGFlags* f=0) const;
  vector<PrimitivePatNode*> split() const;
  ~PrimitivePatNode() {};

  const EventEntry* event() const { return ee_; }
  EventEntry* event() { return ee_; }

  const vector<const VariableEntry*>* params() const { 
	return (vector<const VariableEntry*>*)params_; }
  vector<VariableEntry*>* params() { return params_; }
  void addParam(VariableEntry* ve) { params_->push_back(ve); };
  unsigned nparams() const { return params_->size(); };

  const ExprNode* cond() const { return cond_; }
  ExprNode* cond() { return cond_; }
  void cond(ExprNode* c) { cond_ = c;}

  bool hasSeqOps() const;
  bool hasNeg() const;
  bool hasAnyOrOther() const;

  ExprNode* getSubst(EventEntry*);

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  const Value* evalConst() { return (cond_? cond_->evalConst() : nullptr); }

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const; 
};

/****************************************************************/
class PatNode: public BasePatNode {
 private:
  BasePatNode *pat1_;
  BasePatNode *pat2_;

 public: 
  PatNode(int fl=-1, int line=0, int column=0, string file="");
  PatNode(PatNodeKind pk, BasePatNode *p1, BasePatNode*p2=nullptr, int fl=-1,
          int line=0, int column=0, string file="");
  // PatNode(const PatNode&); // @@@@ Inconsistent: default copy constructor
  AstNode* clone(CGFlags* f=0) const;
  vector<PrimitivePatNode*> split() const;
  ~PatNode() {};

  static BasePatNode* 
	strToPat(char* pat, int line=0, int column=0, string file="");

  const BasePatNode* pat1() const { return pat1_; }
  BasePatNode* pat1() { return pat1_; }
  const BasePatNode* pat2() const { return pat2_; }
  BasePatNode* pat2() { return pat2_; }

  bool hasNeg() const;
  bool hasSeqOps() const;
  bool hasAnyOrOther() const;

  ExprNode* getSubst(EventEntry*);

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const; 
};


/****************************************************************/

// @@@@ Unused as yet ...
/*
class PatRangeNode: public BasePatNode {
 private:
  unsigned max_;
  vector<unsigned> lo_;
  vector<unsigned> hi_;

 public:
  PatRangeNode(unsigned c1, unsigned c2, unsigned max, 
			   int line=0, int col=0, string file="");
  ~PatRangeNode() {};

  void merge(PatRangeNode&);
  void neg();

  bool hasSeqOps() const { return false; };
  bool hasNeg() const { return false; };
  bool hasAnyOrOther() const {return false;};

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;
};
*/
/****************************************************************/

// @@@@ Unused as yet ...
/*
class PatAssignNode: public BasePatNode {
 private:
  VariableEntry* var_;
  BasePatNode*   pat_;
  ExprNode*      cond_;

 public:
  PatAssignNode(VariableEntry* var, BasePatNode* pat, ExprNode* cond=nullptr,
				int line=0, int column=0, string file="");
  ~PatAssignNode() {};

  const VariableEntry* var() const { return var_; };
  VariableEntry* var() { return var_; };
  const BasePatNode* pat() const { return pat_; }
  BasePatNode* pat() { return pat_; }
  const ExprNode* cond() const { return cond_; }
  ExprNode* cond() { return cond_; }

  bool hasSeqOps() const { return ((pat_ != nullptr) && (pat_->hasSeqOps())); };
  bool hasNeg() const { return ((pat_ != nullptr) && (pat_->hasNeg())); };
  bool hasAnyOrOther() const {return((pat_!=nullptr)&&(pat_->hasAnyOrOther()));};

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const; 

 private:
  PatAssignNode(const PatAssignNode&);
};
*/
/****************************************************************/
class StmtNode: public AstNode {
 public:
  enum StmtNodeKind { ILLEGAL=-1, EMPTY, EXPR, IF, COMPOUND, FORALL, WHILE, RETURN};

 private:
  StmtNodeKind skind_;

 public: 
  StmtNode(StmtNodeKind skm, int fl=-1, int line=0, int col=0, string file=""):
	AstNode(STMT_NODE, fl, line,col,file) { skind_ = skm; };
  StmtNode(int fl=-1, int line=0, int column=0, string file=""):
	AstNode(STMT_NODE, fl, line,column,file) { skind_ = EMPTY; };
  AstNode* clone(CGFlags* f=0) const { return new StmtNode(*this); };
  ~StmtNode() {};

  StmtNodeKind stmtNodeKind() const { return skind_;}

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {};
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND) 
     { type(new Type(Type::voidType)); return type(); };
  void codeGen(ostream& os, int indent, CGFlags& flags) { os << ";" ;};
  void print(ostream& os, int indent) const { os << ";" ;};
};

/****************************************************************/

class ReturnStmtNode: public StmtNode {
 private:
  ExprNode* expr_;
  FunctionEntry* fun_;

 public:
  ReturnStmtNode(ExprNode *e, FunctionEntry* fe, int fl=-1, 
				 int line=0, int column=0, string file=""):
    StmtNode(StmtNode::RETURN, fl, line,column,file) { expr_ = e; fun_ = fe;};
  AstNode* clone(CGFlags* f=0) const;
  ~ReturnStmtNode() {};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent) const;
};

/****************************************************************/
class ExprStmtNode: public StmtNode {
 private:
  ExprNode* expr_;

 public:
  ExprStmtNode(ExprNode* e, int fl=-1, int line=0, int column=0, string file=""):
    StmtNode(StmtNode::EXPR, fl, line,column,file) { expr_ = e; };
  AstNode* clone(CGFlags* f=0) const;
  ~ExprStmtNode() {};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  ExprNode* expr() { return expr_; };

  void codeGen(ostream& os, int indent, CGFlags& flags) 
    {expr_->codeGen(os, indent, flags); /*os << ';'*/ endln(os, indent);}
  void print(ostream& os, int indent) const;
};

/****************************************************************/
class CompoundStmtNode: public StmtNode {
 private:
  BlockEntry* decls_;
  list<StmtNode*>   *stmts_;

 public: 
  CompoundStmtNode(BlockEntry* decls, list<StmtNode*> *s, 
                   int fl=-1, int l=0, int c=0, string f=""):
	StmtNode(StmtNode::COMPOUND, fl, l, c, f) { decls_ = decls; stmts_ = s;};
  AstNode* clone(CGFlags* f=0) const;
  ~CompoundStmtNode() { };

  const BlockEntry* decls() const { return decls_; };
  const list<StmtNode*>* stmts() const { return stmts_;}

  BlockEntry* decls() { return decls_; };
  list<StmtNode*>* stmts() { return stmts_;}
  void add(StmtNode *s) 
    { if(stmts_ != nullptr) stmts_->push_back(s); };

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);
  void  emitWithoutBraces(ostream& os, int indent, CGFlags& flags) const;

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void  print(ostream& os, int indent) const;
};

/****************************************************************/

class IfNode: public StmtNode {
 private: 
  ExprNode *cond_;
  StmtNode *then_, *else_;

 public: 
  IfNode(ExprNode* cond, StmtNode* thenStmt, StmtNode* elseStmt=nullptr, 
         int fl=-1, int line=0, int column=0, string file="");
  // IfNode(const IfNode&); // @@@@ Inconsistent: default copy constructor
  AstNode* clone(CGFlags* f=0) const;
  ~IfNode(){};

  const ExprNode* cond() const {return cond_;}
  const StmtNode* elseStmt() const { return else_;};
  const StmtNode* thenStmt() const  { return then_;};

  ExprNode* cond() {return cond_;}      
  StmtNode* elseStmt() { return else_;};
  StmtNode* thenStmt() { return then_;};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent) const;
};

/****************************************************************/

class ListForNode: public StmtNode {
 private:
  RefExprNode *indexVar_;
  ExprNode *list_;
  StmtNode *body_;

 public: 
  ListForNode(RefExprNode* index, ExprNode* list, StmtNode* body=nullptr, 
			  int fl=-1, int line=0, int column=0, string file=""):
	StmtNode(StmtNode::FORALL, fl, line,column,file) {
	indexVar_ = index; list_ = list; body_ = body; }
  // ListForNode(const IfNode&); // @@@@ Inconsistent: default copy constructor
  AstNode* clone(CGFlags* f=0) const;
  ~ListForNode() {};

  const RefExprNode *indexVar() const { return indexVar_; };
  RefExprNode *indexVar() { return indexVar_; };

  const ExprNode *list() const { return list_; };
  ExprNode *list() { return list_; };

  const StmtNode* body() const { return body_; };
  StmtNode* body() { return body_; };
  void body(StmtNode* b) { body_ = b; };
  
  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent) const;
};

/****************************************************************/

class WhileNode: public StmtNode {
 private:
  ExprNode *cond_;
  StmtNode *body_;

 public: 
  WhileNode(ExprNode* cond, StmtNode* body=nullptr, 
			  int fl=-1, int line=0, int column=0, string file=""):
	StmtNode(StmtNode::WHILE, fl, line,column,file) {
	cond_ = cond; body_ = body; }
  // ListForNode(const IfNode&); // @@@@ Inconsistent: default copy constructor
  AstNode* clone(CGFlags* f=0) const;
  ~WhileNode() {};

  const ExprNode *cond() const { return cond_; };
  ExprNode *cond() { return cond_; };

  const StmtNode* body() const { return body_; };
  StmtNode* body() { return body_; };
  void body(StmtNode* b) { body_ = b; };
  
  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent) const;
};

/****************************************************************/

class RuleNode: public AstNode {
 private:
  BlockEntry    *rste_;
  LabelEntry    *label_;
  BasePatNode *pat_;
  StmtNode *reaction_;
   
 public:
  RuleNode(BlockEntry *re, BasePatNode* pat, StmtNode* reaction, int fl=-1, 
	   int line=0, int column=0, string file="");
  // RuleNode(const RuleNode&); // @@@@ Inconsistent: default copy constructor
  AstNode* clone(CGFlags* f=0) const;
  vector<RuleNode*> split() const;
  ~RuleNode() {};

  const BlockEntry* ruleEntry() const { return rste_; };
  BlockEntry* ruleEntry() { return rste_; };

  const BasePatNode* pat() const { return pat_; };
  BasePatNode* pat() { return pat_; };              

  const StmtNode* reaction() const { return reaction_; };   
  StmtNode* reaction() { return reaction_; };
  void reaction(StmtNode* r) { reaction_ = r; };

  const LabelEntry* label() const { return label_;} 
  LabelEntry* label() { return label_;} 
  void label(LabelEntry* le) { label_ = le; };

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes, TypeCheckCtxt tcc=NONCOND);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;
};

/****************************************************************/
#endif
