%option noyywrap
%option never-interactive

%{
#include "ParserUtil.h"
#include "Ast.h"
#include "BMSL.tab.H"
#include "RegExpr.h"

int last_ret;

#ifdef DEBUG_LEXER
extern ostream& lexOut;
#define yylex_ret(x)  do {       lexOut << #x << endl; return x; } while(0)
#define yylex_retval(x, y) do { lexOut << #x << "\t``"<< y << "''\n"; return x; } while(0)
#else
#define yylex_ret(x) \
    do { last_ret = x; return x; } while(0)
#define yylex_retval(x, y) yylex_ret(x)
#endif

#define yylex_retstr { \
  BEGIN 0; \
  yylval.cVal = new char[string_buf_ptr - string_buf + 1]; \
  strncpy(yylval.cVal, string_buf, string_buf_ptr - string_buf); \
  yylval.cVal[string_buf_ptr - string_buf] = '\0'; \
  yylex_retval(TOK_STRCONST, yylval.cVal); \
}

extern bool expectingStmt;
const char* yyfilename="";
char string_buf[10000];
char *string_buf_ptr;
int tabWidth=8;

static CharClass charClass;
bool lex_dquote, lex_dquoteRegexp, lex_squoteRegexp;

static char*
copyStr(int begSkipLen=0, int endSkipLen=0) {
  int slen = yyleng+1-begSkipLen-endSkipLen;
  char *rv = new char[slen];
  strncpy(rv, yytext+begSkipLen, slen-1);
  rv[slen-1] = '\0';
  return rv;
}

static bool
strbuf_space_avl(int s) {
  bool rv = (string_buf_ptr + s) < (string_buf + sizeof(string_buf) - 2);
  if (!rv)
    errMsg("String constant too long");
  return rv;
}

int tokBegCol, tokEndCol, tokBegLn, tokEndLn;
int errTokBegLn, errTokEndLn, errTokBegCol, errTokEndCol;
extern int yyrecovering();
int lparen, lbrace, lbrack;

static void
chkParenMismatch() {
  if (lparen <= 0) {
    errMsg("Too many closing parentheses");
    lparen = 0;
  }
  else lparen--;
}

static void
chkBrackMismatch() {
  if (lbrack <= 0) {
    errMsg("Too many closing brackets");
    lbrack = 0;
  }
  else lbrack--;
}

static void
chkBraceMismatch() {
  if (lbrace <= 0) {
    errMsg("Too many closing braces");
    lbrace = 0;
  }
  else lbrace--;
}

static void
chkRuleScope() {
  if (last_ret == TOK_RBRACE || last_ret == TOK_EV_IDENT)
    return; /* May be the current token is a lookahead token, in which case,
               the scope info won't be correct. So, we skip the check. */
  if (stm.currentScope(RULE_BLOCK_KIND) == NULL)
      errMsg("\"->\" symbol outside a rule");
}

static void
chkGlobalScope(const char* sym) {
  if (last_ret == TOK_RBRACE)
    return; /* May be the current token is a lookahead token, in which case,
               the scope info won't be correct. So, we skip the check. */
  else if ((strcmp(sym, "interface")==0) && (last_ret == TOK_USE))
    return; /* This is syntactically correct */
  if (stm.currentScope()->kind() != GLOBAL_KIND) 
    errMsg(string("Keyword \"") + sym + "\" expected at the top level.");
}

static void
advanceCol() {
  tokBegCol = tokEndCol; tokBegLn = tokEndLn;
  yycolumnno += yyleng; 
  tokEndCol = yycolumnno-1; tokEndLn = yylineno;
}

static void
tabAdvanceCol() {
  tokBegCol = tokEndCol; tokBegLn = tokEndLn;
  yycolumnno = (yycolumnno/tabWidth+1)*tabWidth;
  tokEndCol = yycolumnno-1; tokEndLn = yylineno;
}

static void
advanceLn() {
  yylineno++; yycolumnno = 0;
  tokBegLn = tokEndLn = yylineno;
  tokBegCol = tokEndCol = yycolumnno;
}

static char 
check_and_convert(int base, const char* s) {
   unsigned c =  strtoul(s, NULL, base);
   if (c >= REGEXP_CHMAX) {
      errMsg("Invalid escape sequence");
      c = REGEXP_CHMAX-1;
   }
   return c;
}

static void
check_and_strbuf_spl_char(unsigned long c) {
  if (c < REGEXP_CHMAX) {
    if (strbuf_space_avl(yyleng))
     *string_buf_ptr++ = c;
  } 
  else errMsg("Invalid escape sequence");
}

%}
/* Definitions */

BINARY_DIGIT   [01]
OCTAL_DIGIT    [0-7]
DECIMAL_DIGIT  [0-9]
HEX_DIGIT      [0-9a-fA-F]

BINARY         {BINARY_DIGIT}+(b|B)
OCTNUM         0{OCTAL_DIGIT}+
HEXNUM         0x{HEX_DIGIT}+
UINTNUM        {DECIMAL_DIGIT}+
 /* Only unsigned numbers are recognized by the lexer. Signed numbers are 
    recognized by the parser, or else you have the problem that something 
    like 3-2 will be recognized as 3(-2) instead of (3) - (2)
 */

MANTISSA_EINT  {DECIMAL_DIGIT}*[.]{DECIMAL_DIGIT}+
MANTISSA_EFRAC {DECIMAL_DIGIT}+(([.]{DECIMAL_DIGIT}*)?)
FLOAT_MANTISSA ({MANTISSA_EINT}|{MANTISSA_EFRAC})
FLOAT_EXP      [Ee]([+-]?){DECIMAL_DIGIT}+
FLOATNUM       {FLOAT_MANTISSA}({FLOAT_EXP}?)

ALPHA           [_A-Za-z] 
IDENT           {ALPHA}({ALPHA}|{DECIMAL_DIGIT})*

IP_DECIMAL ({DECIMAL_DIGIT}+[.]){3}{DECIMAL_DIGIT}+

ONELINE_COM_BEG 	"//"
MULTI_LINE_COM_BEG 	"/*"
MULTI_LINE_COM_END 	"*/"

NEWLINE     "\n"

%x STRING INCLUDE FSTART SKIPTOEOL COMMENT TO_EOL_COMMENT REGEXP REGEXP_BEG CHARCLASS
%%

^"#"   {
  BEGIN INCLUDE;
};

 /*********** Remember file/line number info in cpp output ********/

<INCLUDE>[ ][0-9]+[ ]+  {
  yylineno = atoi(yytext+1) - 1;
  BEGIN FSTART;
  yycolumnno = 1;
}

<FSTART>\"([^"]+)\" {
  yyfilename = copyStr(1, 1);
  BEGIN SKIPTOEOL;
}

<SKIPTOEOL>[^\n]*\n {
    advanceLn();
    BEGIN 0;
}

 /**************************** Comment processing ****************************/

{ONELINE_COM_BEG} {
  advanceCol();
  BEGIN TO_EOL_COMMENT;
}

<TO_EOL_COMMENT>"\n" {
   advanceLn();
   BEGIN 0;
}

<TO_EOL_COMMENT>. {
   advanceCol();
}

{MULTI_LINE_COM_BEG} {
   BEGIN COMMENT;
   advanceCol();
}

<COMMENT>{MULTI_LINE_COM_END} {
   advanceCol();
   BEGIN 0;
}

<COMMENT>.|"\n" {
   if (*yytext == '\n') advanceLn();
}

 /**************** String processing ****************/

\" { 
  string_buf_ptr = string_buf;
  lex_dquote = true;
  advanceCol();
  BEGIN(STRING); 
}

\' { 
  string_buf_ptr = string_buf;
  lex_dquote = false;
  advanceCol();
  BEGIN(STRING); 
}

 /* Handle octal/hex codes to input special characters */

<STRING>\\{OCTAL_DIGIT}{1,3} { 
  check_and_strbuf_spl_char(strtoul((yytext+1), NULL, 8));
  advanceCol();
}

<STRING>\\x{HEX_DIGIT}{1,2} { 
  check_and_strbuf_spl_char(strtoul(yytext+2, NULL, 16));
  advanceCol();
}

<STRING>\\. { 
  char c = *(yytext+1);
  switch (c) {
    case 'a': c = '\a'; break;
    case 'b': c = '\b'; break;
    case 'f': c = '\f'; break;
    case 'n': c = '\n'; break;
    case 'r': c = '\r'; break;
    case 't': c = '\t'; break;
    case 'v': c = '\v'; break;
    default: /* treat \x as x when none of the above cases apply */
      break; /* This applies to \\, \", \?, \' and so on.        */
  }
  if (strbuf_space_avl(yyleng))
     *string_buf_ptr++ = c;
  advanceCol();
} 

<STRING>\" { 
  advanceCol();
  if (lex_dquote) yylex_retstr
  else if (strbuf_space_avl(yyleng))
     *string_buf_ptr++ = *yytext;
}

<STRING>\' { 
  advanceCol();
  if (!lex_dquote) yylex_retstr
  else if (strbuf_space_avl(yyleng))
     *string_buf_ptr++ = *yytext;
}

<STRING>\n { 
  unput('\n');
  errMsgLn("String constant cannot span multiple lines");
  yylex_retstr;
}

<STRING>. { 
  if (strbuf_space_avl(yyleng))
     *string_buf_ptr++ = *yytext;
  advanceCol();
}

 /**************** RegExp processing ****************/

[rR]/['"] {
   BEGIN REGEXP_BEG;
}

<REGEXP_BEG>[ \t]*\" {
   advanceCol();
   lex_dquoteRegexp = true; lex_squoteRegexp = false;
   BEGIN REGEXP;
}

<REGEXP_BEG>[ \t]*\' {
   advanceCol();
   lex_squoteRegexp = true; lex_dquoteRegexp = false;
   BEGIN REGEXP;
}

<REGEXP_BEG>[ \t]*/[^\"\'] {
   advanceCol();
   lex_squoteRegexp = false; lex_dquoteRegexp = false;
   BEGIN REGEXP;
}

<REGEXP>"[^" {
   advanceCol();
   charClass = CharClass(true); 
   BEGIN CHARCLASS;
}

<REGEXP>"["/[^^] {
   advanceCol();
   charClass = CharClass(); 
   BEGIN CHARCLASS;
}

<REGEXP>"[:any:]" {
   advanceCol();
   charClass.addChar('\x1');
   charClass.addChar('-');
   charClass.addChar(REGEXP_CHMAX-1);
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>"[:alnum:]" {
   advanceCol();
   charClass = CharClass(); 
   charClass.addChar('A');
   charClass.addChar('-');
   charClass.addChar('Z');
   charClass.addChar('a');
   charClass.addChar('-');
   charClass.addChar('z');
   charClass.addChar('0');
   charClass.addChar('-');
   charClass.addChar('9');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>"[:alpha:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('A');
   charClass.addChar('-');
   charClass.addChar('Z');
   charClass.addChar('a');
   charClass.addChar('-');
   charClass.addChar('z');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:cntrl:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('\x1');
   charClass.addChar('-');
   charClass.addChar('\x1f');
   charClass.addChar('\x7f');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:digit:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('0');
   charClass.addChar('-');
   charClass.addChar('9');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:graph:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('\x21');
   charClass.addChar('-');
   charClass.addChar('\x7e');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:lower:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('a');
   charClass.addChar('-');
   charClass.addChar('z');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:print:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar(' ');
   charClass.addChar('-');
   charClass.addChar('\x7e');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:punct:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('!');
   charClass.addChar('-');
   charClass.addChar('/');
   charClass.addChar(':');
   charClass.addChar('-');
   charClass.addChar('@');
   charClass.addChar('[');
   charClass.addChar('-');
   charClass.addChar('`');
   charClass.addChar('{');
   charClass.addChar('-');
   charClass.addChar('~');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:space:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('\x9');
   charClass.addChar('-');
   charClass.addChar('\xd');
   charClass.addChar(' ');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:upper:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('A');
   charClass.addChar('-');
   charClass.addChar('Z');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}
 
<REGEXP>"[:xdigit:]" {
   advanceCol();
   charClass = CharClass();
   charClass.addChar('0');
   charClass.addChar('-');
   charClass.addChar('9');
   charClass.addChar('A');
   charClass.addChar('-');
   charClass.addChar('F');
   charClass.addChar('a');
   charClass.addChar('-');
   charClass.addChar('f');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>"." {
   advanceCol();
   charClass = CharClass(true);
   charClass.addChar('\n');
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>"(" {
   if (!lex_dquoteRegexp && !lex_squoteRegexp) {
     unput('(');
     errMsg("Unquoted regular expressions must not contain parentheses");
     BEGIN 0;
   }
   else {
     advanceCol();   
     yylex_ret(TOK_RE_LPAREN);
   }
}

<REGEXP>")" {
   if (!lex_dquoteRegexp && !lex_squoteRegexp) {
     unput(')');
     BEGIN 0;
   }
   else {
     advanceCol();   
     yylex_ret(TOK_RE_RPAREN);
   }
}

<REGEXP>"*" {
   advanceCol();   
   yylex_ret(TOK_RE_STAR);
}

<REGEXP>"+" {
   advanceCol();   
   yylex_ret(TOK_RE_PLUS);
}

<REGEXP>"?" {
   advanceCol();   
   yylex_ret(TOK_RE_QMARK);
}

<REGEXP>"|"/[^|] {
   advanceCol();   
   yylex_ret(TOK_RE_OR);
}

  /***************
<REGEXP>"^" {
   advanceCol();   
   yylval.reVal = new RegExpr(RegExpr(RegExpr::BEG));
   yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>"$" {
   advanceCol();   
   yylval.reVal = new RegExpr(RegExpr(RegExpr::END));
   yylex_retval(TOK_RE, yylval.reVal);
}
   ***************/

<REGEXP>" " {
   advanceCol();
   if (!lex_dquoteRegexp && !lex_squoteRegexp)
      BEGIN 0;
   else {
      yylval.reVal = new RegExpr(' ');
      yylex_retval(TOK_RE, yylval.reVal);
   }
}

<REGEXP>\" {
   if (lex_dquoteRegexp) {
      advanceCol();
      BEGIN 0;
   }
   else if (!lex_squoteRegexp) {
      unput('"');
      errMsgLn("Unquoted regular expression cannot contain quotes");
      BEGIN 0;
   }
   else {
      advanceCol();
      yylval.reVal = new RegExpr('\"');
      yylex_retval(TOK_RE, yylval.reVal);
   }
}

<REGEXP>\' {
   if (lex_squoteRegexp) {
      advanceCol();
      BEGIN 0;
   }
   else if (!lex_dquoteRegexp) {
      unput('\'');
      errMsgLn("Unquoted regular expression cannot contain quotes");
      BEGIN 0;
   }
   else {
      advanceCol();
      yylval.reVal = new RegExpr('\'');
      yylex_retval(TOK_RE, yylval.reVal);
   }
}

<REGEXP>"\n" { 
   if (lex_dquoteRegexp || lex_squoteRegexp) {
      unput('\n');
      errMsgLn("Regular expression cannot span multiple lines");
      yylval.reVal = new RegExpr();
      yylex_retval(TOK_RE, yylval.reVal);
   }
   else {
      advanceCol();
      BEGIN 0;
   }      
}


<REGEXP>\\{OCTAL_DIGIT}{1,3} {
  unsigned char c = check_and_convert(8, yytext+1);
  advanceCol();
  yylval.reVal = new RegExpr(c);
  yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>\\x{HEX_DIGIT}{1,2} { 
  char c = check_and_convert(16, yytext+2);
  advanceCol();
  yylval.reVal = new RegExpr(c);
  yylex_retval(TOK_RE, yylval.reVal);
}

<REGEXP>\\. { 
  char c = *(yytext+1);
  switch (c) {
    case 'a': c = '\a'; break;
    case 'b': c = '\b'; break;
    case 'f': c = '\f'; break;
    case 'n': c = '\n'; break;
    case 'r': c = '\r'; break;
    case 't': c = '\t'; break;
    case 'v': c = '\v'; break;
    default: /* treat \x as x when none of the above cases apply */
      break; /* This applies to \\, \", \?, \' and so on.        */
  }
  advanceCol();
  yylval.reVal = new RegExpr(c);
  yylex_retval(TOK_RE, yylval.reVal);
} 

<REGEXP>. { 
   char c = *yytext;
   if (!lex_dquoteRegexp && !lex_squoteRegexp && 
       (c == '|' || c == '&' || c == ',' || c == ';')) {
      unput(c);
      errMsg("Invalid character within unquoted regular expression");
      BEGIN 0;
   }
   else {
      advanceCol();
      yylval.reVal = new RegExpr(c);
      yylex_retval(TOK_RE, yylval.reVal);
   } 
}

<CHARCLASS>\\{OCTAL_DIGIT}{1,3} {
  unsigned char c = check_and_convert(8, yytext+1);
  charClass.addChar(c);
  advanceCol();
}

<CHARCLASS>\\x{HEX_DIGIT}{1,2} { 
  char c = check_and_convert(16, yytext+2);
  charClass.addChar(c);
  advanceCol();
}

<CHARCLASS>\\. { 
  char c = *(yytext+1);
  switch (c) {
    case 'a': c = '\a'; break;
    case 'b': c = '\b'; break;
    case 'f': c = '\f'; break;
    case 'n': c = '\n'; break;
    case 'r': c = '\r'; break;
    case 't': c = '\t'; break;
    case 'v': c = '\v'; break;
    default: /* treat \x as x when none of the above cases apply */
      break; /* This applies to \\, \", \?, \' and so on.        */
  }
  charClass.addChar(c);
  advanceCol();
} 

<CHARCLASS>"]" {
   advanceCol();
   BEGIN REGEXP;
   yylval.reVal = new RegExpr(charClass);
   yylex_retval(TOK_RE, yylval.reVal);
}

<CHARCLASS>. { 
  char c = *yytext;
  charClass.addChar(c);
  advanceCol();
} 

 /**************** Recognize operators ****************/

"-->"   { advanceCol(); chkRuleScope(); yylex_ret(TOK_ARROW);}
"->"    { advanceCol(); chkRuleScope(); yylex_ret(TOK_ARROW); }
"?"     { advanceCol(); chkRuleScope(); yylex_ret(TOK_QMARK); }
"+"     { advanceCol(); yylex_ret(TOK_PLUS);}
"-"     { advanceCol(); yylex_ret(TOK_MINUS);}
"*"     { advanceCol(); yylex_ret(TOK_MULT);}
"/"     { advanceCol(); yylex_ret(TOK_DIV);}
"%"     { advanceCol(); yylex_ret(TOK_MOD);}
">"     { advanceCol(); yylex_ret(TOK_GT); }
"<"     { advanceCol(); yylex_ret(TOK_LT); }
">="    { advanceCol(); yylex_ret(TOK_GE); }
"<="    { advanceCol(); yylex_ret(TOK_LE);}
","     { advanceCol(); yylex_ret(TOK_COMMA);}
"."     { advanceCol(); yylex_ret(TOK_DOT);}
"=="    { advanceCol(); yylex_ret(TOK_EQ);}
"=~"    { advanceCol(); BEGIN REGEXP_BEG; yylex_ret(TOK_EQREG);}
 /* NEQREG should not be a general purpose operator -- what we 
    can efficiently support is 
"!~"    { advanceCol(); BEGIN REGEXP_BEG; yylex_ret(TOK_NEQREG);}
 */
"="     { advanceCol(); yylex_ret(TOK_ASSIGN);}
"!="    { advanceCol(); yylex_ret(TOK_NE);}
";"     { advanceCol(); yylex_ret(TOK_SEMICOLON);}
":"     { advanceCol(); yylex_ret(TOK_COLON);}  
"{"     { advanceCol(); lbrace++; yylex_ret(TOK_LBRACE);}
"}"     { advanceCol(); chkBraceMismatch(); yylex_ret(TOK_RBRACE);}
"("     { advanceCol(); lparen++; yylex_ret(TOK_LPAREN);}
")"     { advanceCol(); chkParenMismatch(); yylex_ret(TOK_RPAREN);}
"["     { advanceCol(); lbrack++; yylex_ret(TOK_LBRACK);}
"]"     { advanceCol(); chkBrackMismatch(); yylex_ret(TOK_RBRACK);}
"&&"    { advanceCol(); yylex_ret(TOK_AND);}
"&&&"   { advanceCol(); yylex_ret(TOK_SEQ_AND);}
"||"    { advanceCol(); yylex_ret(TOK_OR);}
"|||"   { advanceCol(); yylex_ret(TOK_SEQ_OR);}
"\\/"   { advanceCol(); yylex_ret(TOK_PAT_OR);}
"&"     { advanceCol(); yylex_ret(TOK_BITAND);}
"|"     { advanceCol(); yylex_ret(TOK_BITOR);}
"~"     { advanceCol(); yylex_ret(TOK_BITNOT);}
"^"     { advanceCol(); yylex_ret(TOK_BITXOR);}
"!"     { advanceCol(); yylex_ret(TOK_NOT);}
"<<"    { advanceCol(); yylex_ret(TOK_SHL);}
">>"    { advanceCol(); yylex_ret(TOK_SHR);}
"::"    { advanceCol(); yylex_ret(TOK_DOUBLE_COLON); }
"with"  { advanceCol(); yylex_ret(TOK_WITH);}

 /**************** Recognize reserved words ****************/

 /* Basic types */

 /*
"output"         { advanceCol(); yylex_ret(TOK_OUTPUT); }
"input"          { advanceCol(); yylex_ret(TOK_INPUT); }
 */

"enum"           { advanceCol(); yylex_ret(TOK_ENUM); }
"bool"           { advanceCol(); yylex_ret(TOK_BOOL);}
"string"         { advanceCol(); yylex_ret(TOK_STRING);} 
"regexp"         { advanceCol(); yylex_ret(TOK_REGEXP);} 
"short"          { advanceCol(); yylex_ret(TOK_SHORT);}
"long"           { advanceCol(); yylex_ret(TOK_LONG);}
"int"            { advanceCol(); yylex_ret(TOK_INT);}
"char"           { advanceCol(); yylex_ret(TOK_CHAR);}
"bit"            { advanceCol(); yylex_ret(TOK_BIT);}
 /*"float"          { advanceCol(); yylex_ret(TOK_FLOAT);}*/
"double"         { advanceCol(); yylex_ret(TOK_DOUBLE);}
"void"           { advanceCol(); yylex_ret(TOK_VOID);}
"unsigned"       { advanceCol(); yylex_ret(TOK_UNSIGNED); } 
"const"          { advanceCol(); yylex_ret(TOK_CONST);} 
"true"           { advanceCol(); yylex_ret(TOK_TRUE);} 
"false"          { advanceCol(); yylex_ret(TOK_FALSE);} 

 /* Class and struct types */

"class"          { advanceCol(); yylex_ret(TOK_CLASS);}
"struct"         { advanceCol(); yylex_ret(TOK_STRUCT);}

"or"             { advanceCol(); yylex_ret(TOK_KEY_OR);}

 /* Compound types (list and tuple) */

"list"           { advanceCol(); yylex_ret(TOK_LIST); } 
"set"            { advanceCol(); yylex_ret(TOK_SET); } 
"tuple"          { advanceCol(); yylex_ret(TOK_TUPLE); } 
"dict"           { advanceCol(); yylex_ret(TOK_DICT); } 

"foreach"        { advanceCol(); yylex_ret(TOK_FOREACH); }
"while"          { advanceCol(); yylex_ret(TOK_WHILE); }
"do"             { advanceCol(); yylex_ret(TOK_DO); } 

"empty"          { advanceCol(); yylex_ret(TOK_EMPTY); }
"clear"          { advanceCol(); yylex_ret(TOK_CLEAR); } 
"length"         { advanceCol(); yylex_ret(TOK_LENGTH); }
"append"         { advanceCol(); yylex_ret(TOK_APPEND); }
"union"          { advanceCol(); yylex_ret(TOK_UNION); }
"insert"         { advanceCol(); yylex_ret(TOK_INSERT); }
"erase"          { advanceCol(); yylex_ret(TOK_ERASE); } /* by position */
"contains"       { advanceCol(); yylex_ret(TOK_CONTAINS); } 
"remove"         { advanceCol(); yylex_ret(TOK_REMOVE); } /* by key or value */ 
"print"          { advanceCol(); yylex_ret(TOK_PRINT); } 
"printr"          { advanceCol(); yylex_ret(TOK_PRINT_RAW); } 

 /* Events */

"event"          { advanceCol(); yylex_ret(TOK_EVENT);}
 /*
"any"            { advanceCol(); yylex_ret(TOK_ANY); }
"other"          { advanceCol(); yylex_ret(TOK_OTHER); }
 */

 /* if-then-else */

"if"             { advanceCol(); yylex_ret(TOK_IF); }
"else"           { advanceCol(); yylex_ret(TOK_ELSE); }

 /* Modules and state machines */

"module"         { advanceCol(); chkGlobalScope("module"); 
                   yylex_ret(TOK_MODULE);}
"statemachine"   { advanceCol(); chkGlobalScope("statemachine"); 
                   yylex_ret(TOK_STATEMACHINE);}
"states"         { advanceCol(); yylex_ret(TOK_STATES); } 
"start"          { advanceCol(); yylex_ret(TOK_STARTSTATE);} 
"final"          { advanceCol(); yylex_ret(TOK_FINALSTATE);} 
 /*"deterministic"  { advanceCol(); yylex_ret(TOK_DETERMINISTIC); } */
"map"            { advanceCol(); yylex_ret(TOK_MAP);} 
"timeout"        { advanceCol(); yylex_ret(TOK_TIMEOUT);}
"in"             { advanceCol(); yylex_ret(TOK_IN);}
"instantiate"    { advanceCol(); yylex_ret(TOK_INSTANTIATE);}

 /* "On" statements */

"when"           { advanceCol(); yylex_ret(TOK_WHEN);}
"on"             { advanceCol(); yylex_ret(TOK_ON);} 
"all"            { advanceCol(); yylex_ret(TOK_ALL);}
"eachof"         { advanceCol(); yylex_ret(TOK_EACHOF);}
"specialize"     { advanceCol(); yylex_ret(TOK_SPECIALIZE);}
"histogram"      { advanceCol(); yylex_ret(TOK_HISTO);}
"frequency"      { advanceCol(); yylex_ret(TOK_FREQ);}
"falseAlarmRate" { advanceCol(); yylex_ret(TOK_FA_RATE);} 
"falsealarmrate" { advanceCol(); yylex_ret(TOK_FA_RATE);} 
"learn"          { advanceCol(); yylex_ret(TOK_LEARN);}
"detect"         { advanceCol(); yylex_ret(TOK_DETECT);}
"maxsize"        { advanceCol(); yylex_ret(TOK_MAXSIZE);}
 /*
"of"             { advanceCol(); yylex_ret(TOK_OF);}
"to"             { advanceCol(); yylex_ret(TOK_TO);}
"trans"          { advanceCol(); yylex_ret(TOK_TRANS);}
"factor"         { advanceCol(); yylex_ret(TOK_FACTOR);}
"distribution"   { advanceCol(); yylex_ret(TOK_DISTRIB);}
 */

 /* Other */

"sizeof"         { advanceCol(); yylex_ret(TOK_SIZEOF);}

"interface"      { advanceCol(); chkGlobalScope("interface"); 
                   yylex_ret(TOK_INTERFACE); }
 /*
"import"         { advanceCol(); yylex_ret(TOK_USE); }
"optimize"       { advanceCol(); yylex_ret(TOK_OPTIMIZE);}
"at"             { advanceCol(); yylex_ret(TOK_AT); }
 */

"return"         { advanceCol(); yylex_ret(TOK_RETURN); }

 /**************** Recognize Identifiers ****************/

{IDENT} { 
  SymTabEntry * ste;
  yylval.cVal = copyStr(); 
  advanceCol();
  ste = stm.lookUp(yylval.cVal);
  if (ste != NULL)  {
     if (ste->kind() == EVENT_KIND)
        yylex_retval(TOK_EV_IDENT, yylval.cVal);
     else if (ste->kind() == ENUM_ELEM_KIND)
        yylex_retval(TOK_ENUMVAL, yylval.cVal);
  }
  yylex_retval(TOK_IDENT, yylval.cVal);
}

 /**************** Recognize numeric values ****************/

{IP_DECIMAL} {
  unsigned b[4];
  yylval.uVal = 0;
  if ((sscanf(yytext, "%u.%u.%u.%u", &b[3], &b[2], &b[1], &b[0]) != 4) ||
      (b[3] > 255) || (b[2] > 255) || (b[1] >= 255) || (b[0] >= 255)) 
     errMsg(string("Invalid IP address: ") + yytext);
  else yylval.uVal = (b[3] << 24) | (b[2] << 16) | (b[1] << 8) | b[0];
  advanceCol();
  yylex_retval(TOK_UINTNUM, yylval.uVal);
} 

{BINARY} { 
  yylval.ulVal = strtoul(yytext, NULL, 2); 
  advanceCol();
  yylex_retval(TOK_ULONGNUM, yylval.ulVal);
}

{OCTNUM} { 
  yylval.ulVal = strtoul(yytext, NULL, 8); 
  advanceCol();
  yylex_retval(TOK_ULONGNUM, yylval.ulVal);
}

{UINTNUM} { 
  yylval.ulVal = strtoul(yytext, NULL, 10); 
  advanceCol();
  yylex_retval(TOK_ULONGNUM, yylval.ulVal);
}

{HEXNUM} { 
  yylval.ulVal = strtoul(yytext, NULL, 16); 
  advanceCol();
  yylex_retval(TOK_ULONGNUM, yylval.ulVal);
}

{FLOATNUM} { 
  yylval.dVal = strtod(yytext, NULL);
  advanceCol();
  yylex_retval(TOK_DOUBLENUM, yylval.dVal);
}

 /**************** Recognize whitespace characters ****************/

{NEWLINE}   { advanceLn();}

[ ]       { advanceCol();}
[\t]      { tabAdvanceCol();}
[\r]      { advanceCol();}

 /**************** Unrecognized character ==> error ****************/

. { 
  errMsg("Illegal character"); 
  advanceCol();
}

%%
