/*
   This is a copy from nBMSL. That version seems to have gone through a 
   substantial revision to clean up error handling. So, it is likely the
   better version. While there are numerous differences between that version
   and the version that was originally here, almost all changes seem
   relatively minor. The language certainly is different, as some new features
   have been added (rewbr), some operators have changed (e.g., TOK_COLON for
   sequencing), I don't think they are really any major changes.
*/

// @@@@ Currently we get 8 shift-reduce conflicts on error productions.
// @@@@ In addition, two conflicts due to dangling-else are reported.
// @@@@ Two more conflicts on optSemicolon are reported, but these
// @@@@ seem harmless. (Total of 12 conflicts.)
//
// @@@@ For state machines, report error if it uses hasSeqOps() patterns. 
// @@@@ Question: in a state machine, is it OK to instantiate other modules 
// @@@@ that use complex pattern? Shd we say that module-level state vars
// @@@@ shouldn't be changed by complex patterns?
//
// @@@@ About %prec declarations: they seem to work only for prefix
// @@@@ operators. For other operators, it seems that perhaps bison
// @@@@ does not have enough info while parsing to assign the right
// @@@@ precedence to over-loaded operators. To deal with this problem,
// @@@@ had to change the language to use ":" in place of "." so that
// @@@@ precedences will work right.
//
// Error recovery strategies.
// 1. Recover at the level of a delimiter, such as ";", "}", ")" and so on.
//    Every construct that includes these delimiters is expected to 
//	  incorporate error recovery. In addition, error recovery is done
//    where necessary.
// 2. Error recovery is NOT attempted in contexts where delimiters
//    can be nested, e.g., expressions. The general idea is that
//    recovery at the granularity of a line is OK. However, finer grained
//    error recovery is implemented in cases where a clear synchronization
//    symbol can be found, e.g., TOK_ARROW
// 3. Use as much left and right context in error recovery rules. Otherwise
//    you get conflicts in the grammar, or repeated errors. For instance,
//	  if you have a production X -> error TOK_SEMICOLON, which has no
//    left context, then it is likely that this rule will be applied
//    in multiple contexts, such as declarations and statements, causing
//    conflicts. Similarly, if you had a rule X -> error, then you dont
//    skip the content immediately following the error, which usually
//    leads to additional (and usually spurious) error messages. 
// 4. IMPORTANT: The error rules dont quite work as described in the
//    Bison documentation. In particular, it is hard to understand error
//    recovery in terms of the grammar. Instead, one has to look at the
//    state and transitions generated by Bison. It seems that the only
//    way to be sure about error handling is by testing with different
//    errors. It also seems that error handling rules with non-empty left
//    contexts work in a more predictable way than those with empty
//    left context.
// 5. Because of the above point, it is possible to have too many error
//    handling rules that really destroy every thing. Since runtime 
//    transitions that happen when error occurs isn't backed up the
//    static analysis that lies behind parse table construction, the
//    parser is unable to identify equivalences between states. In
//    effect, this causes error recovery rules to interfere with each other
//    in unpredictable ways. Again, the best way to control this is to
//    have non-empty left and right context for error rules.
// 6. It seems best to "anchor" error handling around key delimiters
//    such as semicolons that are used in a very fixed way, and the grammar
//    consists of elements that are separated (or delimited) by these symbols.
//    Using other types of symbols for error handling will cause problems.
//    For instance, if you have a error-recovery rule that tries to
//    take care of unbalanced parenthesis, it may skip semicolons while
//    skipping error input, and end up incorrectly reading text from subsequent
//    statements, and overall doing a poor job.
// 7. For reasons described above, it is better to give up error recovery
//    in cases where (a) error is extreme, such as multiple syntax errors
//    right next to each other, (b) the span over which error recovery
//    isn't done is relatively small, e.g., within an expression. This seems
//    to be the best way to preserve reasonable error recovery for simple,
//    common errors.
// 8. Keep in mind that if you have a rule X -> Y error z, this rule
//    will be used ONLY if the stack already has Y on it near the top
//    at the time of error. If it contains W such that W -> Y is a rule,
//    then the error handling rule wont be applicable.
// 9. Sometimes the right context we use is part of the next part of
//    input. For instance, in a top level error recovery rule for interface
//    declaration, we use TOK_MODULE or TOK_INTERFACE as right contexts.
//    In this case, we use a lexical tie-in to ensure that the token is
//    resent by the lexer to the parser, so that it is included as
//    part of the following interface or module.
//10. We use a number of variables to remember line and column info
//    regarding error context to provide more accurate error reports.
//    (This part still needs to be ironed out -- often, the column/row
//    numbers don't seem right.)
//11. Mismatched parenthesis, bracket or brace can confuse the parser badly.
//    Our only hope is to point user to the likely places where there is
//    a problem, and hope he/she can fix it. In order to provide useful
//    error messages, we (a) track some of the keywords and symbols 
//    (e.g., module) within the lexer, and report if they are found in the 
//    wrong scope, and (b) keep track of the number of left/right paren/brace/
//    bracket to provide useful error messages.
//12. In some cases, the synchronizing symbols occur past the end of the
//    construct being parsed. To handle this, we have a 1-character buffer
//    into which the look-ahead synchronization character is pushed back.
//    yylex() is defined so that it first checks this buffer, and returns
//    the value contained in there, before calling lexer's function.

// @@@@ stmt does not implement error-handling. This means that every user
// @@@@ of stmt needs to handle errors. This is currently not done for rule
// @@@@ and loop statements.

%{

  // @@@@ Note: rule vars are tagged appropriately. This was not done for
  // @@@@ absevents, but now we got rid of them.

#include "Ast.h"
#include "ParserUtil.h"
#include "BMSL.tab.H"
#include <string.h>
#include <assert.h>
#include "RegExpr.h"

extern bool lex_dquoteRegexp, lex_squoteRegexp;
int yysymflags=0;
int extSym = 1;
unsigned int ruleLevel = 0; // are we within a pattern of rule or outside?
unsigned int strHdrEnt = 0; // indicates if we are within a struct defn
bool isMapping = false;     // is this a map statement?
bool doneWithPat = true;    // Used to track whether we are parsing the pattern
                            // component of a rule, or moved into action.
unsigned offSet = 0;
unsigned fnOffSet = 1;           // used to assign a serial number to each fn
bool inModuleHeader = false;
bool inEventHeader = false;

int lastline;
bool seenLparen, seenRbrace;
extern int yylineno, yycolumnno;
int corcol, corline;
int bufferedToken = 0, bufTok;
unsigned vct;
extern int errTokBegLn, errTokEndLn, errTokBegCol, errTokEndCol;

void yyerror(const char *s) {
  if (errTokBegLn == 0) {
	errTokBegLn = tokBegLn;
    errTokBegCol = tokBegCol;
  }
  errTokEndCol = tokEndCol;
  errTokEndLn = tokEndLn;
}

bool canBuffer() {
   return bufferedToken == 0;
};

void buffer(int x) {
   if (!canBuffer())
	  errMsg("Buffer full");
   else {
  	  bufferedToken = 1; bufTok = x;
   }
};

bool isBufNonEmpty() {
   return bufferedToken == 1;
};

int unbuffer() {
   if (bufferedToken == 0) {
	  errMsg("Buffer empty"); 
      return 0;
   }
   else {
      bufferedToken = 0; 
      return bufTok;
   }
};

void resetVars() {
   doneWithPat = true; ruleLevel = 0; isMapping=false;
   seenLparen=seenRbrace=false; //unbuffer();
   inModuleHeader=false;
   inEventHeader=false;
}

/* The following code uses a look ahead symbol to exit out of event scopes. 
   It should be called only at the end of event pattern productions */


#define chkExitEvScope()

/*
#define chkExitEvScope() \
  if (yychar == YYEMPTY) { \
    int tmp; \
	if (canBuffer()) \
      tmp = yylex(); \
    else tmp = unbuffer(); \
    buffer(tmp); \
    if ((tmp==TOK_PAT_OR)||(tmp==TOK_COLON)||(tmp==TOK_QMARK)\
        ||(tmp==TOK_PAT_PLUS)||(tmp==TOK_PAT_STAR)) \
		{leaveToScopeErrCheck(RULE_BLOCK_KIND);} }\
    else if ((yychar==TOK_PAT_OR)||(yychar==TOK_COLON)||(yychar==TOK_QMARK)\
             ||(yychar==TOK_PAT_PLUS)||(yychar==TOK_PAT_STAR))\
    {leaveToScopeErrCheck(RULE_BLOCK_KIND);};
*/

// The purpose of yylex1 below is to distinguish between TOK_(PLUS|STAR) and
// TOK_*PAT*_(PLUS|STAR). It does this by using a look-ahead token that is kept
// in lbTok. Note that lbTok is a distinct buffer from bufTok that is managed by
// functions such as canBuffer() above, and is used to look ahead for 
//    (a) a synchronizing token during error recovery, or 
//    (b) when to exit event scope
// Since yylex1() is called exclusively from inside yylex(), and the latter
// ensures that bufTok is empty at that time, it is clear that bufTok has higher
// priority over lbTok. 
//
// A simpler way to think about this is that lbTok is a look-ahead used entirely
// within the lexer, whereas bufTok is a look-ahead that is maintained and used
// by the parser.

inline int yylex1() {
  static int lbuffered = 0, lbTok;
  static YYSTYPE lbv;
  int curtok;

  if (lbuffered) {
    lbuffered = 0;
	curtok = lbTok;
    yylval = lbv;
  }
  else curtok = yylex();
  // Invariant: lbuffered == 0

  if ((curtok == TOK_MULT) || (curtok == TOK_PLUS)) {
	  lbuffered = 1;
	  lbTok = yylex();
      lbv = yylval;
	  if ((lbTok==TOK_ARROW)||(lbTok==TOK_RPAREN)||(lbTok==TOK_SEMICOLON)
		  || (lbTok == TOK_QMARK) || (lbTok == TOK_MULT) || (lbTok == TOK_PLUS)
          || (lbTok == TOK_PAT_OR) ||  (lbTok == TOK_COLON)) {
		switch (curtok) {
		 case TOK_MULT: return TOK_PAT_STAR;
		 case TOK_PLUS: return TOK_PAT_PLUS;
	    }
      }
  }
  return curtok;
}

#define yylex() (isBufNonEmpty() ? (errTokBegLn = errTokEndLn = errTokBegCol = errTokEndCol = 0, unbuffer()) : (YYRECOVERING() ? yylex1() : (errTokBegLn = errTokEndLn = errTokBegCol = errTokEndCol = 0, yylex1())))

/* The following code uses a look ahead symbol to exit out of event scopes. 
   It should be called only at the end of event pattern productions */

#define leaveToScopeErrCheck(k) {\
  if ((stm.currentScope(k) != nullptr) && (stm.leaveToScope(k) != OK)) {\
	internalWarn(string("current Scope is ") + stm.currentScope()->name()+\
				  " " +  kindTag(stm.currentScope()->kind()) +\
				  ", but trying to go to " + kindTag(k));\
  }\
}

#define leaveScopeErrCheck(k) {\
  if ((stm.currentScope(k) != nullptr) && (stm.leaveScope(k) != OK)) {\
	internalWarn(string("current Scope is ") + stm.currentScope()->name()+\
				  " " +  kindTag(stm.currentScope()->kind()) +\
				  ", but trying to leave " + kindTag(k));\
  }\
}

extern const OpNode::OpInfo opInfo[];
%}

%union{
  unsigned int 	     uVal;
  unsigned long	     ulVal;
  unsigned long      lVal;
  double             dVal;
  char*              cVal;
  RegExpr*           reVal;
  vector<string>*    strVector;
  Value*             valVal;

  ExprNode*          exprVal;
  vector<ExprNode*>* exprList;
  RefExprNode*       refexpVal;

  EnumEntry*         enumEntry;
  vector<EnumElemEntry*>* enumElemList;
  ClassEntry*        classEntry;
  StructEntry*       structEntry;
  Type*              typeVal;
  vector<Type*>*     typeList; 

  LabelEntry*        labelEntry;                        
  EventEntry*        eventEntry;                        
  VariableEntry*     variableEntry;
  FunctionEntry*     functionEntry;
  ModuleEntry*       moduleEntry;
  InterfaceEntry*    interfaceEntry;

  BasePatNode*       patVal;
  PrimitivePatNode*  primPatVal;
  StmtNode*          stmtVal;
  ListForNode*       listForVal;
  WhileNode*         whileVal;
  list<StmtNode*>*   stmtList;
  CompoundStmtNode*  compoundStmtVal;
  RuleNode*          ruleNode;

  vector<RuleNode*>* transList;
  OnEntry*           onVal;
  
  vector<Type*>*     formalTypeVal;
  const OpNode::OpInfo*    funIval;
}

/* Operators */

%token TOK_RE TOK_RE_LPAREN TOK_RE_RPAREN TOK_RE_STAR TOK_RE_PLUS TOK_RE_QMARK
%token TOK_RE_OR TOK_RE_NEG
%token TOK_PAT_STAR TOK_PAT_PLUS TOK_PAT_OR TOK_PAT_NOT
%token TOK_PLUS TOK_MINUS TOK_UMINUS TOK_MULT TOK_DIV TOK_MOD 
%token TOK_BITAND TOK_BITOR 
%token TOK_BITNOT TOK_BITXOR TOK_SHL TOK_SHR TOK_GT TOK_LT TOK_GE 
%token TOK_LE TOK_EQ  TOK_NE TOK_EQREG TOK_IN /*TOK_NEQREG*/
%token TOK_AND TOK_OR TOK_SEQ_AND TOK_SEQ_OR TOK_NOT TOK_COMMA TOK_DOT TOK_ASSIGN
%token TOK_DOUBLE_COLON TOK_SEMICOLON
%token TOK_COLON TOK_ARROW TOK_QMARK
%token TOK_LBRACE TOK_RBRACE TOK_LPAREN TOK_RPAREN TOK_LBRACK TOK_RBRACK 

/* Key words */

%token TOK_SIZEOF TOK_MAXSIZE TOK_IF LOWER_THAN_ELSE TOK_ELSE 
%token TOK_DO TOK_FOREACH TOK_WHILE
%token TOK_INTERFACE TOK_STRUCT TOK_KEY_OR TOK_WITH TOK_CLASS TOK_RETURN
%token TOK_MODULE TOK_STATEMACHINE TOK_USE TOK_OPTIMIZE TOK_AT TOK_INSTANTIATE
%token TOK_EVENT TOK_ANY TOK_OTHER
%token TOK_IDENT TOK_EV_IDENT TOK_ENUMVAL 
%token TOK_CONST TOK_ULONGNUM TOK_UINTNUM TOK_DOUBLENUM 
%token TOK_VOID TOK_DETERMINISTIC TOK_PRINT TOK_PRINT_RAW
%token TOK_APPEND TOK_UNION TOK_LENGTH TOK_EMPTY
%token TOK_REMOVE TOK_CONTAINS TOK_INSERT TOK_ERASE TOK_CLEAR
%token TOK_STRING TOK_STRCONST TOK_REGEXP TOK_INT TOK_UNSIGNED
%token TOK_BOOL TOK_CHAR TOK_BIT TOK_FLOAT TOK_DOUBLE TOK_SHORT 
%token TOK_LONG TOK_LIST TOK_SET TOK_TUPLE TOK_DICT TOK_ENUM TOK_OUTPUT TOK_INPUT
%token TOK_STATES TOK_STARTSTATE TOK_FINALSTATE TOK_MAP 
%token TOK_WHEN TOK_ALL TOK_EACHOF TOK_TIMEOUT
%token TOK_OF TOK_FA_RATE TOK_ON TOK_SPECIALIZE TOK_TRANS
%token TOK_FREQ TOK_DISTRIB TOK_HISTO TOK_TO
%token TOK_TRUE TOK_FALSE TOK_FACTOR TOK_LEARN TOK_DETECT

%type <functionEntry> function_header function_decl function_decl1 function_defn
%type <variableEntry> rewbr_var variable_decl formal_param variable_decl1 array_decl
%type <typeVal> module_formal_param
%type <typeList> formal_param_list formal_param_plus module_formal_param_list module_formal_param_plus 
%type <typeList> type_tuple 
%type <structEntry> struct_header struct_header1
%type <eventEntry> event_decl_hdr event_decl
%type <patVal> event_pattern rule_body1
%type <primPatVal> event event_vbar ev_with_args primitive_pat 

%type <enumEntry> enum_header states_header

%type <enumElemList> state_list

%type <moduleEntry> module_header module_decl module_decl1
%type <typeVal> type type1 baseType 
%type <strVector> ident_list ident_plus states_plus
%type <exprList> expr_list expr_plus specialize wrt1 expr_tuple dictval_plus
%type <exprVal> expr opt_expr opt_init opt_with timeval invocation asg dictval 
%type <exprVal> opt_learn_cond opt_detect_cond
%type <refexpVal> ref_expr list_for_hdr
%type <ruleNode> rule_body

%type <reVal> TOK_RE regexp
%type <cVal> TOK_IDENT TOK_EV_IDENT TOK_ENUMVAL ev_arg opt_ident evnm_lparen  ev_or_ident
%type <funIval> ev_or_fun
%type <uVal> tok_module_or_state_machine lbrace_or_semicol TOK_UINTNUM 
%type <ulVal> TOK_ULONGNUM
%type <dVal> TOK_DOUBLENUM 
%type <cVal> TOK_STRCONST

%type <valVal> literal
%type <cVal>   strconst

%type <onVal>  on_trans_specialize_freq_or_val on_trans_specialize_freq 
       on_trans_specialize_val on_trans_specialize on_trans_spec on_trans_list

%type <stmtList> stmt_plus
%type <stmtVal> stmt simple_stmt opt_else list_for_stmt while_stmt 
                if_then_else_stmt   
%type <listForVal> list_for_hdr1
%type <whileVal> while_hdr
%type <compoundStmtVal> fun_body compoundStmt

%type <transList> labels trans_list

/*********************   Precedence and Associativity    ********************/

%nonassoc TOK_ASSIGN 

%left  TOK_PAT_OR 
%left  TOK_COLON
%left  TOK_QMARK
%left  TOK_PAT_PLUS
%left  TOK_PAT_STAR
%right TOK_PAT_NOT

%left  TOK_OR TOK_SEQ_OR
%left  TOK_AND TOK_SEQ_AND
%right TOK_NOT

%nonassoc TOK_EQ TOK_NE TOK_GT TOK_LT TOK_GE TOK_LE TOK_EQREG TOK_IN

%left  TOK_RE_OR
%left  TOK_RE_SEQ
%left  TOK_RE_QMARK
%left  TOK_RE_PLUS 
%left  TOK_RE_STAR 
%right  TOK_RE_NEG
%nonassoc TOK_RE
%left TOK_RE_LPAREN

%left  TOK_BITOR TOK_BITXOR
%left  TOK_BITAND
%right TOK_BITNOT 
%left  TOK_SHL TOK_SHR

%right TOK_DOUBLE_COLON

%left  TOK_PLUS TOK_MINUS
%left  TOK_MULT TOK_DIV TOK_MOD
%right TOK_UMINUS 

%left  TOK_LBRACK
%left  TOK_DOT

/* 
   It is not a good idea to use precedences simply to eliminate
   conflict messages --- the precedence may eliminate additional
   conflicts beyond what was intended. So it is best to leave the
   conflict reports.

%nonassoc LOWER_THAN_ELSE
%nonassoc TOK_ELSE

%nonassoc LOWER_THAN_SEMICOLON
%nonassoc TOK_SEMICOLON
*/

/**************************************************************************/

%start specification
%%

specification: global_decl_star;

global_decl_star: /* empty */ 
| global_decl_star global_decl
| global_decl_star error tok_module_or_state_machine {
	resetVars();
	if ($3 == 0) 
	  buffer(TOK_MODULE); 
	else buffer(TOK_STATEMACHINE); 
	errMsg("Parse error, skipping to the next module");
	stm.leaveToScope(GLOBAL_KIND);
}
;

global_decl: TOK_SEMICOLON
| event_decl {
  resetVars();
  leaveToScopeErrCheck(GLOBAL_KIND);
}
| class_decl TOK_SEMICOLON 
| function_decl TOK_SEMICOLON {
  leaveScopeErrCheck(FUNCTION_KIND);
}
| function_defn
| enum_decl
| variable_decl {
	errMsg("Variables definitions not permitted in interfaces");
}
| struct_decl
| module_defn
| TOK_INTERFACE TOK_IDENT TOK_LBRACE {
   insertSymbolErrCheck(new InterfaceEntry($2));
   yysymflags |= extSym;
}   
| TOK_RBRACE opt_semicolon {
   yysymflags &= ~extSym;
}
;

/***************************** ENUM ***********************************/

enum_decl: enum_header enum_body {
}
| TOK_ENUM error TOK_SEMICOLON {
  errMsg("parse error: expecting enum name and element list");
}
| TOK_ENUM TOK_IDENT error TOK_SEMICOLON {
  errMsg("Expecting \"{\" following enum name");
}
| enum_header error TOK_SEMICOLON {
  errMsg("Syntax error while parsing enum declaration");
};

enum_header: TOK_ENUM TOK_IDENT TOK_LBRACE {
  string enumName($2);
  EnumEntry* ee = new EnumEntry(enumName);
  Type* et = new Type(ee, Type::ENUM);
  ee->type(et);
  if (insertSymbolErrCheck(ee) == OK) 
    enterScopeErrCheck(ee);
  /* 
     There is no need to enter scope here, but doing so is a
     convenient way to pass ee to the rest of the production
     (enum_body)
  */
  $$ = ee;
}
| TOK_ENUM TOK_EV_IDENT TOK_LBRACE {
   errMsg("Event cannot be redefined to be an enum name");
   $$ = nullptr;
}
;

enum_body: ident_list rbrace_or_semicol {
  if (!seenRbrace)
	errMsg("Missing \"}\" in enum definition");
  if (stm.currentScope()->kind() == ENUM_KIND) {
    EnumEntry *ee = (EnumEntry*)stm.currentScope();
    leaveScopeErrCheck(ENUM_KIND); 
    /* recall that scope was entered simply to pass ee. Besides, you
       want to insert enum elements in the surrounding scope, or else 
       we cannot reference them without being inside the event defn scope! */
    unsigned int i=0;
    if ($1 != nullptr) {
      unsigned int n = $1->size();
	  if (n == 0)
		errMsgLn("Enum type must contain at least one element");
      for (i=0; i < n; i++) {
        EnumElemEntry* se = new EnumElemEntry((*$1)[i], i, ee->type());
        ee->addElem(se);
        insertSymbolErrCheck(se);
      }
      delete $1;
    }
  }
}
| ident_plus error rbrace_or_semicol {
  if (stm.currentScope()->kind() == ENUM_KIND)
    leaveScopeErrCheck(ENUM_KIND); 
  if ($1 != nullptr)
      errMsgLn(string("Error while parsing element ") + itoa($1->size()+1)
				+ " in enum definition");
  else errMsg("Syntax error while parsing enum definition");
  yyerrok;
};

rbrace_or_semicol: TOK_RBRACE {
	seenRbrace=true;
}
| TOK_SEMICOLON{
	seenRbrace=false;
}
;

ident_list: /* empty */ {
  $$ = new vector<string>;
}
| ident_plus {
  $$ = $1; 
}; 

ident_plus: TOK_IDENT {
  $$ = new vector<string>();
  $$->push_back(string($1));
}
| ident_plus TOK_COMMA TOK_IDENT {
  $1->push_back(string($3));
  $$ = $1;
};

/***************************** CLASS ***********************************/

class_decl: class_decl1 opt_class_body
| TOK_CLASS error TOK_SEMICOLON {
	errMsg("Syntax error in class declaration: expecting class name");
}
| class_decl1 error TOK_SEMICOLON {
	errMsg("Syntax error in class declaration");
};

class_decl1: TOK_CLASS TOK_IDENT {
  string className($2);
  ClassEntry *ce = new ClassEntry(className);
  Type *te = new Type(ce, Type::CLASS);
  ce->type(te);
  insertSymbolErrCheck(ce);
};

opt_class_body: /* empty */ 
| TOK_LBRACE error TOK_RBRACE {
  errMsgLn("Class declarations cannot have a body");
};

/***************************** FUNCTION ***********************************/

function_defn: function_decl fun_body {
  if ($1 != nullptr)
     $1->body($2);
  leaveScopeErrCheck(FUNCTION_KIND);
}
| function_header error TOK_LBRACE {
	errMsg("Syntax error in function declaration");
	buffer(TOK_LBRACE);
} fun_body {
	leaveScopeErrCheck(FUNCTION_KIND);
}
| function_header error TOK_SEMICOLON {
  if ($1 != nullptr)
    errMsgLn(string("Error while parsing argument of \"") + $1->name() + 
			"\", possibly due to missing or incorrect type declaration");
  else errMsg("Error encountered while parsing formal parameters");
  leaveScopeErrCheck(FUNCTION_KIND);
};

function_decl: function_decl1 {
   $$ = $1;
}
| function_decl1 TOK_CONST {
    if ($1 != nullptr)
       $1->type()->constType(true);
    $$ = $1;
};
 
function_decl1: function_header formal_param_list TOK_RPAREN { 
  if ($1 != nullptr) {
    $1->type()->argTypes($2);
  }
  $$ = $1;
}
| function_header formal_param_list error TOK_RPAREN {
  $$=nullptr;
  if ($2 != nullptr)
    errMsg(string("Syntax error in parameter declaration ")+itoa($2->size()));
  else errMsg("Syntax error in parameter declaration");
};

function_header: type TOK_IDENT TOK_LPAREN {
  string fnName($2);
  FunctionEntry *fe = new FunctionEntry(fnName, $1, fnOffSet++);
  $$ = nullptr;
  if ($1 != nullptr) {
    Type *te = new Type((vector<Type*>*)nullptr, $1);
    if (insertSymbolErrCheck(fe) == OK) {
      enterScopeErrCheck(fe);
      fe->type(te);
      $$=fe;
    }
  }
} 
| type TOK_EV_IDENT TOK_LPAREN {
     $$=nullptr;
     errMsg("Redeclaration of an event as a function?");
}
| type error TOK_LPAREN {
	$$=nullptr;
	errMsg("Expecting a function declaration -- may be a missing type or a function name");
}
| type1 error TOK_LPAREN {
	$$=nullptr;
	errMsg("Expecting a function declaration -- may be a missing type or a function name");
}
| TOK_IDENT error TOK_LPAREN {
	$$=nullptr;
	errMsg("Expecting a function declaration -- may be a missing type or a function name");
};

formal_param_list: /*empty*/ {
  $$=new vector<Type*>();
}
| formal_param_plus {
  $$ = $1;
};

formal_param_plus: formal_param {
  if ($1 != nullptr) {
    offSet = 0;
    $1->offset(new ValueNode(new Value(offSet))); // First param
    $$ = new vector<Type*>();
    $$->push_back($1->type());
  }
  else $$ = nullptr;
}
| formal_param_plus TOK_COMMA formal_param {
  if (($1 != nullptr) && ($3 != nullptr)) {
    offSet++;
    $3->offset(new ValueNode(new Value(offSet)));
    $1->push_back($3->type());
    $$ = $1;
  }
  else $$ = nullptr;
};

opt_ident: /* empty */ {
  $$ = nullptr;
}
| ev_or_ident  { 
  $$ = $1;
};

formal_param: type opt_ident {
  string v;
  if ($2 == nullptr)
    v = newName("p");
  else v = $2;
  VariableEntry * ve = 
     new VariableEntry(v, inModuleHeader? MODULE_PARAM_VAR : 
                          inEventHeader? EVENT_PARAM_VAR : FN_PARAM_VAR, $1);
  insertSymbolErrCheck(ve);
  $$ = ve;
}
;

fun_body: TOK_LBRACE variable_decls_star stmt_plus TOK_RBRACE {
  $$ = new CompoundStmtNode(nullptr, $3);
}
| TOK_LBRACE error TOK_RBRACE {
  errMsg("Syntax error in function body");
};

/************************ EVENT DECL ***********************/

event_decl: event_decl_hdr formal_param_list TOK_RPAREN { //%%%% CHECK seenLparen
  inEventHeader = false;
  if ($1 != nullptr)
    $1->type()->argTypes($2);
  $$ = $1;
}
| event_decl_hdr TOK_SEMICOLON {
  inEventHeader = false;
  if (seenLparen)
	errMsgLn("Unmatched closing parenthesis");
}
| event_decl_hdr formal_param_plus error TOK_RPAREN {
  inEventHeader = false;
  $$ = nullptr;
  if (($1 != nullptr) && ($2 != nullptr))
    errMsgLn(string("Error while parsing argument ") + itoa($2->size()+1) + 
			" of \"" + $1->name() + "\", possibly due to missing or incorrect "
            "type declaration");
  else errMsg("Error encountered while parsing event parameters");
}
| event_decl_hdr formal_param_plus error TOK_SEMICOLON {
  inEventHeader = false;
  $$ = nullptr;
  if (($1 != nullptr) && ($2 != nullptr))
    errMsgLn(string("Error while parsing argument ") + itoa($2->size()+1) + 
			" of \"" + $1->name() + "\", possibly due to missing \")\"");
  else errMsg("Error in parsing event parameters, may be a missing \")\"");
}
| event_decl_hdr error TOK_RPAREN {
  inEventHeader = false;
  $$ = nullptr;
  if ($1 != nullptr)
    errMsgLn(string("Error while parsing first argument of \"") + $1->name() + 
			"\", possibly due to missing or incorrect type declaration");
  else errMsg("Error encountered while parsing event parameters");
}
| event_decl_hdr error TOK_SEMICOLON {
  inEventHeader = false;
  $$ = nullptr;
  if ($1 != nullptr)
    errMsgLn(string("Error while parsing first argument of \"") + $1->name() + 
			"\", possibly due to missing \")\"");
  else errMsg("Error in parsing event parameters, may be a missing \")\"");
};

opt_lparen: /* empty */ {
  seenLparen=false;
}
| TOK_LPAREN {
  seenLparen=true;
}
;
event_decl_hdr: TOK_EVENT TOK_IDENT opt_lparen  {
  string eventName($2);
  EventEntry *ee = new EventEntry(eventName);
  if (insertSymbolErrCheck(ee) == OK) {
    enterScopeErrCheck(ee);
    Type *te = new Type((vector<Type*>*)nullptr, Type::EVENT);
    ee->type(te);
  }
  inEventHeader = true;
  $$ = ee;
}
| TOK_EVENT TOK_EV_IDENT opt_lparen  {
	errMsg(string("Redeclaration of event ") + $2);
	$$ = nullptr;
}
| TOK_EVENT error TOK_LPAREN {
  // %%%% Note: the case error TOK_LPAREN will be reduced to a function
  // %%%% declaration, which is OK.
  seenLparen=true;
  errMsg("Expected an event name");
  $$=nullptr;
}
| TOK_EVENT error TOK_SEMICOLON {
  seenLparen=false;
  buffer(TOK_SEMICOLON);
  errMsg("Expected an event name");
  $$=nullptr;
};

/******************************** EVENT PATTERN ******************************/
// Has no error recovery capability, since there are no "synchronizing"
// symbols or keywords. So, any production that uses event patterns
// must implement its own error recovery. 

primitive_pat: event {
  $$ = $1;
  chkExitEvScope();
}
| event_vbar expr {
  if (($1 != nullptr) && ($2 != nullptr)) 
    $1->cond($2); 
  $$ = $1;
  chkExitEvScope();
}
;

rewbr_var: TOK_LPAREN TOK_IDENT TOK_ASSIGN {
  SymTabEntry* ste = stm.lookUp($2);
  if ((!isMapping) && (ruleLevel++ == 0)) {
    BlockEntry* re = new RuleBlockEntry();
    if (insertSymbolErrCheck(re) == OK) {
      enterScopeErrCheck(re);
      doneWithPat = false;
    }
  }
  if (ste == nullptr) {
	RuleBlockEntry* rbe = 
	  (RuleBlockEntry*)stm.currentScope(RULE_BLOCK_KIND);
	if (rbe != nullptr) {
    	VariableEntry *ve = new VariableEntry($2, PATTERN_VAR);
	  	rbe->insert(ve);
		$$=ve;
	}
    else $$=nullptr;
  }
  else {
	if (ste->kind() == VARIABLE_KIND) {
	  $$ = (VariableEntry*)ste;
    }
	else {
	  errMsg(string("Expecting a variable name, but found \"")
			+ste->name()+"\"");
      $$ = nullptr;
    }
  }
};

event_pattern: primitive_pat {
  $$ = $1;
  chkExitEvScope();
}
| TOK_LPAREN event_pattern TOK_RPAREN {
  chkExitEvScope();
  $$ = $2;
}
| TOK_NOT event_pattern %prec TOK_PAT_NOT {
  if ($2 != nullptr)
	$$ = new PatNode(BasePatNode::NEG, $2);
  else $$ = nullptr;
  chkExitEvScope();
}
| event_pattern TOK_PAT_OR event_pattern {
  if ($1 != nullptr && $3 != nullptr) 
	$$ = new PatNode(BasePatNode::OR, $1, $3);
  else $$ = nullptr;
  // leaveScopeErrCheck(EVENT_BLOCK_KIND);
}
/*
| event_pattern TOK_COLON event_pattern {
  if ($1 != nullptr && $3 != nullptr)
    $$ = new PatNode(BasePatNode::SEQ, $1, $3);
  else $$ = nullptr;
  chkExitEvScope();
}
| event_pattern TOK_PAT_STAR {
  if ($1 != nullptr)
    $$ = new PatNode(PatNode::STAR, $1);
  else $$ = nullptr;
  leaveScopeErrCheck(EVENT_BLOCK_KIND);
}
| event_pattern TOK_QMARK {
  if ($1 != nullptr) {
    PatNode* pn = new PatNode();
	$$ = new PatNode(BasePatNode::OR, pn, $1);
  }
  else $$ = nullptr;
  leaveScopeErrCheck(EVENT_BLOCK_KIND);
}
| event_pattern TOK_PAT_PLUS {
  if($1 != nullptr) {
	$$ = new PatNode(BasePatNode::PLUS, $1);
  }
  else $$ = nullptr;
  leaveScopeErrCheck(EVENT_BLOCK_KIND);
}
*/
;

// One of the difficulties with the grammar is that there is no keyword that
// identifies the beginning of a rule. This makes it difficult to introduce an
// action to enter rule scope. Our work-around is as follows. Whenever we come
// across the first symbol that needs to be entered into rule scope, and we have
// not entered the rule scope yet (the former possibility arises if we see a
// label for the rule) then we enter the rule scope. Note that the first symbol
// we will encounter in a rule is an event name (including TOK_ANY). However,
// event name may appear in many places: specifically in a mapping, in a rule or
// in the middle of any pattern. So we need to eliminate cases where we are
// processing a mapping (ie isMapping is set), and cases where we are in the
// middle (not beginning) of a pattern (based on the variable ruleLevel).
// Specifically we use ruleLevel to identify whether we are at the outermost
// level (ie preceding first event in a pattern) in a rule. ruleLevel is reset
// on seeing a semicolon at the end of a rule.
//

event: TOK_EV_IDENT {
  // This production seems to be used only when an event name is used
  // in a pattern without a following lparen: basically, event with no args.
  // So, in the very last line below, we enter an argument vector of len=0.
  $$ = nullptr;
  SymTabEntry* ste = stm.lookUp($1);
  if ((!isMapping) && (ruleLevel++ == 0)) {
     BlockEntry* re = new RuleBlockEntry();
     if (insertSymbolErrCheck(re) == OK)
        enterScopeErrCheck(re);
     doneWithPat = false;
  }
  if ((ste == nullptr) || (ste->kind() != EVENT_KIND))
     errMsg("Expected an event name, but got \"" + string($1) +"\"");
  else $$ = new PrimitivePatNode((EventEntry*)ste, new vector<VariableEntry*>);
}
| ev_with_args TOK_RPAREN {
   $$ = $1;
   if ($$ != nullptr) {
      if ($$->nparams() < $$->event()->type()->argTypes()->size())
         errMsgLn("Too few arguments for " + $$->event()->name());
   }
}
;

ev_with_args: evnm_lparen ev_arg {
   $$ = nullptr;
   SymTabEntry* ste = stm.lookUp($1);
   if ((ste == nullptr) || (ste->kind() != EVENT_KIND))
     errMsgLn("Expected an event name, but got \"" + string($1) +"\"");
   else {
      Type *tmp = ste->type();
      const vector<Type*> *argType = (tmp == nullptr)? nullptr : tmp->argTypes();
      if (argType == nullptr || argType->size() < 1)
         errMsgLn("Too many arguments for " + ste->name());
      else {
         VariableEntry *ve = 
            new VariableEntry($2, EVENT_PARAM_VAR, (*argType)[0]);
         ve->offset(new ValueNode(new Value((unsigned)0)));
         insertSymbolErrCheck(ve);
         $$ = new PrimitivePatNode((EventEntry*)ste);
         $$->addParam(ve);
      }
   }
}
| ev_with_args TOK_COMMA ev_arg {
   $$ = $1;
   if ($$ != nullptr) {
      EventEntry *ee = $1->event();
      if (ee != nullptr) {
         const vector<Type*> *argType = ee->type()->argTypes();
         if (argType == nullptr || argType->size() < $$->nparams()+1)
            errMsgLn("Too many arguments for " + ee->name());
         else {
            VariableEntry *ve = 
               new VariableEntry($3, EVENT_PARAM_VAR, (*argType)[$$->nparams()]);
            ve->offset(new ValueNode(new Value($$->nparams())));
            insertSymbolErrCheck(ve);
            $$->addParam(ve);
         }
      }
   }
}
;

ev_arg: TOK_IDENT {
   $$ = $1;
   if ($1[0] == '_' && $1[1] == '\0') {
      $$ = new char[12];
      strcpy($$, $1);
      strcat($$, to_string(vct++).data());
   }
};

evnm_lparen: ev_or_ident TOK_LPAREN {
   if ((!isMapping) && (ruleLevel++ == 0)) {
      BlockEntry* re = new RuleBlockEntry();
      if (insertSymbolErrCheck(re) == OK)
         enterScopeErrCheck(re);
      doneWithPat = false;
   }

   BlockEntry *be = new EventBlockEntry();
   if (insertSymbolErrCheck(be) == OK)
      enterScopeErrCheck(be);
   $$ = $1;
};

ev_or_ident: TOK_IDENT | TOK_EV_IDENT;

event_vbar: event TOK_BITOR {
	$$ = $1;
};

/******************************** STRUCT ********************************/

struct_decl: struct_header2 variable_decls_star TOK_RBRACE  {
  leaveScopeErrCheck(STRUCT_KIND); //leaving the subclass scope
  strHdrEnt = 0;
}
| TOK_STRUCT error TOK_RBRACE {
  errMsg("Syntax error, skipping the entire structure definition");
};

/* It is rather unlikely that we would use these productions unless
   we are actually seeing a struct declaration. For this reason, the
   likelihood of "guessing wrong" wrt to looking for an "{" during
   error recovery is small. As such, we leave error handling rules to
   use TOK_LBRACE rather than lbrace_or_semicol 
*/
struct_header2: struct_header TOK_LBRACE {
}
| struct_header error TOK_LBRACE {
  errMsg("Syntax error in struct declaration");
}
| struct_header1 error TOK_LBRACE {
  errMsg("Syntax error in struct declaration");
}
| TOK_STRUCT error TOK_LBRACE {
  string structName(newName("struct"));
  StructEntry *se = new StructEntry(structName);
  Type *te = new Type(se, Type::STRUCT);
  se->type(te);
  if (insertSymbolErrCheck(se) == OK) {
    enterScopeErrCheck(se);
  } 
  errMsg("Expecting struct name");
};

struct_header: TOK_STRUCT TOK_IDENT {
  strHdrEnt = 1;
  string structName($2);
  StructEntry *se = new StructEntry(structName);
  Type *te = new Type(se, Type::STRUCT);
  se->type(te);
  if (insertSymbolErrCheck(se) == OK) {
    enterScopeErrCheck(se);
    $$ = se;
  } 
  else $$ = nullptr;
} 
| struct_header1 opt_with {
  if ($1 != nullptr) {
    $1->addConstraint($2);
  }
  $$ = $1;
};

struct_header1: struct_header tok_colon_or_tok_or TOK_IDENT {
  if ($1 != nullptr) {
    string s($3);
    SymTabEntry *ste = stm.lookUp(s);
    if ((ste != nullptr) && (ste->kind() == STRUCT_KIND)) {
      $1->addParent((StructEntry*)ste);
      // addChild is implicitly done in addParent
      enterScopeErrCheck(ste); // Entering parent structs scope
	  strHdrEnt++;
      $$ = $1;
    }
    else {
      errMsg(string("Expecting a struct name, but found \"") + $3 + "\"");
      $$ = $1;
    }
  }
  else $$ = nullptr;
};

tok_colon_or_tok_or: TOK_KEY_OR {
  if (strHdrEnt <= 1) 
    errMsg("Expected \":\", but got \"or\"");
} 
| TOK_COLON {
  if (strHdrEnt != 1) 
    errMsg("Expected \"or\", but got \":\"");
};

opt_with: /* empty */ {
  leaveScopeErrCheck(STRUCT_KIND);   // leave parent struct scope
  // Need to add a null constraint to ensure that a parent struct
  // and its associated constrain appear in the same position on
  // the parent and constraint vectors.
  //}
  $$ = nullptr;
}
| TOK_WITH expr {
  leaveScopeErrCheck(STRUCT_KIND);   // leave parent struct scope
  $$ = $2;
};

variable_decls_star:  /* empty */
| variable_decls_star variable_decl TOK_SEMICOLON
| variable_decls_star error TOK_SEMICOLON {
  errMsg("Syntax error in variable declaration");
};

variable_decl: variable_decl1 opt_init {
  if ($1 != nullptr) {
    if (($2 != nullptr) && (strHdrEnt != 0))
      errMsgLn("Initializations not permitted in struct definitions");
	else $1->initVal($2);
  }
  $$ = $1;
}
| type error TOK_SEMICOLON {
  buffer(TOK_SEMICOLON); 
  errMsg("Syntax error while parsing variable declaration");
  $$=nullptr;
}
| variable_decl TOK_COMMA TOK_IDENT opt_init {
  string name($3);
  if ($1 != nullptr) {
    VariableEntry *ve = new VariableEntry(name, $1->varKind(), new Type(*$1->type()));
    insertSymbolErrCheck(ve);
    if (($4 != nullptr) && (strHdrEnt != 0))
      errMsgLn("Initializations not permitted in struct definitions");
    ve->initVal($4);
    $$ = ve;
  }
  else $$ = nullptr;
}
| array_decl {
  $$ = $1;
}
| variable_decl1 error TOK_SEMICOLON {
    buffer(TOK_SEMICOLON); 
	errMsg("Syntax error in variable declaration");
    $$=nullptr;
}
| array_decl error TOK_SEMICOLON {
    buffer(TOK_SEMICOLON); 
	errMsg("Syntax error in array declaration");
    $$=nullptr;
}
;

variable_decl1: type TOK_IDENT {
  string name($2);
  VarKind vk = (stm.currentScope(FUNCTION_KIND)||stm.currentScope(BLOCK_KIND))? 
     LOCAL_VAR: GLOBAL_VAR;
  VariableEntry* ve = new VariableEntry(name, vk, $1);
  insertSymbolErrCheck(ve);
  $$ = ve;
};

array_decl: variable_decl1 TOK_LBRACK expr TOK_RBRACK {
  if ($1 != nullptr) {
     if ($1->type()->tag() == Type::VECTOR || $1->type()->tag() == Type::TUPLE) {
        errMsg("List and Tuple arrays not allowed");
     }
     if ($3 != nullptr)
        $1->addSubscript($3);
  }
  $$ = $1;
}
| array_decl TOK_LBRACK expr TOK_RBRACK {
  if (($1 != nullptr) && ($3 != nullptr)) {
    $1->addSubscript($3);
  }
  $$ = $1;
};

opt_init: /*empty */ {
  $$ = nullptr;
}
| TOK_ASSIGN expr {
  $$ = $2; 
}
| TOK_ASSIGN error TOK_SEMICOLON {
  $$=nullptr;
  buffer(TOK_SEMICOLON); 
  errMsg("Syntax error in initialization expression");
};

/******************************** MODULE ********************************/

module_defn: module_decl module_body {}
| module_decl error TOK_INTERFACE {
	errMsg("Parse error in module body, skipping to next interface");
	resetVars();
	buffer(TOK_INTERFACE); 
	stm.leaveToScope(GLOBAL_KIND);
}
| module_decl error tok_module_or_state_machine {
	errMsg("Parse error in module body, skipping to next module");
	resetVars();
	if ($3 == 0) 
	  buffer(TOK_MODULE); 
	else buffer(TOK_STATEMACHINE); 
	stm.leaveToScope(GLOBAL_KIND);
}
;

rparen_lbrace: TOK_RPAREN TOK_LBRACE {
	if (!seenLparen)
	  errMsg("Missing open parenthesis");
}
| TOK_LBRACE {
   if ((seenLparen) && (!YYRECOVERING()))
	 errMsg("missing closing parenthesis");
}
| TOK_RPAREN {
	errMsg("Missing \"{\"");
	if (!seenLparen)
	  errMsg("Missing open parenthesis");
};

module_decl: module_header module_formal_param_list rparen_lbrace {
  if (($1 != nullptr) && ($2 != nullptr)) {
    if ($1->name() == "main")
       errMsg("Main module cannot take any parameters");
    $1->type()->argTypes($2);
  }
  $$ = $1;
  inModuleHeader = false;
}
| module_header error TOK_LBRACE {
  errMsg("syntax error while parsing module formal parameters");
  $$=nullptr;
};

module_decl1: module_header module_formal_param_list TOK_RPAREN {
  if (($1 != nullptr) && ($2 != nullptr))
    $1->type()->argTypes($2);
  $$ = $1;
};

tok_module_or_state_machine: TOK_MODULE {
	$$=0;
    //yydebug=1;
}
| TOK_STATEMACHINE {
	$$=1;
};

module_header: tok_module_or_state_machine TOK_IDENT opt_lparen {
  string moduleName($2);
  ModuleEntry* me = new ModuleEntry(moduleName);
  me->isStateMachine($1 == 1);
  Type *te = new Type(Type::MODULE);
  me->type(te);
  if (insertSymbolErrCheck(me) == OK) {
    enterScopeErrCheck(me);
  }
  inModuleHeader = true;
  $$ = me;
}
| tok_module_or_state_machine error TOK_LBRACE {
  string moduleName(newName("module"));
  ModuleEntry* me = new ModuleEntry(moduleName);
  me->isStateMachine($1 == 1);
  Type *te = new Type(Type::MODULE);
  me->type(te);
  if (insertSymbolErrCheck(me) == OK) {
    enterScopeErrCheck(me);
  }
  $$ = me;
  errMsg("Expecting module name and parameters");
  seenLparen=false;
  //inModuleHeader = true;
  inModuleHeader = false;
  buffer(TOK_LBRACE);
};

module_formal_param_list: /*empty*/ {
  $$ = nullptr;
}
| module_formal_param_plus {
  $$ = $1;
};

module_formal_param_plus: module_formal_param {
  $$ = new vector<Type*>();
  if ($1 != nullptr) 
    $$->push_back($1);
}
| module_formal_param_plus TOK_COMMA module_formal_param {
  if ($3 != nullptr) 
    $$->push_back($3);
};

module_formal_param: formal_param {
  if ($1 != nullptr)
    $$ = $1->type();
  else $$ = nullptr;
}
| function_decl {
  if ($1 != nullptr) 
    $$ = $1->type();
  else $$ = nullptr;
  leaveScopeErrCheck(FUNCTION_KIND);
}
| module_decl1 {
  if ($1 != nullptr)
    $$ = $1->type();
  else $$ = nullptr;
  leaveScopeErrCheck(MODULE_KIND);
}
| event_decl {
  if ($1 != nullptr)
    $$ = $1->type();
  else $$ = nullptr;
  leaveScopeErrCheck(EVENT_KIND);
};

module_body: module_declaration_star rule_star
                       on_stmt_star TOK_RBRACE opt_semicolon {
  leaveScopeErrCheck(MODULE_KIND);
};

rule_star: /* empty */ {
  /* warnMsg("Module contains no rules."); */
}
| rule_plus {}
;

module_declaration_star: /* empty */
| module_declaration_star module_declaration
| module_declaration_star error TOK_SEMICOLON {
	errMsgLn("Syntax error (Note: statements are permitted "
             "only inside event reactions.");
    resetVars();
}
;

module_declaration: TOK_SEMICOLON
| states {
	checkStateMachine("states");
  }
| startstate {
	checkStateMachine("startstate");
  }
| finalstate {
	checkStateMachine("finalstate");
  }
| variable_decl TOK_SEMICOLON {
}
| function_decl TOK_SEMICOLON {
   leaveScopeErrCheck(FUNCTION_KIND);
   errMsg("Functions cannot be declared within modules");
}
| function_defn {
   errMsg("Functions cannot be declared within modules");
}
| enum_decl {
   errMsg("Enums cannot be declared within modules");
}
| mapping { 
  checkStateMachine("mapping");
  isMapping = false; 
  leaveScopeErrCheck(EVENT_BLOCK_KIND);
}
| timeout {
	checkStateMachine("timeout");
}
| TOK_INSTANTIATE invocation {
   if ($2 != nullptr) {
      if (($2->exprNodeType() != INV_NODE) ||
		  (((InvocationNode*)$2)->symTabEntry() == nullptr) ||
		  (((InvocationNode*)$2)->symTabEntry()->kind() != MODULE_KIND))
         errMsg("Module instantiation expected, but found a non-module expression");
      else {
         SymTabEntry *ste = stm.currentScope();
         if (ste->kind() == MODULE_KIND)
            ((ModuleEntry*)ste)->addInstantiate((InvocationNode*)$2);
      };
   }
};

states: states_header enum_body {
  if ($1 != nullptr) {
    VariableEntry *ve = new VariableEntry("_state", GLOBAL_VAR, 
										  new Type(*$1->type()));
    insertSymbolErrCheck(ve);
  }
}
| TOK_STATES error rbrace_or_semicol {
  errMsg("Syntax error while parsing state list");
};

states_header: TOK_STATES TOK_LBRACE {
  EnumEntry* ee = new EnumEntry("states");
  Type* et = new Type(ee, Type::ENUM);
  ee->type(et);
  if (insertSymbolErrCheck(ee) == OK) 
    enterScopeErrCheck(ee); 
  /* There is no need to enter scope here, but doing so is a convenient 
     way to pass ee to the rest of the production (enum_body) */
  $$ = ee;
};

startstate: TOK_STARTSTATE TOK_ENUMVAL TOK_SEMICOLON {
  string state($2);
  SymTabEntry *start = stm.lookUp(state);
  if ((start !=  nullptr) && (start->kind() == ENUM_ELEM_KIND))
    if (stm.currentScope()->kind() == MODULE_KIND)
      ((ModuleEntry*)stm.currentScope())->startState((EnumElemEntry*)start);
    else errMsgLn("Start statement is valid only at the module scope.");
  else errMsgLn("Invalid type for the argument of start state declaration.");
}
| TOK_STARTSTATE error TOK_SEMICOLON {
  errMsg("parse error, expecting identifier");
};

finalstate: TOK_FINALSTATE TOK_ENUMVAL TOK_SEMICOLON {
  string state($2);
  SymTabEntry *final = stm.lookUp(state);
  if ((final !=  nullptr) && (final->kind() == ENUM_ELEM_KIND))
    if (stm.currentScope()->kind() == MODULE_KIND)
      ((ModuleEntry*)stm.currentScope())->finalState((EnumElemEntry*)final);
    else errMsgLn("Final statement is valid only at the module scope.");
  else errMsgLn("Invalid type for the argument of final state declaration.");
}
| TOK_FINALSTATE error TOK_SEMICOLON {
  errMsg("parse error, in finalstate declaration");
};

mapping: map event TOK_WHEN expr TOK_SEMICOLON {
  SymTabEntry *me = stm.currentScope(MODULE_KIND);
  if (me == nullptr)
    errMsgLn("Map statement is valid only within a module");
  else if (($2 != nullptr) && ($4 != nullptr)) {
    $2->cond($4);
    ((ModuleEntry*)me)->addMap($2);
  }
}
| map event TOK_WHEN error TOK_SEMICOLON {
  resetVars();
  errMsg("parse error within map condition");
}
| map event error TOK_SEMICOLON {
  resetVars();
  errMsg("parse error in map statement, expecting key word 'when'");
}
| map error TOK_SEMICOLON {
  resetVars();
  errMsg("parse error in map statement, expecting an event");
};

map: TOK_MAP {
  isMapping = true;
};

timeout: TOK_TIMEOUT timeval opt_in state_list TOK_SEMICOLON {
  SymTabEntry *me = stm.currentScope(MODULE_KIND);
  if (me == nullptr) 
    errMsgLn("Timout statement must appear within a module");
  else if ($4 != nullptr)
    ((ModuleEntry*)me)->timeout($4, $2);
}
| TOK_TIMEOUT error TOK_SEMICOLON {
  errMsg("parse error in timeout statement");
};

state_list: TOK_ALL {
  SymTabEntry *ste = stm.lookUp("states");
  if ((ste == nullptr) || (ste->kind() != ENUM_KIND)) {
    errMsgLn("Timeout statement must appear after a 'states' statement");
    $$ = nullptr;
  }
  else 
    $$ = ((EnumEntry*)ste)->elems(); // @@@@ ptr ownership a problem
}
| TOK_LBRACE states_plus TOK_RBRACE {
  $$ = nullptr;
  if ($2 != nullptr) {
    $$ = new vector<EnumElemEntry*>();
    int i = 0, n = $2->size();
    for (; i < n;i++) {
      SymTabEntry *ste = stm.lookUp((*$2)[i]);
      if ((ste == nullptr) || (ste->kind() != ENUM_ELEM_KIND)) 
        errMsgLn("\"" + (*$2)[i] + "\" is not a valid state name");
      else $$->push_back((EnumElemEntry*)ste);
    }
  }
};

states_plus: TOK_ENUMVAL {
  $$ = new vector<string>();
  $$->push_back(string($1));
}
| states_plus TOK_COMMA TOK_ENUMVAL {
  $1->push_back(string($3));
  $$ = $1;
};

timeval: expr TOK_IDENT {
  string unit_time($2);
  if ((unit_time == "s") || (unit_time == "sec") || (unit_time == "second") || (unit_time == "seconds"))
    $$ = new OpNode(OpNode::MULT, $1, new ValueNode(new Value(1000000000l)));
  else if ((unit_time == "ms") || (unit_time == "millisecond") || (unit_time == "milliseconds"))
    $$ = new OpNode(OpNode::MULT, $1, new ValueNode(new Value(1000000l)));
  else if ((unit_time == "us") || (unit_time == "microsecond") || (unit_time == "microseconds"))
    $$ = new OpNode(OpNode::MULT, $1, new ValueNode(new Value(1000l)));
  else if ((unit_time == "ns") || (unit_time == "nanosecond") 
           || (unit_time == "nanoseconds"))
    $$ = $1;
  else {
    errMsg("Unknown time unit: please use seconds (s), or "
       "milli/micro/nanoseconds (ms/us/ns)");
    $$ = $1;
  }
};

opt_in: /* empty */
| TOK_IN 
;

/******************************** RULES ********************************/

rule_plus: rule { }
| rule_plus rule { }
;

rule: rule_body {
  leaveScopeErrCheck(RULE_BLOCK_KIND);
  if (stm.currentScope()->kind() == MODULE_KIND) {
    LabelEntry* le = new LabelEntry(newName("rule"), $1);
    insertSymbolErrCheck(le);
    if ($1 != nullptr) 
      $1->label(le);
  }
  ruleLevel = 0;
}
| TOK_IDENT TOK_COLON rule_body {
  leaveScopeErrCheck(RULE_BLOCK_KIND);
  if (stm.currentScope()->kind() == MODULE_KIND) {
    LabelEntry* le = new LabelEntry($1, $3);
    insertSymbolErrCheck(le);
    if ($3 != nullptr) 
      $3->label(le);
  }
  ruleLevel = 0;
};

rule_body: rule_body1 stmt {
  $$ = nullptr;
  if (($1 != nullptr) && ($2 != nullptr)) {
    RuleNode* rn = new RuleNode((BlockEntry*)stm.currentScope(), $1, $2);
    SymTabEntry* me = stm.currentScope(MODULE_KIND);
    SymTabEntry* re = stm.currentScope(RULE_BLOCK_KIND);
    if (re == nullptr || me == nullptr)
      internalErr("rule body is appearing in invalid scope.");
    else {
      ((ModuleEntry*)me)->addRule(rn);
      $$ = rn;
    }
  }
}
;

rule_body1: event_pattern TOK_ARROW {
  $$=$1;
  doneWithPat = true;
}
| event_pattern error TOK_ARROW {
  $$=$1;
  //buffer(TOK_ARROW); 
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| primitive_pat error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| TOK_LPAREN error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| TOK_NOT error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| rewbr_var error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| event error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| strconst error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| event_vbar error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| evnm_lparen error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
| TOK_EV_IDENT error TOK_ARROW {
  $$ = nullptr;
  errMsg("Syntax error, skipping event pattern");
  doneWithPat = true;
}
;

/******************************** ON STATEMENT *****************************/

on_stmt_star: /*empty*/
| on_stmt_plus
;

on_stmt_plus: on_stmt
| on_stmt_plus on_stmt
;

on_stmt: on_trans_specialize_freq_or_val TOK_FA_RATE 
             expr opt_learn_cond opt_detect_cond TOK_SEMICOLON {
  if ($1 != nullptr) {
    SymTabEntry *me = stm.currentScope(MODULE_KIND);
    $1->maxprob($3); // In transition from falseAlarmRate
    if (me == nullptr)
      errMsg("\"on\" statement is valid only within a module");
    else ((ModuleEntry*)me)->addOnEntry($1);
    $1->learnCond($4);
    $1->detectCond($5);
  }
  leaveScopeErrCheck(ON_KIND);
}
| on_trans_specialize_freq_or_val error TOK_SEMICOLON {
  errMsg("Syntax error towards the end of \"on\" statement");
  leaveScopeErrCheck(ON_KIND);
}
| TOK_ON  error TOK_SEMICOLON {
  errMsg("Syntax error in \"on\" statement");
};

on_trans_specialize_freq_or_val: 
  on_trans_specialize_freq 
| on_trans_specialize_val
;

on_trans_specialize_freq:
  on_trans_specialize TOK_HISTO TOK_FREQ timeval TOK_LBRACK expr TOK_RBRACK {
  if (($1 != nullptr) && ($4 != nullptr) && ($6 != nullptr)) {
    $1->interval($4, $6);
  }
  $$ = $1;
};

on_trans_specialize_val: 
  on_trans_specialize TOK_HISTO expr {
  if (($1 != nullptr) && ($3 != nullptr))
     $1->value($3);
  $$ = $1;
};

on_trans_specialize: on_trans_spec {
  $$ = $1;
}
| on_trans_spec specialize { // %%%% CHANGED from onstmt2 to onstmt3
  if (($1 != nullptr) && ($2 != nullptr))
    $1->wrt($2);
  $$ = $1;
};

specialize: wrt1 TOK_MAXSIZE expr {
  SymTabEntry* oe = stm.currentScope();
  if ((oe != nullptr) && (oe->kind() == ON_KIND)) 
	if ($1 == nullptr)
  	  errMsg("Size information is invalid with \"trans\"");
	else ((OnEntry*)oe)->size($3);
  else errMsg("unexpected case");
  $$ = $1;
};

wrt1: TOK_SPECIALIZE TOK_LPAREN expr_list TOK_RPAREN {
  $$ = $3;
/* %%%% CHANGED
}
| TOK_WRT TOK_TRANS {
  $$ = nullptr;
*/
};

on_trans_spec: on_trans_list {
  $$ = $1;
}
| on_trans_list TOK_WHEN expr {
  $$ = $1;
  if (($$ != nullptr) && ($3 != nullptr))
    $$->cond($3);
};

on_trans_list: TOK_ON trans_list {
  /* Return an OnEntry representing this on statement. */
  OnEntry *onEnt = new OnEntry(newName("on"), $2);
  if (insertSymbolErrCheck(onEnt) == OK)
    enterScopeErrCheck(onEnt);
  $$ = onEnt;
}
| TOK_ON TOK_EACHOF trans_list {
  OnEntry *onEnt = new OnEntry(newName("on"), $3);
  if (insertSymbolErrCheck(onEnt) == OK)
    enterScopeErrCheck(onEnt);
  $$ = onEnt;
  $$->onEachOf();
};

trans_list: TOK_ALL { // default is all of the rules
  SymTabEntry *me = stm.currentScope(MODULE_KIND);
  if (me != nullptr) // No need to report error, as it is being done elsewhere
    $$ = new vector<RuleNode*>(((ModuleEntry*)me)->rules());
}
| TOK_LBRACE labels TOK_RBRACE {
  $$ = $2;
};

labels: TOK_IDENT {
  $$ = new vector<RuleNode*>();
  SymTabEntry *le = stm.lookUp($1);
  if ((le == nullptr) || (le->kind() != LABEL_KIND))
    errMsg(string("\"") + $1 + "\" is not a valid rule label");
  else $$->push_back(((LabelEntry*)le)->rule());
}
| labels TOK_COMMA TOK_IDENT {
  $$ = $1;
  SymTabEntry *le = stm.lookUp($3);
  if ((le == nullptr) || (le->kind() != LABEL_KIND))
    errMsg(string("\"") + $3 + "\" is not a valid rule label");
  else if ($$ != nullptr)
     $$->push_back(((LabelEntry*)le)->rule());
};

opt_learn_cond: /* empty */ {
   $$ = nullptr;
}
| TOK_LEARN TOK_IF expr {
   $$ = $3;
};

opt_detect_cond: /* empty */ {
   $$ = nullptr;
}
| TOK_DETECT TOK_IF expr {
   $$ = $3;
};

/******************************** Type ********************************/

type: type1 {
  $$ = $1;
}
| TOK_CONST type1 { 
  if ($2 != nullptr) {
     /*if (Type::isPrimitive($2->tag()))
        errMsg("Primitive types cannot be declared as const");
     else*/ $2->constType(true);
  }
  $$ = $2;
};

type1: baseType {
  $$ = $1;
}
| TOK_IDENT { 
  string ident($1);
  SymTabEntry* ste = stm.lookUp(ident);
  if (ste != nullptr) {
    switch(ste->kind()) {
      /* Note: currently, there is no way to give a name to list, tuple,
     module, event, or function type; hence only 3 cases to handle */
    case ENUM_KIND:
    case STRUCT_KIND:
    case CLASS_KIND:
      $$ = new Type(*ste->type());
      break;
    default:
      errMsgLn(string("Unknown type \"")+$1+"\"");
      $$ = nullptr;
      break;
    }
  } 
  else {
    errMsgLn(string("Undeclared type \"")+$1+"\"");
    $$ = nullptr;
  }
}
| type1 TOK_LIST  {
  if ($1 != nullptr) 
    $$ = new Type($1, Type::TypeTag(Type::VECTOR));
  else $$ = nullptr;
}
| type1 TOK_SET  {
  if ($1 != nullptr) 
    $$ = new Type($1, Type::TypeTag(Type::SET));
  else $$ = nullptr;
}
| TOK_TUPLE TOK_LPAREN type_tuple TOK_RPAREN {
  if ($3 != nullptr) 
    $$ = new Type($3, Type::TypeTag(Type::TUPLE));
  else $$ = nullptr;
}
| TOK_TUPLE TOK_LPAREN error TOK_RPAREN {
   $$ = nullptr;
   errMsgLn("Invalid tuple type declaration");
}
| TOK_DICT TOK_LPAREN type TOK_COMMA type TOK_RPAREN {
  if ($3 != nullptr && $5 != nullptr) {
    $$ = new Type($3, $5);
  }
  else $$ = nullptr;
}
;

type_tuple: type TOK_COMMA type {
  $$ = new vector<Type*>();
  if (($1 != nullptr) && ($3 != nullptr)) {
    $$->push_back($1);
    $$->push_back($3);
  }
}
| type_tuple TOK_COMMA type {
  $1->push_back($3);
  $$ = $1;
};

baseType: TOK_VOID { 
  $$ = new Type(Type::VOID);
}
| TOK_BOOL {
  $$ = new Type(Type::BOOL);
}
| TOK_STRING { 
  $$ = new Type(Type::STRING);
}
| TOK_REGEXP { 
  $$ = new Type(Type::REGEXPR);
}
| TOK_BIT { 
  $$ = new Type(Type::BIT);
}
| TOK_CHAR { 
  $$ = new Type(Type::CHAR);
}
| TOK_SHORT { 
  $$ = new Type(Type::SHORT);
}
| TOK_UNSIGNED TOK_CHAR { 
  $$ = new Type(Type::UCHAR);
}
| TOK_UNSIGNED TOK_SHORT {
  $$ = new Type(Type::USHORT);
}
| TOK_UNSIGNED TOK_INT {
  $$ = new Type(Type::UINT);
}
| TOK_UNSIGNED {
  $$ = new Type(Type::UINT);
}  
| TOK_INT {
  $$ = new Type(Type::INT);
}
| TOK_UNSIGNED TOK_LONG {
  $$ = new Type(Type::ULONG);
}
| TOK_LONG {
  $$ = new Type(Type::LONG);
}
| TOK_DOUBLE { 
  $$ = new Type(Type::DOUBLE);
} 
;
/******************************** EXPR ********************************/

expr_list: /* empty */ {
  $$ = nullptr;
}
| expr_plus {
  $$ = $1;
}; 

expr_plus: expr {
  $$ = new vector<ExprNode*>();
  $$->push_back($1);
}
| expr_plus TOK_COMMA expr {
  $1->push_back($3);
  $$ = $1;
};

expr: TOK_LPAREN expr TOK_RPAREN {
  $$ = $2;
}
| ref_expr {
  $$ = $1;
}
/*
| asg { // Permitted within event condition, to assign to rule vars
  if ((stm.currentScope(RULE_BLOCK_KIND) != nullptr &&
       !doneWithPat) || (isMapping == true)) {
     if ($1 != nullptr && $1->checkRuleVarAsg())
        $$ = $1;
     else errMsg("This variable cannot be assigned in this context");
  }
  else errMsg("Assignments are invalid in this context");
  $$ = $1;
}
*/
/*
| method_invocation {
  $$ = $1;
}*/
| invocation {
  $$ = $1;
}
| literal {
  if ($1 != nullptr)
    $$ = new ValueNode($1);
  else $$ = nullptr;
}
/*
| TOK_SIZEOF TOK_LPAREN TOK_IDENT TOK_RPAREN {
   // @@@@ Do we need sizeof? BMSL has no arrays or other data structure
   // @@@@ whose size needs to be specified in advance. As a result, motivation
   // @@@@ for sizeof is unclear. If it is really necessary, then we probably
   // @@@@ needs to accept a refExpr argument rather than TOK_IDENT
  SymTabEntry* ste = stm.lookUp($3);
  if ((ste == nullptr) || 
      ((ste->kind() != STRUCT_KIND) && 
       (ste->kind() != VARIABLE_KIND))) {
    errMsg(string("Invalid operand for sizeof operator: \"")+$3+"\"");
    $$ = nullptr;
  }
  else $$ = new SizeOfNode(ste);
}
*/
| TOK_MINUS expr %prec TOK_UMINUS { 
  if ($2 != nullptr && $2->exprNodeType() == VALUE_NODE) {
     const Value* v = $2->value();
     if (Type::isSigned(v->type()->tag()))
        $$ = new ValueNode(new Value(((long)-(v->lval()))));
     else if (Type::isNumeric(v->type()->tag()))
        $$ = new ValueNode(new Value(((long)(-(v->ulval())))));
     else $$ = new OpNode(OpNode::UMINUS,$2);
  }
  else $$ = new OpNode(OpNode::UMINUS,$2);   
}
| expr TOK_PLUS expr { 
  $$= new OpNode(OpNode::PLUS,$1,$3);
}
| expr TOK_MINUS expr { 
  $$= new OpNode(OpNode::MINUS,$1,$3);
}
| expr TOK_MULT  expr { 
  $$= new OpNode(OpNode::MULT,$1,$3);
}
| expr TOK_DIV  expr { 
  $$= new OpNode(OpNode::DIV,$1,$3);
}
| expr TOK_MOD  expr { 
  $$= new OpNode(OpNode::MOD,$1,$3);
}
| expr TOK_EQ expr { 
  $$= new OpNode(OpNode::EQ,$1,$3);
}
| expr TOK_NE expr { 
  $$= new OpNode(OpNode::NE,$1,$3);
}
| expr TOK_EQREG regexp {
   SymTabEntry* ste; VariableEntry* ve; 
   if ($3 != nullptr) {
      if (!lex_squoteRegexp && !lex_dquoteRegexp &&
          $3->isString() && (ste = stm.lookUp($3->getPrtString())) &&
          ste->kind() == VARIABLE_KIND && (ve = (VariableEntry*)ste) &&
          ve->varKind() == MODULE_PARAM_VAR) {
         // 
         $$ = new OpNode(OpNode::EQREG, $1, 
                         new RefExprNode(nullptr, ste->name(), ste));
      }
      else {      
         /* RegExp matches can be announced only at the end of strings */
         RegExpr* re = new RegExpr(RegExpr::SEQ, $3, new RegExpr(RegExpr::END));
         $$= new OpNode(OpNode::EQREG, $1, new ValueNode(new Value(re)));
      }
   }
} 
/*
| expr TOK_NEQREG regexp {
   SymTabEntry* ste; VariableEntry* ve; 
   if (!lex_squoteRegexp && !lex_dquoteRegexp &&
       $3->isString() && (ste = stm.lookUp($3->getPrtString())) &&
       ste->kind() == VARIABLE_KIND && (ve = (VariableEntry*)ste) &&
       ve->varKind() == MODULE_PARAM_VAR) {
      // 
      $$ = new OpNode(OpNode::NEQREG, $1, 
                      new RefExprNode(nullptr, ste->name(), ste));
   }
   else {      
      RegExpr* re = new RegExpr(RegExpr::SEQ, $3, new RegExpr(RegExpr::END));
      $$= new OpNode(OpNode::NEQREG, $1, new ValueNode(new Value(re)));
   }
}
*/
| expr TOK_GT expr { 
  $$= new OpNode(OpNode::GT,$1,$3);
}
| expr TOK_LT expr { 
  $$= new OpNode(OpNode::LT,$1,$3);
}
| expr TOK_GE expr { 
  $$= new OpNode(OpNode::GE,$1,$3);
}
| expr TOK_LE expr { 
  $$= new OpNode(OpNode::LE,$1,$3);
}
| expr TOK_IN expr {
  $$ = new OpNode(OpNode::CONTAINS, $3, $1);
}
| expr TOK_AND expr { 
  $$= new OpNode(OpNode::AND,$1,$3);
}
| expr TOK_OR  expr { 
  $$= new OpNode(OpNode::OR,$1,$3);
}
| expr TOK_SEQ_AND expr { 
  $$= new OpNode(OpNode::SEQ_AND,$1,$3);
}
| expr TOK_SEQ_OR  expr { 
  $$= new OpNode(OpNode::SEQ_OR,$1,$3);
}
| TOK_NOT expr { 
  $$= new OpNode(OpNode::NOT,$2);
}
| TOK_BITNOT expr { 
  $$= new OpNode(OpNode::BITNOT,$2);
}
| expr TOK_BITAND expr { 
  $$= new OpNode(OpNode::BITAND,$1,$3);
}
| expr TOK_BITOR expr { 
  $$= new OpNode(OpNode::BITOR,$1,$3);
}
| expr TOK_BITXOR expr { 
  $$= new OpNode(OpNode::BITXOR,$1,$3);
}
| expr TOK_SHL expr { 
  $$= new OpNode(OpNode::SHL,$1,$3);
}
| expr TOK_SHR expr{ 
  $$= new OpNode(OpNode::SHR,$1,$3);
}
| TOK_LPAREN expr_tuple TOK_RPAREN {
  $$ = new OpNode(OpNode::TUPLE, $2);
}
| TOK_LBRACE expr_list TOK_RBRACE {
  $$ = new OpNode(OpNode::SET, $2);
}
| TOK_LBRACE dictval_plus TOK_RBRACE {
  $$ = new OpNode(OpNode::DICT, $2);
}
| TOK_LBRACK expr_list TOK_RBRACK {
  $$ = new OpNode(OpNode::VECTOR, $2);
}
;

dictval_plus: dictval {
   $$ =  new vector<ExprNode*>();
   $$->push_back($1);
}
| dictval_plus TOK_COMMA dictval {
   if ($1 != nullptr)
      $1->push_back($3);
   $$ = $1;
}
;

dictval: expr TOK_COLON expr {
   auto t = new vector<ExprNode*>();
   if ($1 != nullptr ) 
      t->push_back($1);
   if($3 != nullptr)
      t->push_back($3);
   $$ = new OpNode(OpNode::TUPLE, t);
};

expr_tuple: expr TOK_COMMA expr {
  $$ = new vector<ExprNode*>();
  if ($1 != nullptr ) 
    $$->push_back($1);
  if($3 != nullptr)
    $$->push_back($3);
}
| expr_tuple TOK_COMMA expr {
  if ($1 != nullptr && $3 != nullptr)
    $$->push_back($3);
  $$ = $1;
};

ref_expr: TOK_IDENT {
  $$ = nullptr;
  SymTabEntry* ste = stm.lookUp($1);
  if (ste != nullptr) {
    $$ = new RefExprNode(nullptr, $1, ste);
  } 
  else if ((((ste = stm.currentScope(RULE_BLOCK_KIND)) != nullptr) &&
             (!doneWithPat)) || (isMapping == true)) {
    //
    // If we are within rule scope, then it is OK to have undeclared
    // variables either as args of event patterns, or as lhs of
    // assignments within patterns. In either case, we should be within a
    // rule block scope and doneWithPat should be false.
    //
    // For such undeclared variables, we dont create an implicit declaration at
    // this point. Instead we leave the symTabEntry() field nullptr. Later, in
    // asg production, we deal with unset symTabEntry() field correctly. In
    // other places, (e.g., in the rhs of an assignment or within a condition)
    // the symTabEntry() is not dealt with, and it will lead to generation of an
    // error at a later point in compilation.
    //
    $$ = new RefExprNode(nullptr, $1, nullptr);
  } 
  else {
    if ((stm.currentScope() != nullptr) && 
        (stm.currentScope()->kind() == STRUCT_KIND)) {
      ste = stm.lookUpWithInheritance($1);
      if (ste != nullptr)
        $$ = new RefExprNode(nullptr, $1, ste);
    }
    if ($$ == nullptr)
      errMsgLn("Undeclared identifier \"" + string($1) + "\"");
  }
}
| ref_expr TOK_DOT TOK_IDENT {
  if ($1 != nullptr)
    $$ = new RefExprNode($1, $3); // $3 is not resolved until type check time
  else $$ = nullptr;
}
| ref_expr TOK_LBRACK expr TOK_RBRACK {
  if (($1 != nullptr) && ($3 != nullptr))
    $1->addSubs($3);
  $$ = $1;
};

regexp: TOK_RE {
   $$ = $1;
}
| TOK_RE_LPAREN regexp TOK_RE_RPAREN {
   $$ = $2;
}
| regexp TOK_RE_STAR {
   $$ = new RegExpr(RegExpr::STAR, $1);
}
| regexp TOK_RE_PLUS {
   $$ = new RegExpr(RegExpr::PLUS, $1);
}
| regexp TOK_RE_QMARK {
   $$ = new RegExpr(RegExpr::QMARK, $1);
}
| regexp regexp %prec TOK_RE_SEQ {
   $$ = new RegExpr(RegExpr::SEQ, $1, $2);
}
| regexp TOK_RE_OR regexp {
   $$ = new RegExpr(RegExpr::ALT, $1, $3);
};

ev_or_fun: ev_or_ident {
  // To capture built-in and user-defined functions in a consistent way,
  // we encode the latter as OpNodes with op->code == INVALID
  OpNode::OpInfo* op = new OpNode::OpInfo();
  op->code_ = OpNode::INVALID;
  op->name_ = $1;
  $$ = op;
}
| TOK_EMPTY {
  $$ = &opInfo[OpNode::EMPTY];
}
| TOK_CLEAR {
  $$ = &opInfo[OpNode::CLEAR];
}
| TOK_LENGTH {
  $$ = &opInfo[OpNode::LENGTH];
}
| TOK_APPEND {
  $$ = &opInfo[OpNode::APPEND];
}
| TOK_UNION {
  $$ = &opInfo[OpNode::UNION];
}
| TOK_INSERT {
  $$ = &opInfo[OpNode::INS_VEC];
}
| TOK_ERASE {
  $$ = &opInfo[OpNode::ERASE];
}
| TOK_CONTAINS {
  $$ = &opInfo[OpNode::CONTAINS];
}
| TOK_REMOVE {
  $$ = &opInfo[OpNode::REMOVE];
}
| TOK_PRINT {
  $$ = &opInfo[OpNode::PRINT];
}
| TOK_PRINT_RAW {
  $$ = &opInfo[OpNode::PRINT_RAW];
}
;

invocation: ev_or_fun TOK_LPAREN expr_list TOK_RPAREN {
  $$ = nullptr;
  if ($1 != nullptr) {
     if ($1->code_ == OpNode::INVALID) {
        // To capture built-in and user-defined functions in a consistent way,
        // we encoded the latter as OpNodes with op->code == INVALID
        SymTabEntry *fe = stm.lookUp($1->name_);
        if ((fe != nullptr) && ((fe->kind() == FUNCTION_KIND) || 
                                (fe->kind() == MODULE_KIND)   ||
                                (fe->kind() == EVENT_KIND)))
	   $$ = new InvocationNode(fe, $3);
	else errMsgLn(string("\"") + $1->name_ + 
                      "\" is not a function or module name");
     }
     else {
	switch ($1->arity_) {
	case 1: 
	   if ($3->size() < 1)
              errMsg(string("Missing operand for \"")+$1->name_+"\"");
	   else $$ = new OpNode($1->code_, (*$3)[0]);
	   break;
	case 2:
	   if ($3->size() < 2)
              errMsg(string("Missing operand(s) for \"")+$1->name_+"\"");
	   else $$ = new OpNode($1->code_, (*$3)[0], (*$3)[1]);
	   break;

        default:
	   $$ = new OpNode($1->code_, $3);
	   break;
        }
     }
  }
};

literal: TOK_UINTNUM {
  $$ = new Value($1);
}
| TOK_ULONGNUM {
  $$ = new Value($1);
}
| TOK_DOUBLENUM {
  $$ = new Value($1);
}
| strconst {
  $$ = new Value($1);
}
| regexp {
   RegExpr* re = new RegExpr(RegExpr::SEQ, $1, new RegExpr(RegExpr::END));
   $$ = new Value(re);
}
| TOK_TRUE {
  $$ = new Value(Type::BaseTypeTag(Type::BOOL), true);
}
| TOK_FALSE {
  $$ = new Value(Type::BaseTypeTag(Type::BOOL), false);
}
| TOK_ENUMVAL {
  SymTabEntry* ste = stm.lookUp($1);
  assert(ste->kind() == ENUM_ELEM_KIND);
  $$ = new Value(((EnumElemEntry*)ste)->num(), ste);
};

strconst: TOK_STRCONST {
  lastline = yylineno;
  $$ = $1;
}
| strconst TOK_STRCONST {
  if ((lastline != yylineno) && ($1 != nullptr) && ($2 != nullptr)) {
	$$ = new char[strlen($1)+strlen($2)+4];
    strcpy($$, $1);
    strcat($$, $2);
  }
  else errMsg("Missing operator between string constants");
};
/******************************** STMT ********************************/

stmt_plus:  stmt {
  $$ = new list<StmtNode*>();
  if ($1 != nullptr)
    $$->push_back($1);
}
| stmt_plus stmt {
  if (($1 != nullptr) && ($2 != nullptr))
    $1->push_back($2);
  $$ = $1;
}
| stmt_plus variable_decl TOK_SEMICOLON {
   errMsg("Declarations should precede all statements");
   $$ = $1;
}
| stmt_plus error rbrace_or_semicol {
  errMsg("Syntax error in statement");
  $$=nullptr;
  if (seenRbrace) {
    buffer(TOK_RBRACE);
  }
};

// @@@@ Need to check what happens if a rule variable is assigned twice.
// @@@@ Such assignments can complicate semantics in some cases, so we might
// @@@@ as well disallow them.

asg: ref_expr TOK_ASSIGN expr {
  if (($1 != nullptr) && ($3 != nullptr)) {
    RefExprNode& rn = *$1;
    SymTabEntry* ste;
    if ((rn.base() == nullptr) && (rn.symTabEntry() == nullptr)) {
      if (((ste = stm.currentScope(RULE_BLOCK_KIND)) != nullptr) &&
        (!doneWithPat)) {
        // Undeclared variable within rule scope. We need to create a
        // declaration, and enter it into the rule scope.
        VariableEntry *ve=new VariableEntry(rn.ext(), RULE_VAR);
        ste->insert(ve);
        rn.symTabEntry(ve);
      }
    }
    $$ = new OpNode(OpNode::ASSIGN, $1, $3);
  }
  else $$ = nullptr;
}
| ref_expr TOK_ASSIGN error rbrace_or_semicol {
  $$ = nullptr;
  errMsg("Syntax error in assignment");
  if (seenRbrace)
    buffer(TOK_RBRACE);
  else buffer(TOK_SEMICOLON);
}
// @@@@ The following rule is highly questionable. On seeing ref_expr error,
// @@@@ we don't know if we are in a statement at all, so this error handling
// @@@@ is inappropriate here. But then, we dont have any good error 
// @@@@ error recovery rule for statements in general. So, if we dont include
// @@@@ this rule here, we end up with syntax errors being caught at higher
// @@@@ levels, where good recovery doesn't seem possible (since the error
// @@@@ is at the statement level). So, for now, we will leave this rule
// @@@@ enabled. Note that this rule can cause inappropriate error recovery,
// @@@@ in places where we aren't even seeing a statement, e.g., err12a.asl:43

/*| ref_expr error rbrace_or_semicol {
  $$ = nullptr;
  errMsg("Syntax error in statement or expression");
  if (seenRbrace)
    buffer(TOK_RBRACE);
  else buffer(TOK_SEMICOLON);
}*/
;

simple_stmt: TOK_SEMICOLON {
  $$ = new StmtNode();
}
| invocation TOK_SEMICOLON {
  if ($1 != nullptr)
    $$ = new ExprStmtNode($1);
  else $$ = new StmtNode();
}
| asg  TOK_SEMICOLON {
	if ($1 != nullptr)
  	  $$ = new ExprStmtNode($1);
	else new StmtNode();
}
| TOK_RETURN opt_expr rbrace_or_semicol {
  SymTabEntry *fe = stm.currentScope(FUNCTION_KIND);
  if (fe == nullptr)
	errMsg("Return statement is valid only within a function");
  $$ = new ReturnStmtNode($2, (FunctionEntry*)fe);
  if (seenRbrace) {
    buffer(TOK_RBRACE);
  }
}
| TOK_RETURN error rbrace_or_semicol  {
  errMsg("Syntax error in return expression");
  $$=nullptr;
  if (seenRbrace) {
    buffer(TOK_RBRACE);
  }
}
;

opt_expr: expr {
   $$ = $1;
}
| /*EMPTY*/ {
   $$ = nullptr;
}
;

stmt: simple_stmt {
  $$ = $1;
}
| compoundStmt opt_semicolon {
  $$ = $1;
}
| if_then_else_stmt {
  $$ = $1;
}
| list_for_stmt {
  $$ = $1;
}
| while_stmt {
};

lbrace_or_semicol: TOK_LBRACE {
  $$=0;
}
| TOK_SEMICOLON {
  $$=1;
}
;

compoundStmt: cstmt_lbrace variable_decls_star stmt_plus TOK_RBRACE {
  $$ = new CompoundStmtNode((BlockEntry*)stm.currentScope(), $3);
  leaveScopeErrCheck(BLOCK_KIND); 
}
| cstmt_lbrace error TOK_RBRACE {
  errMsg("Syntax error, expecting a statement");
  $$=nullptr;
  leaveScopeErrCheck(BLOCK_KIND); 
};

cstmt_lbrace: TOK_LBRACE {
  BlockEntry* be = new BlockEntry(newName("compound_stmt"));
  if (insertSymbolErrCheck(be) == OK) 
     enterScopeErrCheck(be);
}
;

opt_else: /* empty */ {
  $$=nullptr;
}
| TOK_ELSE stmt {
  $$=$2;
}
| TOK_ELSE error rbrace_or_semicol {
  $$=nullptr;
  errMsg("Syntax error in else statement");
  if (seenRbrace)
    buffer(TOK_RBRACE);
};

if_then_else_stmt: TOK_IF expr stmt opt_else {
  $$ = new IfNode($2, $3, $4); 
}
| TOK_IF error lbrace_or_semicol {
	errMsg("Syntax error in the condition part of \"if\" statement");
	if ($3 == 0)
		buffer(TOK_LBRACE);
    else buffer(TOK_SEMICOLON);
  } if_rest {
    $$=nullptr;
};

if_rest: stmt opt_else {} 
;

list_for_stmt: list_for_hdr1 stmt {
   if ($1 != nullptr && $2 != nullptr)
      $1->body($2);
   $$ = $1;
  leaveScopeErrCheck(BLOCK_KIND); 
}

list_for_hdr1: list_for_hdr expr opt_do {
  if (($1 != nullptr) && ($2 != nullptr))
    $$ = new ListForNode($1, $2);
}
| list_for_hdr error lbrace_or_semicol {
   if ($3 == 0)
      buffer(TOK_LBRACE);
   else buffer(TOK_SEMICOLON);
   errMsg("Syntax error in for statement");
   $$ = nullptr;
}
;

opt_do: TOK_DO {
}
|
/* empty */ {
   errMsg("Expected keyword 'do'");
}
;
list_for_hdr: TOK_FOREACH TOK_IDENT TOK_IN {
  BlockEntry* be = new BlockEntry(newName("list_for"));
  if (insertSymbolErrCheck(be) == OK) 
     enterScopeErrCheck(be);
  VariableEntry* ve = new VariableEntry($2, FOREACH_VAR);
  insertSymbolErrCheck(ve);
  RefExprNode *re = new RefExprNode(nullptr, $2, ve);
  $$ = re;
}
;

while_stmt: while_hdr stmt {
   if ($1 != nullptr && $2 != nullptr)
      $1->body($2);
   $$ = $1;
}
;

while_hdr: TOK_WHILE expr opt_do {
   $$ = new WhileNode($2);
}
| TOK_WHILE error lbrace_or_semicol {
   if ($3 == 0)
      buffer(TOK_LBRACE);
   else buffer(TOK_SEMICOLON);
   errMsg("Syntax error in while-statement condition");
   $$ = nullptr;
}
;   

opt_semicolon: /* empty */
| TOK_SEMICOLON
;
%%

