#include "util/STLutils.h"
#include "Ast.h"
#include "CondFac.h"
#include <algorithm>
#include "STEClasses.h"
#include <math.h>

 /******************* Ast operations related to MatchCond ********************/

MatchCond* RefExprNode::
normalize(bool neg, ostream& os, int indent) {
   // Invocation of normalize on RefExprNode is possible only when it isn't
   // inside a comparison. In this case, we must be dealing with a boolean var,
   // as in x or !x

   Value v = Value(neg? 0l : 1l);
   return new MatchCond(MatchCond::EQ, ext_, v);
}

MatchCond* ValueNode::
normalize(bool neg, ostream& os, int indent) {
   // Invocation of normalize on ValueNode is possible only when it isn't
   // inside a comparison. In this case, we must be dealing with a boolean cond,
   // as in "true" or "0"

   bool res = value()->bval();
   if (neg) res = !res;
   return new MatchCond(res);
}

IndexAsg<string> strToVarTx;
void emitDecl(const Type* t, const string& nm, ostream& os, int indent) {
   os << Type::name(t->tag()) << ' ' << nm; 
   endln(os, indent);
   if (Type::isString(t->tag())) {
      os << "const unsigned char* _c" << nm;
      endln(os, indent);
   }
}

MatchCond* InvocationNode::
normalize(bool neg, ostream& os, int indent) {
   bool isNew;
   string var = strToVarTx.getName(prtString(), isNew);
   if (isNew) emitDecl(type(), var, os, indent);
   MatchCond* asg = new MatchCond(MatchCond::ASG, var, this);
   OpNode cmpop(OpNode::EQ, new RefExprNode(nullptr, var), 
                new ValueNode(new Value(1l)));
   MatchCond* cmp = cmpop.normalize(neg, os, indent);
   return new MatchCond(MatchCond::SEQ_AND, asg, cmp);
}

MatchCond*
OpNode::normalize(bool neg, ostream& os, int indent) {
  string var; bool isNew;
  if (arg_[0]->exprNodeType() == REF_EXPR_NODE)
     var = ((RefExprNode*)arg_[0])->ext();
  else if (arg_[0]->exprNodeType() == INV_NODE) { 
     // Must be an expression consisting of const fn invocations
     if (!Type::isPrimitive(arg_[0]->type()->tag())) {
        errMsg("Left-hand side must have primitive type", arg_[0]);
        return nullptr;
     }
     var = strToVarTx.getName(arg_[0]->prtString(), isNew);
     if (isNew) emitDecl(arg_[0]->type(), var, os, indent);
     MatchCond* asg = new MatchCond(MatchCond::ASG, var, arg_[0]);
     OpNode cmpop(*this);
     cmpop.arg_[0] = new RefExprNode(nullptr, var);
     MatchCond* cmp = cmpop.normalize(neg, os, indent);
     return new MatchCond(MatchCond::SEQ_AND, asg, cmp);
  }

  OpCode oc = opCode_;
  OpNode* on = nullptr;

  if (neg && oc >= EQ && oc <= NEQREG) {
     switch (oc) {
     case EQ: oc = NE; break;
     case NE: oc = EQ; break;
     case GT: oc = LE; break;
     case LT: oc = GE; break;
     case GE: oc = LT; break;
     case LE: oc = GT; break;
     case EQREG: //oc = NEQREG; break;
     case NEQREG: //oc = EQREG; break;
        errMsg("Negation of regular expression matches within event patterns "
               "is unsupported, aborting", this);
        
     default: assert(false);
     }
     neg = false;
  }

  switch (oc) {
  case EQ:
  case NE:
     if (!arg_[1]->value())
        errMsg("Internal error: Right-hand side must be a constant", arg_[1]);
     else if ((arg_[0]->exprNodeType() != REF_EXPR_NODE) && 
         ((arg_[0]->exprNodeType() != OP_NODE) ||
          ((on = (OpNode*)arg_[0])->opCode_ != BITAND) ||
            (on->arg_[0]->exprNodeType() != REF_EXPR_NODE) ||
            (!on->arg_[1]->value()))) {
        errMsg("Left-hand side must be an event argument, a const function"
               " invocation, or be of the form 'var & const'", arg_[0]);
        return nullptr;
     }
     else if (on != nullptr) {
        var = ((RefExprNode*)on->arg_[0])->ext();
        MatchCond::FacOp c = (oc == EQ? MatchCond::EQBIT : MatchCond::NEQBIT);
        return new MatchCond(c, var, *on->arg_[1]->value(), *arg_[1]->value());
     }
     else {
        MatchCond::FacOp c = (oc == EQ? MatchCond::EQ : MatchCond::NE);
        return new MatchCond(c, var, *arg_[1]->value());
     }
     return nullptr;
 
   case EQREG:
   case NEQREG:
     if (arg_[0]->exprNodeType() != REF_EXPR_NODE) {
        errMsg("Left-hand side must be an event argument or a const function",
               arg_[0]);
       return nullptr;
     }
     else {
       arg_[0]->normalize(neg, os, indent);
       if (oc == EQREG) 
          return new MatchCond(MatchCond::EQREG, var, *arg_[1]->value());
       else return new MatchCond(MatchCond::NEQREG, var, *arg_[1]->value());
     }

  case GT:
  case LT:
  case GE:
  case LE:
     if (arg_[0]->exprNodeType() != REF_EXPR_NODE)
       errMsg("Left-hand side must be an event argument or a const function",
              arg_[0]);
     else if (!arg_[1]->value() || !arg_[1]->type()->isScalar())
       errMsg("Internal error: rhs must be a scalar constant", arg_[1]);
     else {
       MatchCond::FacOp c = (oc==GT || oc==GE)? MatchCond::GT : MatchCond::LT;
       MatchCond* rv = new MatchCond(c, var, *arg_[1]->value());
       if (oc == GE || oc == LE) {
          MatchCond* rv1 = new MatchCond(MatchCond::EQ, var, *arg_[1]->value());
          rv = new MatchCond(MatchCond::OR, rv, rv1);
       }
       return rv;
     }
     return nullptr;

#define mkBoolCond(op, neg) \
  new MatchCond(MatchCond:: op, arg_[0]->normalize(neg, os, indent), arg_[1]->normalize(neg, os, indent))

  case AND:
     return (neg? mkBoolCond(OR, true): mkBoolCond(AND, false));

  case OR:
    return (neg? mkBoolCond(AND, true): mkBoolCond(OR, false));

  case NOT:
     return arg_[0]->normalize(!neg, os, indent);

  case SEQ_AND: 
     return (neg? mkBoolCond(SEQ_OR, true): mkBoolCond(SEQ_AND, false));
  case SEQ_OR:
     return (neg? mkBoolCond(SEQ_AND, true): mkBoolCond(SEQ_OR, false));

  case CONTAINS:
    if (arg_[0]->exprNodeType() != REF_EXPR_NODE || !arg_[0]->type()->isScalar())
       errMsg("Left-hand side must be an event argument or a const function", 
              arg_[0]);
    else if (arg_[1]->exprNodeType() != OP_NODE || 
             (on = (OpNode*)arg_[0])->opCode_ != SET)
      errMsg("Internal error: rhs must be a constant-valued set", arg_[1]);
    else {
       set<Value> v;
       bool nonconst = false;
       for (auto a: on->arg_) {
          if (a->value() == nullptr)
             nonconst = true;
          else v.insert(*a->value());
       }
       if (nonconst)
          errMsg("Right-hand side must be a constant-valued set", arg_[1]);
       if (v.size() == 0 || v.size() > 8) {
          errMsg("Right-hand side must have 1 to 8 elements", arg_[1]);
          return nullptr;
       }
       else {
          MatchCond::FacOp op = neg? MatchCond::NE : MatchCond::EQ;
          MatchCond::FacOp connect = neg? MatchCond::AND : MatchCond::OR;
          auto it = v.begin(); 
          MatchCond *rv = new MatchCond(op, var, *it); it++;
          for (; it != v.end(); ++it ) {
             rv = new MatchCond(connect, new MatchCond(op, var, *it), rv);
          }
          return rv;
       }
    }
    return nullptr;

   default:
      errMsg(string("Operator '") + opInfo[oc].name_ + 
             "' is not supported in event pattern condition", this);
      return nullptr;
  }
}
#undef mkBoolCond

/************************ MatchCond utility functions *************************/

void MatchCond::
releaseResources() {
   if (op_ == IN || op_ == NOTIN) set_.clear();
   else if (op_ == AND || op_ == OR || op_ == SEQ_AND || op_ == SEQ_OR) {
      if (ch_[0]) {
         delete ch_[0]; 
         ch_[0] = nullptr;
      }
      if (ch_[1]) {
         delete ch_[1]; 
         ch_[1] = nullptr;
      }
   }
   else if (op_ >= EQBIT && op_ < ASG) {
      if (v_[0]) {
         delete v_[0]; 
         v_[0] = nullptr;
      }
      if (v_[1]) {
         delete v_[1]; 
         v_[1] = nullptr;
      }
   }
   // Not clearing expr_, we did not allocate it.
}

void MatchCond::
deepCopy(const MatchCond& fc) {
   releaseResources();
   op_ = fc.op_;
   charOp_ = fc.charOp_;
   var_ = fc.var_;
   switch (op_) {
   case AND: case OR: case SEQ_OR: case SEQ_AND:
      ch_[0] = new MatchCond();
      ch_[0]->deepCopy(*fc.ch_[0]); 
      ch_[1] = new MatchCond();
      ch_[1]->deepCopy(*fc.ch_[1]); 
      break;
   case IN: case NOTIN:
      set_ = fc.set_; break;
   case ASG:
      expr_ = fc.expr_; break; // only shallow copy, we don't own it.
   default:
      v_[0] = v_[1] = nullptr;
      if (fc.v_[0])
         v_[0] = new Value(*fc.v_[0]); 
      if (fc.v_[1])
         v_[1] = new Value(*fc.v_[1]); 
      break;
   }
}

bool MatchCond::
operator==(const MatchCond& fc) const {
   if (op_ != fc.op_ || var_ != fc.var_ || charOp_ != fc.charOp_) 
      return false;
   switch (op_) {
   case TRUE:
   case FALSE:
      return true;

   case AND:
   case OR:
   case SEQ_AND:
   case SEQ_OR:
      return ((ch_[0] == fc.ch_[0] 
                 || (ch_[0] && fc.ch_[0] && *ch_[0] == *fc.ch_[0])) && 
              (ch_[1] == fc.ch_[1] 
                 || (ch_[1] && fc.ch_[1] && *ch_[1] == *fc.ch_[1])));

   case IN:
   case NOTIN:
      return (set_ == fc.set_);

   case EQ:
   case NE:
   case GT:
   case LT:
   case EQREG:
   case NEQREG:
      return (*v_[0] == *fc.v_[0]);

   case EQBIT:
   case NEQBIT:
      return (*v_[0] == *fc.v_[0] && *v_[1] == *fc.v_[1]);

   case ASG:
      return true;
      // Because of the way we generate var names, the expressions will match
      // when the variable names match.

   default:
      assert(false);
   }
}

bool MatchCond::
operator<(const MatchCond& fc) const {
   if (op_ != fc.op_) return (op_ < fc.op_);
   if (charOp_ != fc.charOp_) return (charOp_ < fc.charOp_);
   if (var_ != fc.var_) return (var_ < fc.var_);

   switch (op_) {
   case AND:
   case OR:
   case SEQ_AND:
   case SEQ_OR:
      if (*ch_[0] != *fc.ch_[0]) return (*ch_[0] < *fc.ch_[0]);
      if (*ch_[1] != *fc.ch_[1]) return (*ch_[1] < *fc.ch_[1]);
      break;

   case IN:
   case NOTIN:
      if (set_ != fc.set_) return (set_ < fc.set_);
      break;

   case EQ:
   case NE:
   case GT:
   case LT:
   case EQREG:
   case NEQREG:
   case EQBIT:
   case NEQBIT:
      if (*v_[0] != *fc.v_[0]) return (*v_[0] < *fc.v_[0]);
      if (op_ == EQBIT || op_ == NEQBIT)
         if (*v_[1] != *fc.v_[1]) return (*v_[1] < *fc.v_[1]);
      break;

   default:
      break;
   }
   return false;
}

size_t MatchCond::
hash(size_t seed) const {
   unsigned char buf[256];
   unsigned j=0;
   size_t h;
   assert(op_ < AND || var_.size() < 224);
   buf[j++] = op_;
   buf[j++] = charOp_;
   if (op_ >= AND && op_ <= SEQ_OR) {
      h = ch_[0]->hash();
      memcpy(&buf[j], &h, 8); j += 8;
      h = ch_[1]->hash();
      memcpy(&buf[j], &h, 8); j += 8;
   }
   else if (op_ > SEQ_OR) {
      memcpy(&buf[j], var_.c_str(), var_.size()); j += var_.size();
      if (op_ == IN || op_ == NOTIN) {
         h = std::hash<set<Value>>()(set_);
         memcpy(&buf[j], &h, 8); j+= 8;
      }
      else if (op_ == EQBIT || op_ == NEQBIT) {
         h = v_[0]->ulval();
         memcpy(&buf[j], &h, 8); j+= 8;
         h = v_[1]->ulval();
         memcpy(&buf[j], &h, 8); j+= 8;
      }
      else if (op_ != ASG) { // For asg, the varname is unique for expr_,
         assert(op_ <= LT);  // so it is enough.
         if (v_[0]->isString()) {
            const string& s = v_[0]->sval();
            memcpy(&buf[j], s.c_str(), s.size()); j += s.size();
         }
         else if (v_[0]->isRegExpr()) {
            h = v_[0]->reval()->hash();
            memcpy(&buf[j], &h, 8); j += 8;
         }
         else if (v_[0]->isIntegral()) {
            h = v_[0]->ulval();
            memcpy(&buf[j], &h, 8); j += 8;
         }
         else if (v_[0]->isDouble()) {
            double d = v_[0]->dval();
            memcpy(&buf[j], &d, 8); j += 8;
         }
      }
   }
   return fasthash64(buf, j, seed);
};

void MatchCond::
print(ostream& os, int indent) const {
   string opname;
   switch (op_) {
   case TRUE:
      os << " true"; break;
   case FALSE:
      os << " false"; break;

   case AND:
      ch_[0]->print(os); os << " &&"; ch_[1]->print(os); break;

   case OR:
      os << '('; ch_[0]->print(os); os << " ||"; ch_[1]->print(os); os << ')'; 
      break;

   case SEQ_AND:
      ch_[0]->print(os); os << " &&&"; ch_[1]->print(os); break;

   case SEQ_OR:
      ch_[0]->print(os); os << " |||"; ch_[1]->print(os); break;

   case IN: {
      os << ' ' << var_ << " in ";
      ::print(set_, os); break;
   }

   case NOTIN: {
      os << ' ' << var_ << " not_in ";
      ::print(set_, os); break;
   }

   case EQ:
   case NE:
   case GT:
   case LT:
   case EQREG:
   case NEQREG:
   case EQBIT:
   case NEQBIT:
   case ASG:
      os << ' '; 
      if (charOp_) os << "*_c";
      os << var_;
      if (op_ == EQ) opname = "==";
      else if (op_ == NE) opname = "!=";
      else if (op_ == GT) opname = ">";
      else if (op_ == LT) opname = "<";
      else if (op_ == EQREG) opname = (charOp_? "==" : "=~");
      else if (op_ == NEQREG) opname = (charOp_? "!=" : "!~");
      else if (op_ == EQBIT) { opname = "=="; os << " & "; v_[1]->print(os); }
      else if (op_ == NEQBIT) { opname = "!="; os << " & "; v_[1]->print(os); }
      else if (op_ == ASG) { os << " ="; expr_->print(os, indent); break; }
      os << ' ' << opname;
      if (charOp_) {
         unsigned char c = cval();
         if (isprint(c)) {
            os << '\'';
            if (c == '\\')
               os << '\\';
            os << c << '\'';
         }
         else os << (0xff & (int)c);
      }
      else v_[0]->print(os);
      break;

   default:
      assert(false);
   }   
}

/************************ MatchCond main functions ************************/

bool MatchCond::
negate() { // Negate match condition, return false if it can't be negated
   switch (op_) {
   case TRUE: op_ = FALSE; break;
   case FALSE: op_ = TRUE; break;
   case IN: op_ = NOTIN; break;
   case NOTIN: op_ = IN; break;
   case EQBIT: op_ = NEQBIT; break;
   case NEQBIT: op_ = EQBIT; break;
   case EQ: op_ = NE; break;
   case NE: op_ = EQ; break;
   case EQREG: //op_ = NEQREG; break;
   case NEQREG: //op_ = EQREG; break;
        errMsg("Negation of regular expression matches within event patterns "
               "is unsupported, aborting");

   default:
      return false;
   }
   return true;
}

bool MatchCond::
opPresent(unsigned opset, string vname) const {
   switch (op_) {
   case AND:
   case OR:
   case SEQ_AND:
   case SEQ_OR:
      if (opset & (1<<op_))
         return true;
      else (ch_[0]->opPresent(opset, vname) || 
            ch_[1]->opPresent(opset, vname));

   case IN:
   case NOTIN:
   case EQBIT:
   case NEQBIT:
   case EQ:
   case NE:
   case EQREG:
   case NEQREG:
   case GT:
   case LT:
   case ASG:
      return ((vname == "" || vname == var_) && (opset & (1<<op_)));

   default:
      return (opset & (1<<op_));
   }
}

void MatchCond::
getVars(set<string>& s) const {
   switch (op_) {
   case AND:
   case OR:
   case SEQ_AND:
   case SEQ_OR:
      ch_[0]->getVars(s);
      ch_[1]->getVars(s);
      break;

   case IN:
   case NOTIN:
   case EQBIT:
   case NEQBIT:
   case EQ:
   case NE:
   case EQREG:
   case NEQREG:
   case GT:
   case LT:
   case ASG:
      s.insert(var_);
      break;

   default:
      break;
   }
}

//
// See Fig. 3 in Condition Factorization paper. In some cases, the results from
// that table have been generalized or modified, e.g., to support "in" 
// constraints, and to support inequalities rather than interval constraints.
//
// Returns false if residue value is somehow not recommended. This may indicate
// that the residue is inaccurate, or that the residue is even more expensive to
// test than this MatchCond. It may also be the case that the test fc is somehow
// not the right test for this MatchCond. On return, this MatchCond will
// represent a sound residue value (i.e., residue && fc => this), even though a
// return value of false should discourage the caller from using the residue.
//
bool 
MatchCond::residue(const MatchCond& fc, unsigned char& nextc) { 
   if (*this == fc) { 
      deepCopy(MatchCond(true));
      return true;
      // This is a very simple implementation of residue that works for all
      // types of tests, even those that don't have a constant rhs. It hence
      // serves as a useful fallback. Although it may not share subcomputations
      // across tests, it is (a) sound, and (b) ensure terminiation of factorizn
   }

   if (op_==AND || op_==OR || op_==SEQ_AND || op_==SEQ_OR) {
      bool rv1 = ch_[0]->residue(fc, nextc);
      bool rv2 = ch_[1]->residue(fc, nextc);
      FacOp elim = FALSE, drop = TRUE;
      if (op_ == OR || op_ == SEQ_OR) {
         elim = TRUE; drop = FALSE;
      }

      if (ch_[0]->op_ == elim || ch_[1]->op_ == elim) {
         delete ch_[0];
         delete ch_[1];
         ch_[0]=ch_[1]=nullptr;
         op_ = elim;
      }
      else if (ch_[0]->op_ == drop) {
         MatchCond c(*ch_[1]);
         deepCopy(c);
      }
      else if (ch_[1]->op_ == drop) {
         MatchCond c(*ch_[0]);
         deepCopy(c);
      }
      return (rv1 && rv2);
   }
   else if (var_ != fc.var_ || fc.op_ == ASG)
      return true;
   else if (fc.charOp_)
      return charOpResidue(fc, nextc);
   else {
      assert(fc.op_ >= NOTIN && fc.op_ < ASG); // IN unimplemented yet --- we
      // support only small sets, which are converted into a disjunction of
      // equalities. Because of the way variable names are generated for ASG,
      // two MatchCond with same var_ should be equal, so we won't reach here.
      unsigned long c, c1, c2, c3, c4;
      switch (op_) {
      case EQ:
         switch (fc.op_) {
         case IN:
            if (fc.set_.count(*v_[0]) == 0)
              deepCopy(MatchCond(false));
            break;

         case NOTIN:
            if (fc.set_.count(*v_[0]) == 1)
              deepCopy(MatchCond(false));
            break;

         case EQ: 
            deepCopy(MatchCond(*v_[0] == *fc.v_[0]));
            break;

         case NE:
            if (*fc.v_[0] == *v_[0])
               deepCopy(MatchCond(false));
            break;

         case EQBIT:
            c = v_[0]->ulval();
            c1 = fc.v_[0]->ulval();
            c2 = fc.v_[1]->ulval();
            if ((c & c1)  == c2)
               deepCopy(MatchCond(EQBIT, var_, Value(~c1), Value(c & (~c1))));
            else deepCopy(MatchCond(false));
            break;

         case NEQBIT:
            c = v_[0]->ulval();
            c1 = fc.v_[0]->ulval();
            c2 = fc.v_[1]->ulval();
            if ((c & c1)  == c2)
               deepCopy(MatchCond(false));
            break;

         case LT:
            if (*fc.v_[0] <= *v_[0])
               deepCopy(MatchCond(false));
            break;

         case GT:
            if (*fc.v_[0] >= *v_[0])
               deepCopy(MatchCond(false));
            break;

         default:
            assert(false);
         }
         break;

      case NE:
         switch (fc.op_) {
         case EQ:
            deepCopy(MatchCond((*fc.v_[0] == *v_[0])? TRUE : FALSE));
            break;

         case NE:
            if (*fc.v_[0] == *v_[0])
               deepCopy(MatchCond(true));
            break;

         case EQBIT:
            c = v_[0]->ulval();
            c1 = fc.v_[0]->ulval();
            c2 = fc.v_[1]->ulval();
            if ((c & c1)  == c2)
              deepCopy(MatchCond(NEQBIT, var_, Value(~c1), Value(c & (~c1))));
            else deepCopy(MatchCond(true));
            break;

         case NEQBIT:
            c = v_[0]->ulval();
            c1 = fc.v_[0]->ulval();
            c2 = fc.v_[1]->ulval();
            if ((c & c1)  == c2)
               deepCopy(MatchCond(true));
            break;

         case LT:
            if (*fc.v_[0] <= *v_[0])
               deepCopy(MatchCond(true));
            break;

         case GT:
            if (*fc.v_[0] >= *v_[0])
               deepCopy(MatchCond(true));
            break;

         case NOTIN: // x != v/x \notin V = true if v \in V, else x != v
            if (fc.set_.count(*v_[0]))
               deepCopy(MatchCond(true));
            break;

         default:
            assert(false);
         }
         break;

      case LT:
         switch (fc.op_) {
         case EQ:
            deepCopy(MatchCond(*fc.v_[0] < *v_[0]));
            break;

         case EQBIT:
            if (*fc.v_[1] >= *v_[0])
               deepCopy(MatchCond(false));
            break;

         case LT:
            if (*fc.v_[0] <= *v_[0])
               deepCopy(MatchCond(true));
            break;

         case GT:
            if (*fc.v_[0] >= *v_[0])
               deepCopy(MatchCond(false));
            break;

         default:
            break;
         }
         break;

      case GT:
         switch (fc.op_) {
         case EQ:
            deepCopy(MatchCond(*fc.v_[0] > *v_[0]));
            break;

         case LT:
            if (*fc.v_[0] <= *v_[0])
               deepCopy(MatchCond(false));
            break;

         case GT:
            if (*fc.v_[0] >= *v_[0])
               deepCopy(MatchCond(true));
            break;

         default:
            break;
         }
         break;

      case EQBIT:
         c1 = v_[0]->ulval();
         c2 = v_[1]->ulval();
         switch (fc.op_) {
         case EQ:
            c4 = fc.v_[0]->ulval();
            deepCopy(MatchCond((c1 & c4)  == c2));
            break;

         case EQBIT:
            c3 = fc.v_[0]->ulval();
            c4 = fc.v_[1]->ulval();
            if ((c1 & c4) == (c2 & c3)) {
               if ((c1 & ~c3) == 0)
                  deepCopy(MatchCond(true));
               else deepCopy(
                       MatchCond(EQBIT,var_,Value(c1 & ~c3),Value(c2 & ~c3)));
            }
            else deepCopy(MatchCond(true));
            break;

         case NEQBIT:
            c3 = fc.v_[0]->ulval();
            c4 = fc.v_[1]->ulval();
            if (c1 == c3 && c2 == c4)
               deepCopy(MatchCond(false));
            break;

         case LT:
            if (fc.v_[0]->ulval() >= c2)
               deepCopy(MatchCond(false));
            break;

         default:
            break;
         }
         break;

      case NEQBIT:
         c1 = v_[0]->ulval();
         c2 = v_[1]->ulval();
         switch (fc.op_) {
         case EQ:
            c4 = fc.v_[0]->ulval();
            deepCopy(MatchCond(!((c1 & c4)  == c2)));
            break;

         case EQBIT:
            c3 = fc.v_[0]->ulval();
            c4 = fc.v_[1]->ulval();
            if ((c1 & c4) == (c2 & c3)) {
               if ((c1 & ~c3) == 0)
                  deepCopy(MatchCond(false));
               else deepCopy(
                      MatchCond(NEQBIT,var_,Value(c1 & ~c3),Value(c2 & ~c3)));
            }
            else deepCopy(MatchCond(true));
            break;

         case NEQBIT:
            c3 = fc.v_[0]->ulval();
            c4 = fc.v_[1]->ulval();
            if ((c1 == c3) && (c2 == c4))
               deepCopy(MatchCond(true));
            break;

         default:
            break;
         }
         break;

      default:
         assert(false);
      }
   }
   return true;
}

bool
MatchCond::charOpResidue(const MatchCond& fc, unsigned char& nextc) { 
   // A charOp test is selected only when there is at least one regexp match on
   // this variable. In this case, we limit fc to be an EQ or NOTIN test, so we
   // define residue only for these cases.

   RegExpr* rv=nullptr;
   string sv;

   if (op_ == EQREG || op_ == NEQREG)
      rv = v_[0]->reval();
   else sv = v_[0]->sval();

   switch (fc.op_) {
   case EQ: {
      unsigned char cv = fc.cval();
      switch (op_) {
      case EQ:
         if (sv.size() == 0)
            deepCopy(MatchCond(cv == '\0'));
         else if (sv[0] == cv) { // Convert to regexp in order to continue char
            op_ = EQREG;         // by char comparison.
            RegExpr r(sv.substr(1));
            *v_[0] = Value(&r);
         }
         else deepCopy(MatchCond(false));
         break;

      case NE:
         // We could return false, since there NE will succeed on numerous
         // transitions, leading to explosion. But then, it is inconsistent
         // to support NEQREG while outlawing NE. Let us be consistent: either
         // return false in both cases, or ignore the possibilty of explosion
         // and do our best.
         if (sv.size() == 0 && cv != '\0') 
            deepCopy(MatchCond(true)); // NOTE: x!="" / *x=c is TRUE
         else if (sv[0] == cv) {
            op_ = NEQREG;
            RegExpr r(sv.substr(1));
            *v_[0] = Value(&r); // x!=as / *x=a is ++x!=s
         }
         else deepCopy(MatchCond(true));  // x!=as / *x!=a is TRUE
         // break;
         return false;

      case EQREG:
         rv->residue(cv, nextc);
         if (rv->isPhi())
            deepCopy(MatchCond(false));
         else if (rv->isTrue() || rv->inclEpsilon())
            deepCopy(MatchCond(true));
         break;

      case NEQREG:
         rv->residue(cv, nextc);
         if (rv->isPhi())
            deepCopy(MatchCond(true));
         else if (rv->isTrue() || rv->inclEpsilon()) //@@@@ inclEpsilon: wrong?
            deepCopy(MatchCond(false));
         // break;
         return false;

      case GT:
      // Although we can accurately define residue for GT and LT, typically,
      // they will succeed on half of the many eq transitions out of a node that
      // is designed for regexp matching (i.e., it has a multiway branch on
      // several characters).  So, we should avoid a regexp match when there are
      // GT and LT conditions present. We return false so that the caller can
      // take alternate action (instead of proceeding with the residue), if
      // it chooses to.
      //
         if (sv.size() == 0)
            deepCopy(MatchCond(cv > '\0')); // x > "" / *x = a is true if a > 0
         else if (cv > sv[0])
             deepCopy(MatchCond(true)); // x > as / (*x = b > a) is TRUE
         else if (cv == sv[0])
            *v_[0] = Value(sv.substr(1)); // x>as / *x=a is ++x > s
         else deepCopy(MatchCond(false));  // x>as / (*x = b < a) is FALSE
         return false;

      case LT:
         if (sv.size() == 0)
            deepCopy(MatchCond(false)); // x < "" / *x = a is false for all a.
         else if (cv < sv[0])
            deepCopy(MatchCond(true)); // x < as / (*x = b < a) is TRUE
         else if (cv == sv[0])
            *v_[0] = Value(sv.substr(1)); // x<as / *x=a is ++x < s
         else deepCopy(MatchCond(false));  // x<as / *x=b>a is FALSE
         return false;

      default:
         assert(false);
         break;
      }
      break;
   }
   case NOTIN: {
      auto& setv = fc.set_;
      switch (op_) {
      case EQ:
         if (sv.size() == 0) { // x = "" / x notin s <=> 0 notin s
            if (setv.count(Value::zero()) == 1)
               deepCopy(MatchCond(false));
            else return false; // we cannot advance pointer
         }
         else if (setv.count(Value((long)sv[0]))==1)
            deepCopy(MatchCond(false));
         else assert(false); // Should never happen.
         break;

      case EQREG: {
         bitset<REGEXP_CHMAX> f = rv->first();
         for (unsigned j=0; j < REGEXP_CHMAX; j++)
            if (f.test(j))
               assert(setv.count(j)==1);
         deepCopy(MatchCond(false));
         break;
      }

      case NEQREG: {
         bitset<REGEXP_CHMAX> f = rv->first();
         for (unsigned j=0; j < REGEXP_CHMAX; j++)
            if (f.test(j))
               assert(setv.count(j)==1);
         deepCopy(MatchCond(true));
         break;
      }

      default:
         assert(false);
         break;
      }
      break;
   }
   default:
      assert(false);         
   }
   return true;
}

/*********** Helper functions to pick ("select") branching conditions *********/

Merit<string, Value> MatchCond::
getEqIneqReMerit(FacOp fop, float weight) const {
   Merit<string, Value> rv;
   Value *v = nullptr; // the value collected from this MatchCond
   float det=0, red=0; // determinism, reduction in work

   assert(fop == EQ || fop == GT || fop == EQREG);
   switch(op_) {
   case AND: 
      rv = ch_[0]->getEqIneqReMerit(fop, weight);
      rv.and_trans(ch_[1]->getEqIneqReMerit(fop, weight));
      return rv;

   case OR:
      rv = ch_[0]->getEqIneqReMerit(fop, weight/2);
      rv.or_trans(ch_[1]->getEqIneqReMerit(fop, weight/2));
      return rv;

   case SEQ_AND: 
   case SEQ_OR: 
      // For sequential and/or, we first need to check ch_[0], so tests in
      // ch_[1] don't contribute any thing to usefulness
      rv = ch_[0]->getEqIneqReMerit(fop, weight);
      return rv;

   case EQ: 
      v = v_[0];
      if (fop == GT) {
         // EQ will lead to deterministic branching on a GT test. This is
         // unaffected by the values used in GT comparisons, but we accumulate
         // the EQ value in transInfo any way: this will help choose a median
         // such that the current ruleset (candidates for matches) will get
         // split evenly.
         det = weight;
         red = 0; 
         // On the branch that v= remains viable, the test still needs to be done

         // red = 0.6*weight; 
         // on two of the 3 branches of GT test, eq test fails, and
         // hence need not be repeated in the future.
      }
      else if (fop == EQ) {
         det = red = weight; // We can treat EQ as if it is EQREG with regexp 
         // being simple string. So, EQREG also gives 100% det and red.
      }
      else if (fop == EQREG) {
         det = weight; red = 0.9*weight;
         //det = red = 0.9*weight; // We can treat EQ as EQREG with regexp 
         // being simple string. So, EQREG also gives 100% det and red. But
         // scale by 0.9 to express a preference for EQ over EQREG
      }
      break;

   case NE: 
      // In general, NE will follow multiple branches of EQ, GT, etc.
      // So, we set det to 0. For EQ branches, NE test will not have
      // to be made in the future (except the else branch), so red is high
      det = 0; 
      if (fop == EQ)
         red = weight; // x != v test is eliminated on the viable branch.
         // 0.8*weight; // old
      if (fop == EQREG) {
         // NE can be converted into a regexp match, so the value should be
         // accumulated in transInfo.
         v = v_[0];
         red = 0.9*weight; // 0.7*weight;         
      }
      else if (fop == GT) {
         red = 0.6*weight; // NE test won't need test on 2 of 3 GT branches.
      }
      break;

   case GT:
   case LT:
      if (fop == EQ) {
         // GT and LT will succeed on multiple EQ branches, but
         // probabilistically, we may expect it to fail on half the branches, so
         // det = 0.5. On all except the else branch, GT or LT test won't have
         // to be repeated, so red is high.
         det = 0.5*weight; red = weight; // 0.7*weight;
         // We don't accumulate v_[0] because constant values on EQ branches
         // are not based on any GT test.
      }
      else if (fop == EQREG) {
         // We could convert LT and GT into regexp matches if necessary. Such
         // a choice would be appropriate if there are a large number of string
         // inequalities across the rules. For now, we use the simpler option of
         // using a nondeterministic branching, where one of the branches will
         // match on regexp, and the other on inequalities. 
         det = weight*0.45; red = 0.9*weight; // 0.7*weight;
         // We don't accumulate v_[0] because constant values on EQ branches
         // are not based on any GT test.
      }
      else if (fop == GT) {
         // < n is deterministic branch for < m if m <= n. Taking m <= n prob
         // to be 0.5, det = 0.5. If m <= n then x < m still needs to be
         // performed on the < n branch, i.e. red = 0 in this case. If m > n
         // then <n => <m so the test can be avoided on that branch. So, red
         // is 0.5*0 + 0.5*0.5*1 = 0.25. 

         det = 0.5*weight; red = 0.25*weight;
         // det = 0.8*weight; red = 0.7*weight;
         //On the otherhand, if you have a bunch of 
         // inequalities, the best you can do is to make a GT/LT branch. So.
         // we set det=0.8. Regardless of the value of m, testing < m
         // can be avoided on one of the branches.
         v = v_[0];
      }
      break;

   case EQREG:
   case NEQREG:
      if (fop == EQREG || fop == NEQREG) {
         det = red = weight;
         v = v_[0];
      }
      else { // only EQREG branch can deal with regexp operators.
         det = 0;
         red = -1e10;
      }
      break;

   default:
      // If we haven't explicitly specified det/red for an op then we
      // default to zero. 
      det = 0; red = 0;
      break;
   }

   rv = Merit<string, Value>(var_, weight, det, red, v);
#if CG_DBG >= 3
   cout << "getEqInEqMerit(" << *this << ", " << fop << ", " << weight << ") = ";
   rv.print(cout); cout << endl;
#endif
   return rv;
}

Merit<string, Value> 
getEqIneqReMerit(const vector<const MatchCond*>& pat, MatchCond::FacOp fop) {
   // If we set weight = 1/pat.size, then the base for rv's weight will be 1.0,
   // so the det and use values in transInfo can be used directly without 
   // having to be scaled (which requires weight, a field not within transInfo)
   auto rv = pat[0]->getEqIneqReMerit(fop, 1.0/pat.size());
   rv.upd_remsz();
   for (unsigned j=1; j < pat.size(); j++) {
      auto vm = pat[j]->getEqIneqReMerit(fop, 1.0/pat.size());
      vm.upd_remsz();
      rv.or_trans(vm);
   }
#if CG_DBG >= 3
   cout << "Top level getEqIneqReMerit(fop=" << facOpNm[fop] << ")=\n";
   rv.print(cout); cout << endl;
#endif
   return rv;
}

Merit<MatchCond, bool> MatchCond::
getIfMerit(float weight) const {
   Merit<MatchCond, bool> rv;

   switch(op_) {
   case AND: 
      rv = ch_[0]->getIfMerit(weight);
      rv.and_trans(ch_[1]->getIfMerit(weight));
      break;

   case OR:
      rv = ch_[0]->getIfMerit(weight/2);
      rv.or_trans(ch_[1]->getIfMerit(weight/2));
      break;

   case SEQ_AND: 
   case SEQ_OR: 
      // For sequential and/or, we first need to check ch_[0], so tests in
      // ch_[1] don't contribute any thing to usefulness
      rv = ch_[0]->getIfMerit(weight);
      break;

   case EQREG:
   case NEQREG:
      rv = Merit<MatchCond, bool>(*this, weight, 0, 0, false);
      break;

   default:
      rv = Merit<MatchCond, bool>(*this, weight, weight, weight, true);
      break;
   }

   return rv;
}

Merit<MatchCond, bool> 
getIfMerit(const vector<const MatchCond*>& pat) {
   // If we set weight = 1/pat.size, then the base for rv's weight will be 1.0,
   // so the det and use values in transInfo can be used directly without 
   // having to be scaled (which requires weight, a field not within transInfo)
   auto rv = pat[0]->getIfMerit(1.0/pat.size());
   rv.upd_remsz();
   for (unsigned j=1; j < pat.size(); j++) {
      auto vm = pat[j]->getIfMerit(1.0/pat.size());
      vm.upd_remsz();
      rv.or_trans(vm);
   }
#if CG_DBG >= 3
   cout << "Top level getIfMerit\n";
   rv.print(cout); cout << endl;
#endif
   return rv;
}

#if false
/* Order first on the basis of determinism, then on the basis of work reduction
   and then by branching factor, and then by type of variable */
bool operator<(const Merit<string, Value>::TransInfo& m1, 
               const Merit<string, Value>::TransInfo& m2) {
#if Merit == OldMerit
#define w(m) (m.remsz() + (m.elems.size() == 0? 0 : \
                (0.22/m.elems.size() + ((m.elems[0].isIntegral())? 0.03 : 0.0))))
#else
#define w(m) (m.remsz() * m.
#endif
   return (w(m1) < w(m2));
#undef w
/*
   return (m1.det < m2.det ||
             (m1.det == m2.det && 
                (m1.red < m2.red ||
                   (m1.red == m2.red && 
                      (m1.elems.size() > m2.elems.size() ||
                         (m1.elems.size() == m2.elems.size() && 
                            (m1.elems.size() != 0 && 
                               (!m1.elems[0].isIntegral() &&
                                   m2.elems[0].isIntegral()))))))));
*/
}

bool operator<(const Merit<MatchCond, bool>::TransInfo& m1, 
               const Merit<MatchCond, bool>::TransInfo& m2) {
   return ((m1.det < m2.det) || ((m1.det == m2.det)&&(m1.remsz() < m2.remsz())));
}

template <class K, class E>
bool operator>(const typename Merit<K, E>::TransInfo& m1, 
               const typename Merit<K, E>::TransInfo& m2) { return m2 < m1; };
#endif

#ifdef OLDM
#define rem(x, y) 1/((x).util() + ((x).elems.size() == 0? 0 : \
           (0.22/(x).elems.size())) + 0.03)
#else
#define rem(x, nbr) ((x).remsz+0.01)*(nbr+0.01)
#endif

template <class K, class E>
tuple<K, float, const typename Merit<K, E>::TransInfo*> 
getMaxTi(const Merit<K, E>& in, MatchCond::FacOp fop) {
   const typename Merit<K, E>::TransInfo *rvti = nullptr; 
   unsigned maxnbr=1000;
   K rvk; float rvremsz = 1e10;
   for (auto& it: in.trans()) {
      unsigned nbr = it.second.elems.size();
      if (nbr == 0) continue;
      switch (fop) {
      case MatchCond::EQ: 
      case MatchCond::EQREG: 
         nbr--; break;
      case MatchCond::LT:
      case MatchCond::GT: 
         nbr = 2; break;
      case MatchCond::IF:
         nbr = 1; break;
      default:
         assert(false); break;
      }

      if (rvti == nullptr || rem(*rvti, maxnbr) > rem(it.second, nbr)) {
         rvk = it.first;
         rvti = &it.second;
         rvremsz = rem(it.second, nbr);
         maxnbr = nbr;
      }
   }
#if CG_DBG >= 2
   cout << "getMaxTi(" << rvk << " " << facOpNm[fop] << ") rvremsz=" << rvremsz;
   if (rvti) {
      cout << " det=" << rvti->det << " ";
      for (const auto& v: rvti->elems) cout << v << " ";
   }
   cout << endl;
#endif
   return make_tuple(rvk, rvremsz, rvti);
}

struct BranchInfo {
   bool exhaustive; // false => branches[branches.size()] can be a compound test
   string var;
   float det;
   float remsz;
   MatchCond::FacOp maxOp;
   vector<Value> vals;
   vector<MatchCond> branches;

   BranchInfo() {
      exhaustive = false;
      det  = 0.0;
      remsz = 1e10;
      maxOp = MatchCond::INVALID;
   }
};

struct MatchRule {
   int ruleNum;
   MatchCond pat;

   MatchRule(int sq=0): ruleNum(sq) {};

   MatchRule(int sq, const MatchCond& c): ruleNum(sq), pat(c) {};

   bool operator==(const MatchRule& mr) const {
      return (ruleNum == mr.ruleNum && pat == mr.pat);
   }

   bool operator<(MatchRule& mr) const {
      if (ruleNum == mr.ruleNum)
         return (pat < mr.pat);
      else return (ruleNum < mr.ruleNum);
   }
};

namespace std {
template <> struct hash<MatchRule> {
 public :
   size_t operator()(const MatchRule& m) const { 
      return m.pat.hash(m.ruleNum); 
   }
};
};

ostream& operator<<(ostream& os, const MatchRule& m) {
   os << m.ruleNum << ": " << m.pat; return os;
}

struct BranchInfo
findBranches(const vector<MatchRule>& rules) {
   // We want high determinism (i.e., det() == 1.0). Among the tests that have
   // the highest determinism, we pick one with highest red(). Among those, we
   // go for branches with minimum degree. Among those, we prefer branches on
   // integer type values. (This preference is captured in the operator<
   // function on Merit.)
   //
   // We could pick every possible test, evaluate its Merit best based on the
   // above criteria, and then select the very best. But this may take a lot of
   // time. For deterministic tests, it is unnecessary to pick the absolute
   // best, as all deterministic tests lead to almost the same code size and
   // matching time. So, we can speed up test selection by stopping as soon as
   // we find a test with determinism == 1. In reality, what we do is that we
   // look for different types of tests in the following order. Within each test
   // type, all possible tests are tried. If one test type gives us det() == 1
   // we stop immediately. If we don't get any with determinism == 1 then we end
   // up cycling through all possible tests.
   //
   // 1. Equality transition is the most preferred transition type. It cannot
   //    deal with EQREG. GT/LT are handled but less effectively.
   // 2. Next, we try inequality transition. Inequality transition can
   //    deal with all operators except EQREG. (NE handling is not effective.)
   // 3. Next, we should try EQBIT, but we have deferred that implementation.
   //    (But they should get handled as IF.)
   // 4. Next, we try a branch on IF
   // 5. Next, we go for a regexp transition.
   //

   struct BranchInfo rv;

   tuple<string, float, const Merit<string, Value>::TransInfo*> 
      maxti, maxeqti, maxgtti, maxltti, maxreti;
   tuple<MatchCond, float,
        const Merit<MatchCond, bool>::TransInfo*> maxifti;

   vector<const MatchCond*> pat(rules.size()); unsigned j=0;
   for (auto& rs: rules)
      pat[j++] = &rs.pat;

#if CG_DBG >= 3
   cout << "findBranches called on\n";
   for (auto& p: pat)
      cout << *p << endl;
#endif
   {
      // Look for an equality transition
      const auto& t1 = getEqIneqReMerit(pat, MatchCond::EQ); 
      maxeqti = getMaxTi(t1, MatchCond::EQ);

      if (get<2>(maxeqti)) {
         maxti = maxeqti;
         rv.exhaustive = ((get<2>(maxeqti)->det - 1) > -1e-12);
         rv.var = get<0>(maxeqti);
         rv.det = get<2>(maxeqti)->det;
         rv.remsz = get<1>(maxeqti);
         rv.maxOp = MatchCond::EQ;
         rv.vals.assign(get<2>(maxti)->elems.begin(),get<2>(maxti)->elems.end());
         if (get<2>(maxti)->det > 0.99)
            goto done;
      }

      // Look for a GT transition. We thought LT and GT should be considered
      // separately because the use of a median across them may result in a bad
      // branch point choice. For instance, consider >1, >2,...,>8 and <12,
      // <13,...,<20. The median-based branch will be (<,=,>)10, but all of
      // these 17 patterns will remain in all three child states. However, it
      // can be seen that on the LT branch, the <12 ... <20 tests will be
      // eliminated, so the next choice, if it is made on the same variable,
      // will only be based on the values {1..8}, so this bad choice won't
      // further propagate. A similar argument can be made for the other two
      // branches. So, although it is possible to make a bad choice one time, it
      // remains a one-time problen. Moreover, the likelihood of this problem is
      // also low because it requires *all* LT values to be greater than *all*
      // GT values. Instead, it is more likely that the median of the combined
      // set will be close to the middle of LT as well as GT sets.

      const auto& t21 = getEqIneqReMerit(pat, MatchCond::GT);
      maxgtti = getMaxTi(t21, MatchCond::GT);

      if (get<2>(maxgtti)) {
         if (!get<2>(maxti) || get<1>(maxti) > get<1>(maxgtti)) {
            maxti = maxgtti;
            rv.exhaustive = true;
            rv.var = get<0>(maxti);
            rv.det = get<2>(maxgtti)->det;
            rv.remsz = get<1>(maxgtti);
            rv.maxOp = MatchCond::GT;
            rv.vals.assign(get<2>(maxti)->elems.begin(),
                           get<2>(maxti)->elems.end());
            if (get<2>(maxti)->det > 0.999)
               goto done;
         }
      }

      // Look for an if transition
      const auto& t3 = getIfMerit(pat);
      maxifti = getMaxTi(t3, MatchCond::IF);
      if (get<2>(maxifti) && get<0>(maxifti).op() != MatchCond::EQREG
            && get<0>(maxifti).op() != MatchCond::NEQREG 
            && (!get<2>(maxti) || get<1>(maxti) > get<1>(maxifti))) {
         rv.exhaustive = true;
         rv.var = get<0>(maxifti).var();
         rv.det = get<2>(maxifti)->det;
         rv.remsz = get<1>(maxifti);
         rv.maxOp = MatchCond::IF;
         if (get<2>(maxifti)->det > 0.999)
            goto done;
      }
      else get<1>(maxifti) = 1e12;

      bool hasregexp = false;
      for (const auto& p: pat)
         if ((hasregexp = p->hasRegexpOp())) break;
      if (!hasregexp) goto done;

      // Look for a regexp transition
      const auto& t4 = getEqIneqReMerit(pat, MatchCond::EQREG);
      maxreti = getMaxTi(t4, MatchCond::EQREG);
      if (get<2>(maxreti) &&
          ((rv.maxOp==MatchCond::IF && get<1>(maxifti) > get<1>(maxreti))
           || (rv.maxOp!=MatchCond::IF && 
               (!get<2>(maxti) || get<1>(maxti) > get<1>(maxreti))))) {
         maxti = maxreti;
         rv.var = get<0>(maxreti);
         rv.det = get<2>(maxreti)->det;
         rv.remsz = get<1>(maxreti);
         rv.exhaustive = (rv.det > 0.999);// 1e-12?
         rv.maxOp = MatchCond::EQREG;
         rv.vals.assign(get<2>(maxti)->elems.begin(),get<2>(maxti)->elems.end());
         if (get<2>(maxreti)->det > 0.999)
            goto done;
      }
   }

 done:
   if (rv.maxOp == MatchCond::EQ) {
      //sort(rv.vals.begin(), rv.vals.begin()+rv.vals.size());
      //auto it = unique(rv.vals.begin(), rv.vals.begin()+rv.vals.size());
      //rv.vals.resize(distance(rv.vals.begin(), it));
      for (const auto& val: rv.vals) 
         rv.branches.emplace_back(MatchCond::EQ, rv.var, val);
      if (!rv.exhaustive)
         rv.branches.emplace_back(MatchCond::NOTIN, rv.var, 
                   set<Value>(rv.vals.begin(), rv.vals.begin()+rv.vals.size()));
   }
   else if (rv.maxOp == MatchCond::GT || rv.maxOp == MatchCond::LT) {
      //sort(rv.vals.begin(), rv.vals.begin()+rv.vals.size());
      Value median = rv.vals[(rv.vals.size()-1)/2];
      rv.branches.emplace_back(MatchCond::LT, rv.var, median);
      rv.branches.emplace_back(MatchCond::EQ, rv.var, median);
      rv.branches.emplace_back(MatchCond::GT, rv.var, median);
   }
   else if (rv.maxOp == MatchCond::IF) {
      // NOTE: rv.vals is not set (not applicable for IF)
      MatchCond mc_then = get<0>(maxifti);
      MatchCond mc_else(mc_then);
      if (mc_else.negate()) {
         rv.branches.push_back(mc_then);
         rv.branches.push_back(mc_else);
      }
      else if (mc_then.op() == MatchCond::ASG) {
         rv.maxOp = MatchCond::ASG;
         rv.branches.push_back(mc_then);
      }
      else {
         // Other than ASG, at present, only GT/LT cannot be negated.
         string var = mc_then.var();
         const Value& val = mc_then.val();
         assert(mc_then.op()==MatchCond::GT || mc_then.op()==MatchCond::LT);
         rv.maxOp = MatchCond::GT;
         rv.branches.emplace_back(MatchCond::LT, var, val);
         rv.branches.emplace_back(MatchCond::EQ, var, val);
         rv.branches.emplace_back(MatchCond::GT, var, val);
      }
      return rv;
   }
   else if (rv.maxOp == MatchCond::EQREG) {
      // Note that NEQREG values will also be in the list of elems
      //sort(rv.vals.begin(), rv.vals.begin()+rv.vals.size());
      //auto it = unique(rv.vals.begin(), rv.vals.begin()+rv.vals.size());
      //rv.vals.resize(distance(rv.vals.begin(), it));

      bitset<REGEXP_CHMAX> firsts;
      for (auto& val: rv.vals) {
         if (val.isString()) {
            const string& s = val.sval();
            if (s.size() == 0)
               firsts.set(0);
            else firsts.set(s[0]);
         }
         else if (val.isRegExpr())
            firsts |= val.reval()->first();
      }

      //set<Value> elseSet;
      bool hasElse = !rv.exhaustive;
      for (const auto& p: pat)
         hasElse |= p->opPresent((1<<MatchCond::NE)|(1<<MatchCond::NEQREG), 
                                 rv.var);
      rv.vals.clear();
      for (unsigned j=0; j < REGEXP_CHMAX; j++) {
         if (firsts.test(j)) {
            rv.branches.emplace_back(MatchCond::EQ, rv.var, (char)j);
            rv.vals.push_back(Value(j));
         }
         //else if (hasElse)
            // elseSet.insert(Value(j));
      }
      if (hasElse)
         rv.branches.emplace_back(MatchCond::NOTIN, rv.var, 
                         set<Value>(rv.vals.begin(), rv.vals.end()), true);
      else rv.exhaustive = true;
   }
#if CG_DBG >= 1
   cout<< "findBranches: " << rv.var << " " << facOpNm[rv.maxOp] << " nbr=" 
       << rv.branches.size()<< " det=" << rv.det<< " remsz=" << rv.remsz << endl;
#endif
   assert (rv.maxOp != MatchCond::INVALID);
   return rv;
}

/***************** Helper function to generate branching code *****************/

struct MatchStats {
   int64_t cur_sz, tot_sz;
   int64_t cur_dups, tot_dups;
   int64_t n_nodes, n_ndnodes, n_tests, n_ranges;
   int64_t residues, residue_saves;
   int64_t residueFalseND;
   int64_t proactiveND;
   int64_t polyND;
} mstats;

#if CG_DBG >= 2
unordered_map<unsigned, stringstream> rangeInfo;
#endif

void prtMatchStats(ostream& os, string prefix) { 
   os << "// " << prefix << " Rule size: " << mstats.tot_sz 
      << ", states: " << mstats.n_nodes; 
   if (mstats.n_ndnodes+mstats.residueFalseND > 0)
      os << " (ND states=" << mstats.n_ndnodes+mstats.residueFalseND << ")";
   os << " tests: " << mstats.n_tests;
   int excess = (100*(mstats.n_ranges-mstats.n_tests))/(mstats.n_tests+.01);
   if (excess >= 10 || excess <= -10)
      os << " (ranges=" << (excess > 0? "+" : "") << excess << "%)";
   os << " residues=" 
      << (int)(0.5+100/(1 + (mstats.residue_saves+.01)/(mstats.residues+.01))) 
      << "%\n";
   if (mstats.residueFalseND > 0)
      os << "// ND branches created due to residue returning false: " 
         << mstats.residueFalseND << endl;
   if (mstats.n_ndnodes > 0) 
      os << "// Rules made non-deterministic by proactive ND: " 
         << mstats.proactiveND << ", by polynomial bound: " 
         << mstats.polyND << endl;
}

#define beginblk os << " {"; indent += STEP_INDENT; prtln(os, indent)
#define prtif(c) \
   os << "if ("; mstats.n_tests++; \
   c.print(os, indent);    \
   os << ") "; beginblk
#define prtnif(c) \
   os << "if (!("; mstats.n_tests++; \
   c.print(os, indent);    \
   os << "))"; beginblk
#define prtelse os << "else "
#define endblk \
   indent -= STEP_INDENT;                 \
   prtRewind(os, STEP_INDENT, indent);    \
   os << "}"; prtln(os, indent)

#define gencode(j, incflag) \
   if (incflag) os << "_c" << br.var << "++"; endln(os, indent); \
   if (chIdx[j] != j || chLbl[j] <= baselbl) { \
      os << "goto Lms" << chLbl[(chIdx[j] < chIdx.size()? chIdx[j] : j)]; \
      endln(os, indent); \
   } \
   else { \
      os << (isDet[j]? "/*d*/" : "/*nd*/") << "Lms" << chLbl[j] << ":"; prtln(os, indent); \
      /* For regexp matching, pointer to string arg needs to be saved and
         restored if there is a non-det branch */ \
      if (br.maxOp == MatchCond::EQREG) { \
         if (!isDet[j]) { \
            beginblk; \
            os << "const unsigned char* savedCh = _c" << br.var; \
            endln(os, indent); \
         } \
      } \
      codeGen(chRule[j], chLbl[j], os, indent, nfl); \
      if (br.maxOp == MatchCond::EQREG && !isDet[j]) { \
         os << "_c" << br.var << " = savedCh"; \
         endln(os, indent); \
         endblk; \
      } \
   }

#define prtcont(j) { \
      os << "goto " << (isDet[j]? "Ld" : "Lnd") << mylbl; \
      endln(os, indent);           \
   }

#define prtelsebr \
   if (elsePresent) { \
      os << /*"else */ "goto Lelse" << mylbl; endln(os, indent); \
   }

static void 
codeGen(vector<MatchRule>& rules, unsigned mylbl, 
        ostream& os, int indent, CGFlags& nfl);
void
genIfBranch(const BranchInfo& br, vector<vector<MatchRule>>& chRule, 
            const vector<bool>& isDet, const vector<unsigned>& chLbl,
            const vector<unsigned>& chIdx, unsigned baselbl, unsigned mylbl,
            ostream& os, int indent, CGFlags& nfl) {
   if (br.maxOp == MatchCond::ASG) {
      const MatchCond& mc = br.branches[0];
      mc.print(os, indent); endln(os, indent);
      const ExprNode* e = mc.expr();
      if (Type::isString(e->type()->tag())) {
         os << "_c" << mc.var() 
            << " = (const unsigned char*)" << mc.var() << ".data()"; 
         endln(os, indent);
      }
      gencode(0, 0);
      return;
   }
   prtif(br.branches[0]);
   gencode(0, 0);
   prtcont(0);
   endblk;

   prtelse;
   if (br.maxOp == MatchCond::GT || br.maxOp == MatchCond::LT) {
      prtif(br.branches[1]);
   }
   else { beginblk; }

   gencode(1, 0); prtcont(1); endblk;

   if (br.maxOp == MatchCond::GT || br.maxOp == MatchCond::LT) {
      prtelse; beginblk;
      gencode(2, 0); prtcont(2); endblk;
   }
}

/***************************** Branching Code ******************************/
//
// @@@@ Ideally, a range will first be divided up into subranges that can be
// efficiently dispatched using jump tables. We could formulate this as an
// optimization problem in terms of runtime, or space. 
//
// Time-based formulation is a bit tricky because we don't have a good handle
// on the relative costs of an if-then-else binary tree versus a jump table.
// Nevertheless, let us go ahead and construct a strawman. Dispatching a
// range m with n subranges and k targets using a binary tree of if-then-else
// requires log(2n) time. Note that the time cost is independent of n and k.
// For a jump table, we may set the constant as fixed, say, 4. This choice
// means that for n > 8, jump table will be used. Of course, this choice
// isn't meaningful if the space cost of the jump table is too high. So, the
// only sensible options are to (a) optimize purely on the basis of code
// size, or (b) optimize for a combination of code and space. We could opt
// for some linear combination of these two costs.
//
// Memory-based formulation is a bit easier because memory costs can be more
// accurately estimated. Again, for a range m with n subranges and k targets,
// let us say that a balanced binary tree requires 8*2n space. (8 bytes of
// instructions per branch.) For a jump table approach, we need, say, 16
// bytes of code to dispatch using the jump table and m*log(k) bits for the
// jump table, for a total of 16+m*log(k)/8 bytes. 
//
// A hybrid approach will first use if-then-else to narrow down n, m and k
// until jump tables become a better option over if-then-else. Note that
// there will be cases where if-then-else remains the choice all through, or
// within some subtrees. It looks like this problem can be formulated as a
// dynamic programming problem. Let us start by writing down the optimal cost
// C as a function of the range of values [i,j]. The number of subranges and
// the number of distinct branches --- n and k in the discussion above ---
// are still significant, but they are not explicitly included in the
// recurrences in order to avoid clutter.
//
//         C[i,j] = min(J[i, j], min_(i <= l <= j) (C[i,l] + c + C[l+1,j])
//
// The equation says that we either use a jump table for the entire range,
// thereby incurring the cost J[i,j], or we have an if (x <= l) test, and
// then generate optimal code on the two branches. A fixed cost of c is
// assigned for the (x <= l), with c reflecting some linear combination of
// the code size and runtime costs. Note that C[r,s]=0 whenever r and s are
// in the same subrange: a single subrange has exactly only action, which can
// be launched as soon as you know you are in that range. For this reason,
// there is no reason to consider any point that is in the middle of a
// subrange, and hence the complexity of the algorithm for computing the
// optimal solution will be in terms of n.
//
// The optimal solution C[i, j] can record both the cost and the split point
// l. For a dynamic programming solution, we compute C[i, i+l] in a nested
// loop, iterating over l in the outer loop and i in the inner loop. Thus,
// the dynamic programming algorithm is O(n^2).
//
// We will leave the implementation of this for future. It will be
// interesting to implement and experiment with this dynamic programming
// algorithm for optimal jump table construction, and also compare it with a
// full hash table based solution.
//

pair<unsigned, unsigned>
findGap(const vector<tuple<Value, Value, unsigned>>& range, bool chkLeft, 
        bool chkRight, unsigned l, unsigned r, unsigned stop_at) {
   // if stop_at == 0 then returns the number of gaps found. Otherwise,
   // returns the branch that will divide at the gap number specified by ngaps.
   // (We count gaps starting at 1, so 0th gap does not exist.) Note that the
   // branch needs two components for full specification. The first specifies
   // the index i in the range array, while the second indicates whether we are
   // are interested in the rhs end of this range (1, 1) or the left end (i, 0)

   unsigned ngaps = 0;
   if (chkLeft && get<0>(range[l]) != get<1>(range[l])) {
      ngaps++;
      if (stop_at == ngaps) return make_pair(l, 0);
   }
   for (unsigned j=l+1; j <= r; j++) {
      const Value& l1 = get<1>(range[j-1]);
      const Value& r0 = get<0>(range[j]);
      if (!r0.isIntegral() || l1.lval()+1 != r0.lval()) {
         ngaps++;
         if (stop_at == ngaps) return make_pair(j-1, 1);
      }

      // if (r0 != r1) {
      ngaps++;
      if (stop_at == ngaps) return make_pair(j, 0);
      //}
   }
   if (r > l && chkRight && (get<0>(range[r]) != get<1>(range[r]))) {
      ngaps++;
      if (stop_at == ngaps) return make_pair(r, 1);
   }
   return make_pair(ngaps, 2); 
}

void       
genRangeCode(const vector<tuple<Value, Value, unsigned>>& range, bool chkLeft, 
      bool chkRight, bool elsePresent, unsigned l, unsigned r, 
      const BranchInfo& br, vector<vector<MatchRule>>& chRule, 
      const vector<bool>& isDet, const vector<unsigned>& chLbl, 
      const vector<unsigned>& chIdx, unsigned baselbl, unsigned mylbl, 
      ostream& os, int indent, CGFlags& nfl) {

   bool charop = (br.maxOp == MatchCond::EQREG);
   if (charop) {
      if (get<0>(range[l]).lval() == 0) chkLeft = false;
      if (get<1>(range[r]).lval() == REGEXP_CHMAX-1) chkRight = false;
   }

   if (l+1 >= r) { // Leaf case, no recursion here
      // Generate a branch for range[l]. Many cases to consider here.
      if (get<0>(range[l]) == get<1>(range[l]))
         { prtif(MatchCond(MatchCond::EQ, br.var, get<0>(range[l]), charop)); }
      else if (!chkLeft)
         { prtnif(MatchCond(MatchCond::GT, br.var, get<1>(range[l]), charop)); }
      else if (l==r && !chkRight)
         { prtnif(MatchCond(MatchCond::LT, br.var, get<0>(range[l]), charop)); }
      else {
         prtnif(MatchCond(MatchCond::OR, 
                new MatchCond(MatchCond::LT, br.var, get<0>(range[l]), charop),
                new MatchCond(MatchCond::GT, br.var, get<1>(range[l]), charop)));
      }
      // Now, generate the code for range[l]
      unsigned j = get<2>(range[l]);
      gencode(j, charop); prtcont(j); endblk;

      // Generate the code for range[l]. This applies only if r > l
      if (l < r) {
         prtelse;
         if (get<0>(range[r]) == get<1>(range[r])) {
            prtif(MatchCond(MatchCond::EQ, br.var, get<0>(range[r]), charop));
         }
         else if (!chkRight) {
            prtnif(MatchCond(MatchCond::LT, br.var, get<0>(range[r]),charop));
         }
         else {
            prtnif(MatchCond(MatchCond::OR, 
                new MatchCond(MatchCond::LT, br.var, get<0>(range[r]), charop),
                new MatchCond(MatchCond::GT, br.var, get<1>(range[r]), charop)));
         }

         j = get<2>(range[r]);
         gencode(j, charop); prtcont(j); endblk;
      }
      prtelsebr;
   }
   else {
#define label(a) ((chIdx[a] < br.branches.size()+1)? chLbl[chIdx[a]] : chLbl[a])
#define samestate(a, b) (label(a) == label(b))

      if (chkLeft+chkRight+(r-l)+1 < 10) {
         unsigned elsebr = 1<<30; int nbr = chkLeft+chkRight;
         for (unsigned j=l; j <= r && nbr <= 4; j++) {
            if (get<0>(range[j]) == get<1>(range[j]))
               nbr++;
            else if (elsebr >= (1<<30))
               elsebr = j;
            else {
               unsigned a = get<2>(range[elsebr]);
               unsigned b = get<2>(range[j]);
               if (!samestate(a, b))
                   nbr = 100;
            }
         }
         if (nbr <= 4) {
            nbr -= chkRight + chkLeft;
            for (unsigned j=l; j <= r; j++, nbr--) {
               Value k; 
               if ((k = get<0>(range[j])) == get<1>(range[j])) {
                  prtif(MatchCond(MatchCond::EQ, br.var, k, charop));
                  gencode(get<2>(range[j]), charop); 
                  prtcont(get<2>(range[j])); endblk;
                  //if (nbr > 1 || elsebr < (1<<30)) {
                  //   prtelse; 
                  //}
               }
            }
            if (elsebr < (1<<30)) {
               MatchCond *lc, *rc;
               if (chkLeft)
                 lc = new MatchCond(MatchCond::OR,
                   new MatchCond(MatchCond::EQ,br.var,get<0>(range[l]),charop),
                   new MatchCond(MatchCond::GT,br.var,get<0>(range[l]),charop));
               else lc = new MatchCond(true);
               if (chkRight)
                 rc = new MatchCond(MatchCond::OR,
                   new MatchCond(MatchCond::EQ,br.var,get<1>(range[r]),charop),
                   new MatchCond(MatchCond::LT,br.var,get<1>(range[r]),charop));
               else rc = new MatchCond(true);
               if (chkLeft || chkRight)
                  { prtif(MatchCond(MatchCond::AND, lc, rc)); }
               else {
                  beginblk;
                  delete lc; delete rc;
               }
               gencode(get<2>(range[elsebr]), charop); 
               prtcont(get<2>(range[elsebr])); endblk;               
            }
            prtelsebr;
            return;
         }
      }

      // Invariant: r >= l+2
      unsigned mid = (l+r)/2;

#if CG_DBG >= 2
      rangeInfo[mylbl] << "(<=" << get<1>(range[mid]) << ' ';
#endif

      // divide at the right end (<1>) of range specified by mid
      prtnif(MatchCond(MatchCond::GT, br.var, get<1>(range[mid]), charop));
      genRangeCode(range, chkLeft, false, elsePresent, l, mid, br, chRule,
                   isDet, chLbl, chIdx, baselbl, mylbl, os, indent, nfl);
      endblk;
      prtelse; beginblk;
      bool chklt = (!get<0>(range[mid]).isIntegral() || 
                    get<1>(range[mid]).lval()+1!=get<0>(range[mid+1]).lval());
      genRangeCode(range, chklt, chkRight, elsePresent, mid+1, r, br, chRule,
                   isDet, chLbl, chIdx, baselbl, mylbl, os, indent, nfl);
      endblk;

#if CG_DBG >= 2
      rangeInfo[mylbl] << ") ";
#endif
   }
}

void
doGenCode(const BranchInfo& br, vector<vector<MatchRule>>& chRule, 
      const vector<bool>& isDet, const vector<unsigned>& chLbl, 
      const vector<unsigned>& chIdx, unsigned baselbl, unsigned& mylbl,
      unsigned nbr, bool ndbranch,
      ostream& os, int indent, CGFlags& nfl) {

   bool elsePresent = !br.exhaustive;

#if CG_DBG >= 2
   cout << "doGenCode(nbr=" << nbr << ", type=" << facOpNm[br.maxOp] 
        << ", label=" << mylbl;
   for (unsigned j=0; j < isDet.size(); j++)
      if (!isDet[j]) cout << "nd[" << j << "] ";
   cout << "\nchIdx[k] <> k: ";
   for (unsigned k=0; k < nbr; k++) 
      if (chIdx[k] != k) {
         cout << k << "(";
         if (k < br.vals.size()) 
            cout << br.vals[k];
         else if (k == br.vals.size())
            cout << "else";
         else if (k < nbr) cout << "nd";
         cout << ")->" << chIdx[k] << "(";
         if (chIdx[k] < br.vals.size())
            cout << br.vals[chIdx[k]];
         else cout << "Lms" <<  chLbl[k];
         cout << ") ";
      }
   cout << endl;
   for (unsigned j=0; j < br.branches.size(); j++) 
      cout << br.branches[j] << ": " << chRule[j] << endl;
   if (br.branches.size() < nbr) cout << "ND: " << chRule[nbr-1] << endl;
#endif

   if (br.maxOp == MatchCond::IF || br.maxOp == MatchCond::GT 
       || br.maxOp == MatchCond::LT || br.maxOp == MatchCond::ASG)
      genIfBranch(br, chRule, isDet, chLbl, chIdx, 
                  baselbl, mylbl, os, indent, nfl);
   else {
      assert(br.maxOp == MatchCond::EQ || br.maxOp == MatchCond::EQREG);

      vector<tuple<Value, Value, unsigned>> range; 
      Value prevVal;
      unsigned nextr=0; 
      // for EQ and EQREG, br.vals will have one fewer entry than br.branches
      // if the last branch is an else-branch
      for (unsigned j=0; j < br.vals.size(); j++) {
         const Value& v = br.vals[j];
         if ((j > 0) && (((chIdx[j] != j-1 && chIdx[j] != chIdx[j-1]) || 
                          (chIdx[j] > nbr && chLbl[j] != chLbl[j-1])) ||
                         (!v.isIntegral() || prevVal.lval()+1 != v.lval()))) {
            range.emplace_back(br.vals[nextr], prevVal, nextr);
            nextr = j;
         }
         prevVal = v;
      }
      range.emplace_back(br.vals[nextr], br.vals[br.vals.size()-1], nextr);
      mstats.n_ranges += range.size();
#if CG_DBG >= 2
      rangeInfo[mylbl] << "Ranges: ";
      for (auto& r: range)
       rangeInfo[mylbl] << get<0>(r) << '-' << get<1>(r) << ':' <<
        ((chIdx[get<2>(r)]< nbr)? chLbl[chIdx[get<2>(r)]]:chLbl[get<2>(r)])<<' ';
#endif
      genRangeCode(range, true, true, elsePresent, 0, range.size()-1, br, 
                   chRule, isDet, chLbl, chIdx, baselbl, mylbl, os, indent, nfl);
   }

   if (elsePresent) {
      os << "Lelse" << mylbl/*chLbl[chIdx[nbr-1]]*/ << ":"; prtln(os, indent);
      unsigned i = nbr-1-(ndbranch);
      gencode(i, 0);
      prtcont(i);
   }

   if (ndbranch) {
      os << "Lnd" << mylbl << ":"; // prtln(os, indent);
      gencode(nbr-1, 0);
   }
   os << "Ld" << mylbl << ":"; //prtln(os, indent);
}

#undef prtif
#undef prtnif
#undef beginblk
#undef endblk
#undef prtcont
#undef prtelsebr
#undef prtelse
#undef gencode

/****************** Proceed to construct branches and recurse *****************
 * @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * Regular expression matching is much more expensive in terms of automaton
 * size. We can do two things to tackle this. First, we can share regexp
 * matching code across events. This requires parameterization of matching code,
 * as the the variables (and reactions) involved will be different across
 * events. It will also likely require regexp matching to be done at the
 * beginning or the end, i.e., before or after all other tests. "After" makes
 * sense in that it avoids grouping regexps that don't have to be matched
 * together, which is better in terms of space. "Before" makes sense because it
 * allows regexp matching results to be convered into a simple x=c conditions,
 * which can be effectively factorized in the second stage. 
 *
 * Alternatively, we should devise regexp matching techniques that is much
 * stricter with space, opting to use deterministic branching ONLY when it is
 * extremely advantageous. It will also require the code to use NFA emulation by
 * keeping track of all the states the automaton could be in. The current
 * approach of ND-transitions may not be powerful enough.
 * @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * The current approach of using first and residue is not the best approach for
 * automata construction, as we spend a lot of time in identifying ranges where
 * the target state is the same. This is particularly true for regexp, where we
 * first generate a branch for i=1 to 255 and then merge equivalent ones. A
 * better approach is to define an alternate operation NEXT that returns
 *
 *      {(range1, residue1), (range2, residue2), ..., (rangeN, residueN)}, 
 *
 * where the ranges are in sorted order. Now the NEXTs of two MatchRules can be
 * combined using a merge-like operation. For instance:
 *
 *      {(1-10, r1), (11-30, r2)] and [(1-20, r3), (21-30, r4)} 
 *
 * will yield
 *
 *       {(1-10, [r1,r3]), (11-20, [r2,r3]), (21-30, [r2,r4])}
 *
 * At least for regexp, NEXT can likely be implemented a bit more efficiently 
 * as compared to first and residue since it combines the two operations. More
 * dramatic speed increase will come from the fact that we don't iterate 
 * explicitly through i=1 to 255, thereby calling residue 255 times. 
 *
 * @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * We should think if this approach yields benefits for non-regexp branches.
 * If so, does that suggest a different perspective at the next higher level,
 * as opposed to the way we consider different types of branches now? Is there a
 * more general way to unify these branch types and thereby reduce the level of
 * ad-hocness and heuristics in branch selection? 
 ******************************************************************************
 ****************** Proceed to construct branches and recurse *****************/

static unordered_map<vector<MatchRule>, int> matchTab;
static int lbl=1000;

static unsigned 
getsize(const vector<MatchRule>& rs) {
   unsigned rv=0;
   for (const auto& r: rs) rv += r.pat.size();
   return rv;
}

/*
static unsigned 
getsize(const set<unsigned> rns,
        const unordered_map<unsigned, const MatchCond*>& rmap) {
   unsigned rv=0;
   for (const auto& rn: rns) {
      const auto& ri = rmap.find(rn);
      if (ri != rmap.end())
         rv += (*ri).second->size();
   }
   return rv;
}
*/

static unsigned 
getsize(const set<unsigned> rns, const unordered_map<unsigned,MatchCond>& rmap) {
   unsigned rv=0;
   for (const auto& rn: rns) {
      const auto& ri = rmap.find(rn);
      if (ri != rmap.end())
         rv += (*ri).second.size();
   }
   return rv;
}

bool no_nondet = true;

void 
codeGen(vector<MatchRule>& rules, unsigned mylbl, 
        ostream& os, int indent, CGFlags& nfl) /*cg*/{
   // Check for matched rules, and eliminate them.

   unsigned k=0;
   mstats.n_nodes++;
   for (unsigned j=0; j < rules.size(); j++) {
      auto& ruleNum = rules[j].ruleNum;
      auto& pat = rules[j].pat;
      if (pat.op() == MatchCond::TRUE) {
         os << "__matched[__mcount++] = " << ruleNum; 
         endln(os, indent);
      }
      else {
         //assert(pat.op() != MatchCond::FALSE);
         if (k < j)
            rules[k] = rules[j];
         k++;
      }
   }

   rules.erase(rules.begin()+k, rules.begin()+rules.size());
   if (k == 0) {
      endln(os, indent);
      return;
   }

#if CG_DBG >= 2
      cout << "\n------------------------\ncodeGen mylbl=" << mylbl << " rules=\n";
      print(rules, cout, ' ', ' ', "\n"); cout << endl;
#endif

   // Select the next tests to perform

   const BranchInfo& br = findBranches(rules);
   unsigned nbr = br.branches.size();

   // Below, we will reference a rule by its ruleNum. This avoids the need to
   // copy over MatchConds several times. 
   unordered_map<unsigned, MatchCond> rulemap;
   for (unsigned j=0; j < rules.size(); j++) 
      rulemap[rules[j].ruleNum] = rules[j].pat;

   // Keep track of whether the kth branch is deterministic or not.
   vector<bool> isDet(nbr);
   for (unsigned j=0; j < nbr; j++) isDet[j] = true;

   // Collect rules that will definitely be nondeterministic
   set<unsigned/*rule#*/> ndChRule;

   // Compute residues reaching each branch once, store for later use
   vector<vector<MatchRule>> residue_store(nbr); 
   // Outer index: branch# (nbr+1: nondet branch); inner vector sorted by rule#

   vector<unordered_map<unsigned/*rule#*/, MatchCond>> residueMap(nbr);
   // Also store the residues as a map for convenient access

   vector<unsigned> chIdx(nbr+1); // chIdx[x]: lowest index w/ same match state.
   for (unsigned j=0; j < chIdx.size(); j++) chIdx[j] = j;
   // We use chIdx[x] > nbr+1 to indicate that the match state was prev seen.
   // For such cases, we will do nothing at all. (We ignore the contribution
   // to state explosion by branches to such states.)

   vector<unsigned> chLbl(nbr+1); // global state label for each child state
   for (unsigned j=0; j < chLbl.size(); j++) chLbl[j] = (1<<30)+j;

   int nextc=-1; unsigned lastk=0;
   for (unsigned k=0; k < nbr; k++) {
      if (chIdx[k] != k) continue;
      if (br.maxOp == MatchCond::EQREG && br.branches[k].op() == MatchCond::EQ) {
         int curc = br.branches[k].cval();
         if (nextc >= curc) {
            mstats.residue_saves += rulemap.size();
            chIdx[k] = lastk;
            continue;
         }
         lastk = k;
         nextc = REGEXP_CHMAX-1;
      }

      for (auto& sp: rulemap) {
         unsigned ruleNum = sp.first;
         MatchCond c(sp.second);
         const MatchCond& test = br.branches[k];
         unsigned char cn = REGEXP_CHMAX-1;
         mstats.residues++;
         if (!c.residue(test, cn)) {
            // False return code: advice to make this branch nondeterministic 
            assert(!no_nondet);
            isDet[k] = false;
            // and the rule nondeterministic (by not inserting into residueMap)
            mstats.residueFalseND++;
            ndChRule.insert(ruleNum); 
         }
         else if (c.op() != MatchCond::FALSE) {
            residue_store[k].emplace_back(ruleNum, c);
            residueMap[k][ruleNum] = c;
         }
#if CG_DBG >= 3
         cout << sp.second << "/" << test << " = " << c << endl;
#endif
         nextc = min(nextc, (int)cn);
      }
   }

   const vector<vector<MatchRule>>& residues = residue_store; 

   // Residues[k] aren't distinct for distinct k due to the use of disjunction
   // (especially common for regexpr), so identify unique child states.
   unsigned ndc=0;              // # of distinct children
   unordered_map<vector<MatchRule>, unsigned> detChTab, ndetChTab;

   for (unsigned k=0; k < nbr; k++) {
      bool found = false;
      if (chIdx[k] != k) continue;
      if (isDet[k]) {
         // For nondeterministic branches, we need to consider the set of rules
         // in the nd branch as part of the match state, but this is difficult
         // to do here since the ndbranch isn't finalized yet. So, at this
         // point, we look for equivalent state only on deterministic branches.
         auto it = matchTab.find(residues[k]);
         if ((found = (it != matchTab.end()))) {
            chIdx[k] = nbr+2;
            chLbl[k] = it->second;
         }
      }
#if CG_DBG >= 2
      else cout << "nd[" << k << "] ";
#endif

      if (!found) { // Not found in global table, insert/lookup in local tables
         auto& mtab = isDet[k]? detChTab : ndetChTab;
         const auto& st = mtab.find(residues[k]);
         if (st == mtab.end()) {
            mtab[residues[k]] = k;
            chIdx[k] = k;
            ndc++;
         }
         else chIdx[k] = st->second;
      }
   }

#if CG_DBG >= 2
   cout << "\nchIdx[k] <> k: ";
   for (unsigned k=0; k < nbr; k++)
      if (chIdx[k] != k) cout << k << "->" << chIdx[k] << " ";
   cout << endl;

   for (unsigned k=0; k < nbr; k++)
      if (chIdx[k] == k)
         cout << "rmap[" << br.branches[k] << "]= " << residueMap[k] << endl;
#endif

   // From here on, we work only on branches such that chIdx[j] == j, as the
   // other branches reach states that have already been examined.

   // Store the rule numbers for rules that deterministically reach one branch
   vector<set<unsigned/*rule#*/>> detChRule(nbr);

   // Numbers for rules that reach a branch possibly nondeterministically
   vector<set<unsigned/*rule#*/>> posNDChRule(nbr);

   for (auto& sp: rulemap) {
      unsigned ruleNum = sp.first;
      unsigned nchild = 0;
      for (unsigned k=0; k < nbr; k++) {
         if (chIdx[k] != k) continue;

         const auto& rmapk = residueMap[k];
         if (rmapk.find(ruleNum) != rmapk.end()) {
            if (nchild++ == 0 || no_nondet)
               detChRule[k].insert(ruleNum);
            else {
               if (nchild == 2)
                  for (unsigned i=0; i < k; i++) {
                     if (chIdx[i] != i) continue;
                     if (detChRule[i].count(ruleNum) == 1) {
                        detChRule[i].erase(ruleNum);
                        posNDChRule[i].insert(ruleNum);
                     }
                  }
               posNDChRule[k].insert(ruleNum);
            }

         }
      }
   }

#if CG_DBG >= 2
   cout << "detCh: "; print(detChRule, cout); cout << "\nposNdCh: " << posNDChRule << endl;
   if (ndChRule.size()) cout << "nondetCh: " << ndChRule << endl;
#endif

   // Now, examine detChRule and posNDChRule and decide whether to move some of
   // the latter into deterministic or nondeterministic branch. The first
   // criteria we use is whether there are a sufficient number of common
   // variables between patterns in detChRule[k] and a pattern p in
   // posNDChRule[k]. If the common fraction is less than thresh, then
   // (1-thresh) fraction of the work to check p cannot be shared with those in
   // detChRule[k]. In other words, even if we try build a subautomaton for p +
   // detChRule[k], most of the work for matching p will still be left even
   // after we finished matching (or ruling out) all other patterns in
   // detChRule[k]. By moving p to the nondeterministic branch, we are
   // indicating our willingness to repeat _all_ of the tests in p, as opposed
   // to just (1-thresh) fraction of them. If thresh is relatively small, say,
   // 0.3, then we are foregoing only a small potential reduction in time. In
   // return, we may gain a lot in terms of space.
   //
   // Note that benign nondeterminism will fall of as a special case where we
   // set thresh=0.0. For any thresh > 0, all patterns that satisfy the benign
   // nondeterminism criteria will be moved to the nondeterministic branch.
   // Specifically, when benign nondeterminism criteria is satisfied, then the
   // set of patterns P can be partitioned into disjoint P_1, P_2,...,P_r such
   // that (a) their union equals P, and (b) no P_i shares any common variables
   // with another P_j. Let P_m be the partition that consists of rules that
   // reference the variable x being examined in the current branch. Then, note
   // that neither x nor any of the other variables in P_m appear in other P_k
   // for k != m. So, they will all satisfy the above criteria based on thresh
   // (for any value of thresh >= 0), so \union_{i != m} P_i will be put in the
   // state S reached by nondeterministic branch. If r > 2, it is easy to see,
   // by following the same argument, that a nondeterministic brnach will be
   // created from S as well. In fact, (at least) r-1 successive
   // nondeterministic descendants of the current state will themselves have
   // nondeterministic descendants.
   //
   // Clearly, the criteria based on variable thresh value (as opposed to benign
   // nondeterminism that corresponds to a hard-code value of thresh=0) is
   // better. Not only does it provide flexibility, but in addition, it provides
   // a more general argument about increasing the total work only by a constant
   // factor of 1/(1-thresh).

   const float thresh = 0.3; 
   unsigned ndchanges=0;

   for (unsigned k=0; k < nbr; k++) {
      if (chIdx[k] != k || !detChRule[k].size() || !posNDChRule[k].size())
         continue;

      set<string> detvars;
      for (auto r: detChRule[k]) 
         residueMap[k][r].getVars(detvars);

      set<unsigned> movToNd;
      for (auto r: posNDChRule[k]) {
         set<string> ndvars;
         residueMap[k][r].getVars(ndvars);
         unsigned totv=ndvars.size();
         unsigned foundv=0;
         for (auto ndvar: ndvars)
            foundv += detvars.count(ndvar);
         if (foundv < thresh*totv) // Less than thresh overlap in variables
            movToNd.insert(r);
      }

      for (auto r: movToNd) {
         ndchanges++;
         isDet[k] = false;
         posNDChRule[k].erase(r);
         ndChRule.insert(r);
      }
   }

   // For a branch k with isDet[k]=false, avoid duplication between kth branch
   // and the nd branch. Otherwise, we would be needlessly repeating the work of
   // matching some pattern on the kth branch, as as well as the ND-branch (when
   // we eventually get to it.) Apply this after every change to ndChRule
   //
   for (unsigned k=0; k < nbr; k++) {
      if (chIdx[k] != k) continue;
      if (!isDet[k]) 
         for (unsigned rn: ndChRule)
            if (detChRule[k].erase(rn) || posNDChRule[k].erase(rn))
               ndchanges++;
   }
   mstats.proactiveND += ndchanges;

#if CG_DBG >= 2
   if (ndchanges > 0) {
      cout << ndchanges << " benign ND moves, nondetCh: " << ndChRule
           << "\ndetCh: " << detChRule << "\nposNdCh: " << posNDChRule << endl;
   }
#endif

   // We have avoided needless explosion, i.e., size explosion without reduction
   // in matching time. But we want to avoid an undue size blow up even if it
   // yields matching time improvement at every state. To do this, start with
   // the recurrence for size of matching code. (As in the paper, we focus on
   // breadth of the code tree rather than the number of nodes in it.)
   //
   //             S(|P|) = S(|P_1|) + ... + S(|P_a|)
   // 
   // where P is the match set for the current state, and P_1,...,P_a
   // are the matchsets for the child states. I can think of two ways to
   // bound explosion:
   //
   //    1. We can define a polynomial p(n) and bound S(n) by p(n). This bound
   //       will hold if 
   //
   //             p(|P|) > p(|P_1|) + p(|P_2|) + ... + p(|P_a|)
   // 
   //    We can force this by moving elements from posNDChRule to ndChRule. In
   //    fact, we can use a lower degree polynomial when n is large.
   //
   //    2. Limit duplication globally. By limiting total duplication to dn,
   //       we can ensure that the code size (breadth) is O(dn). In practice,
   //       it may be good to set two thresholds, a threshold d_i for each rule,
   //       and then an overall threshold d. Or better, set an initial global
   //       d' < d (e.g., d' = 0.5d) and ignore per-pattern threshold until 
   //       d' is reached. After this we enforce d_i as well as the global d.
   //
   // While (2) seems to be a good mechanism because it ensures linear upper
   // bound, it does not provide good control until we hit the limit. By that
   // time, it may be too late, and the only option left is to drastically cut
   // back any further duplication. In contrast, (1) is a local criterion
   // that bounds explosion locally. So, we will implement it for now. 
   //

   // Check criteria 1
   // @@@@ For size, we should consider the size of rules, not just count
   // @@@@ each rule as 1. Requires some coding (lookup sizes by looking at
   // @@@@ residuemap etc.

   unsigned l2rs = 32-__builtin_clz(rules.size());
   double degree = (l2rs >= 10)? 1.3 : (l2rs <= 3? 2 : (2.3 - 0.1*l2rs));
   if (no_nondet) goto done;
   while (true) {
      double rhs = pow(getsize(ndChRule, rulemap), degree);
      double max_diff_k = 0, max_diff=0;
      for (unsigned k=0; k < nbr; k++) {
         if (chIdx[k] != k) continue;
         unsigned detsz = getsize(detChRule[k], residueMap[k]);
         unsigned posndsz = getsize(posNDChRule[k], residueMap[k]);
         double term = pow(detsz + posndsz, degree);
         double diff = term - pow(detsz, degree);
         if (diff > max_diff) {
            max_diff = diff;
            max_diff_k = k;
         }
         rhs += term;
      }
      unsigned lhs = pow(getsize(rules), degree);
      if (rhs < lhs+1 || no_nondet ||
          (rules.size() <= posNDChRule[max_diff_k].size() + ndChRule.size())) {
         // If we keep moving rules to the ND branch, we can sometimes end up
         // in a degenerate situation where the ND branch has the same match
         // state as the current state, and if so, we will get into an
         // infinite loop. We can forestall this by ensuring that the match 
         // states are different. It is simpler to check for the stronger 
         // condition that the max number of rules in the ndbranch after the
         // move from posND (see below) will still be less than that for
         // the current state.         
#if CG_DBG >= 2
      cout<<"Size criteria 1 success: lhs=" << lhs << ", rhs=" << rhs
         << ", deg="<<(log(rhs)/log(lhs))<<'\n';
#endif
         goto done;
      }
      assert(max_diff > 0);

      // Polynomial bound with degree does not hold. Retry after moving the
      // posNDChRule[k] that would contribute most to reduction in rhs. This k
      // was computed in the loop above (max_diff_k).

      isDet[max_diff_k] = false;
      auto& posnd = posNDChRule[max_diff_k];
      for (auto r: posnd) {
         ndchanges++;
         mstats.polyND++;
         ndChRule.insert(r);
      }
#if CG_DBG >= 2
      cout << "Size criteria 1: cleared posNd[" << max_diff_k 
           << "], " << posnd.size() << " rules\n";
#endif
      posnd.clear();
   }

done: 
   unsigned nerased=0;
   if (ndchanges)
      mstats.n_ndnodes++;
   for (unsigned k=0; k < nbr; k++) {
      if (chIdx[k] != k) continue;
      if (!isDet[k]) 
         for (unsigned rn: ndChRule) {
            nerased += detChRule[k].erase(rn);
            nerased += posNDChRule[k].erase(rn);
         }
   }
#if CG_DBG >= 3
   cout<<"Erased "<<nerased<<" rules from det, posND as they are in ND branch\n";
#endif
   // Now, construct the match sets for child states.
   vector<vector<MatchRule>> chRule(nbr+1);
   for (unsigned k=0; k < nbr; k++) {
      if (chIdx[k] != k) continue;
      for (auto r: detChRule[k])
         chRule[k].emplace_back(r, residueMap[k][r]);
      for (auto r: posNDChRule[k])
         chRule[k].emplace_back(r, residueMap[k][r]);
      sort(chRule[k].begin(), chRule[k].begin()+chRule[k].size());
   }
   bool ndbranch = ndChRule.size() > 0;
   if (ndbranch) {
      for (auto r: ndChRule)
         chRule[nbr].emplace_back(r, rulemap[r]);
      sort(chRule[nbr].begin(), chRule[nbr].begin()+chRule[nbr].size());
      nbr++;
      //brnc.branches.push_back(MatchCond(true));
   }

   // Some chRule[k] that were distinct before may no longer be distinct
   // because of moving some elements from posND to nd branch. So, we need to
   // recheck for duplication. 

   unsigned baselbl = lbl;
   for (unsigned k=0; k < nbr; k++) {
      if (chIdx[k] != k) continue;
      if (chRule[k].size() == 0) {
         chLbl[k] = ++lbl;
         continue;
      }

      vector<MatchRule> *augmRules;
      if ((k < nbr-1) && !isDet[k]) {
        // If this is a non-deterministic state, then the rules in the
        // nondeterministic branch are logically part of this state.
        augmRules = new vector<MatchRule>(chRule[k].size()+chRule[nbr-1].size());
        merge(chRule[k].begin(), chRule[k].end(), chRule[nbr-1].begin(),
               chRule[nbr-1].end(), augmRules->begin());
        auto it = unique(augmRules->begin(), 
                         augmRules->begin()+augmRules->size());
        augmRules->resize(distance(augmRules->begin(), it));
      }
      else augmRules = &chRule[k];

      const auto& st = matchTab.find(*augmRules);
      if (st == matchTab.end()) {
         chLbl[k] = ++lbl;
         matchTab[*augmRules] = lbl;
      }
      else {
         chLbl[k] = st->second;
         chIdx[k] = nbr+1;
         for (unsigned j=0; j < k; j++)
            if (chLbl[k] == chLbl[j]) {
               chIdx[k] = j;
               break;
            }
      }
      if (augmRules != &chRule[k]) delete augmRules;
   }

   // Now the branches and matchsets are constructed. 

   doGenCode(br, chRule, isDet, chLbl, chIdx, baselbl, mylbl, 
             nbr, ndbranch, os, indent, nfl);
}

void evCodeGen(const vector<RuleNode*>& rs, EventEntry* ee, ostream& os, 
               int indent, CGFlags& fl) {

   // The loop below identifies rules that can potentially match this event. 
#if CG_DBG >= 2
   cout << "evCodeGen(" << ee->name() << ") with\n";
#endif

   strToVarTx.clear();
   strToVarTx.setPrefix("__qzx");
   matchTab.clear();
   fl.defn_ = true;
   ee->codeGen(os, indent, fl);
   os << " {";
   indent += STEP_INDENT; prtln(os, indent);

   vector<MatchRule> rules; vector<unsigned> rn;
   for (unsigned j=0, k=0; j < rs.size(); j++) {
      ExprNode* cond = rs[j]->pat()->getSubst(ee);
      if (cond) cond->evalConst();
      if (cond && cond->value())
         cond = new ValueNode(cond->value(), cond);
      if (cond) {
         auto nc = cond->normalize(false, os, indent);
         if (nc && nc->op() != MatchCond::FALSE) {
#if CG_DBG >= 4
            cout << "normalized " << *cond << " to " << *nc << endl;
#endif
            rules.emplace_back(k, *nc); rn.push_back(j); k++;
            // We use k to preserve ordering among rules, so that reactions can
            // be executed in the same order as the list of instantiated rules.
         }
      }
   }

   if (errCount() > 0) { cerr << "Errors encountered, exiting.\n"; exit(1); }

   if (rules.size() == 0) {
      os << "}"; endln(os, indent);
      return;;
   }

#if CG_DBG >= 2
   for (auto r: rules) cout << r.pat << " --> " << r.ruleNum << endl;
#endif

   // Create a char* version of string-type vars for use in regexp matching
   unsigned jj=0;
   for (auto it =ee->symTab()->begin(); it !=ee->symTab()->end(); ++it,jj++) {
      VariableEntry* ve = (VariableEntry*)(*it);
      if (ve->type()->isString()) {
         os << "const unsigned char* _c__p" << jj << "= (const unsigned char*)__p" << jj << ".data()";
         endln(os, indent);
      }
   }

   // We generate the code for checking the conditions in the rule on event
   // arguments, and if satisfied, then the associated action will be executed.
   // We defer the execution of these actions till the end, i.e., until all
   // matches are determined. This way, if the action has side-effects, e.g.,
   // modifies data that is passed by reference into the matching code, those
   // won't affect match results. We collect the matched rules in the array
   // __matched that is computed during matching. This array simply stores the
   // index of the rule that matched. By doing this, we are able to sort the
   // rules based on this index, and then execute them in the indexed order.
   // This ensures that actions are executed in the order in which the rules are
   // listed. (Note that the top-level instantiate creates rules in the order of
   // instantiation of modules, with the module's own rules appearing after
   // those of the submodules it instantiated.)

   os << "int __matched[" << rn.size() << "]; int __mcount = 0;"; 
   prtln(os, indent);

   mstats.cur_sz = getsize(rules);
   mstats.tot_sz += mstats.cur_sz;

   codeGen(rules, lbl, os, indent, fl); // Generates code for matching

   // At this point, the matched rules are stored in __matched, and the
   // the length of this array is in __mcount. We first sort it.
   os << "sort(&__matched[0], &__matched[__mcount]);"; prtln(os, indent);

   // Go through the __matched array, generate code for each matched action.
   os << "for (unsigned i=0; i < __mcount; i++) {"; 
   indent += STEP_INDENT; prtln(os, indent);
   // Eliminate duplicate matches
   os << "if (i > 0 && __matched[i] == __matched[i-1]) continue;";       
   prtln(os, indent);

   // The actions will be all in a giant switch statement, with one case
   // corresponding to each action. This approach gives us full flexibility,
   // because the code for each case can be arbitrarily different from others.
   os << "switch(__matched[i]) {"; 
   prtln(os, indent);
   for (unsigned j=0; j < rn.size(); j++) {
      os << "case " << j << ": ";
      rs[rn[j]]->reaction()->codeGen(os, indent, fl);
      prtln(os, indent+STEP_INDENT);
      os << "break;";
      prtln(os, indent);
   }

   os << "default: assert_abort(false); break;";
   prtln(os, indent);
   os << "}"; indent -= STEP_INDENT;
   prtln(os, indent);
   os << "}";
   prtln(os, indent);

   prtRewind(os, STEP_INDENT, indent);
   os << "}";
   prtMatchStats(os, "After "+ee->name());
   indent -= STEP_INDENT; prtln(os, indent); 

#if CG_DBG >= 2
   for (const auto& ri: rangeInfo)
      cout << ri.first << ": " << ri.second.str() << endl;
   rangeInfo.clear();
#endif
}
