#ifndef COND_FAC_H
#define COND_FAC_H

#include <functional> 
#include "Ast.h"
#include "RegExpr.h"

void prtMatchStats(ostream& os, string prfx);

// See "condition factorization" (TIFS 2016) paper for the background on this
// code. Terminology here is mostly consistent with that paper. 

template <class K, class E> class OldMerit;
template <class K, class E> class NewMerit;
#ifdef OLDM
#define Merit OldMerit
#else
#define Merit NewMerit
#endif

class MatchCond {
 /* Expressions to which condition factorization can be applied must satisfy
    some restrictions. For instance, all comparisons ("tests") must be of the
    form <var> <op> <value> where <op> is of type FacOp defined below.
    MatchCond represents such factorizable expressions. */

 public:
   enum FacOp: unsigned char {
      TRUE, FALSE,              // degenerate case, no variables
      AND, OR, SEQ_AND, SEQ_OR, // conjunction/disjunction
      IN, NOTIN,                // conditions on sets
      EQBIT, NEQBIT,            // eq/diseq on selected bits
      EQ, NE,                   // equality/disequality
      EQREG, NEQREG,            // regular expressions
      GT, LT,                   // inequalities
      ASG,                      // to permit expression evaluation
      IF,                       // place holder to support arbitrary conditions
      INVALID                   //    in the future. 
   };

 private:
   FacOp op_;                   // comparison operator
   bool charOp_;                // true means we compare only with leading char
   string var_;                 // variable involved in the test
   union {
      Value* v_[2];             // v[0] is value to be compared, v[1] is optional
                                // and specifies bits for EQBIT/NEQBIT
      MatchCond* ch_[2];        // used for disj/conj
      set<Value> set_;          // used for set comparisons
      ExprNode* expr_;          // used as rhs of assignments.
   };

 public:
   // Constructors to construct all of the above-mentioned condition types
   MatchCond(bool b=false) 
      { op_ = (b? TRUE: FALSE); charOp_ = false; v_[0]=v_[1]=nullptr;};
   MatchCond(FacOp op, string n, const Value& v, bool charop=false): 
             op_(op), charOp_(charop), var_(n) { 
      assert (op==EQ || op==NE || op==EQREG || op==NEQREG || op==GT || op==LT);
      v_[0] = new Value(v); v_[1] = nullptr;
   };
   MatchCond(FacOp op, string n, char c): op_(op), charOp_(true), var_(n) { 
      v_[0] = new Value((long)c); v_[1]=nullptr;
   };
   MatchCond(FacOp op, string n, const Value& v1, const Value& v2):
             op_(op), charOp_(0), var_(n) { 
      assert(op == EQBIT || op == NEQBIT);
      v_[0] = new Value(v1); v_[1] = new Value(v2);
   };
   MatchCond(FacOp op, string n, const set<Value>& v, bool charOp=false):
              op_(op), charOp_(charOp), var_(n), set_(v) {
      assert(op == IN || op == NOTIN);
   };
   MatchCond(FacOp op, string n, ExprNode* rhs): op_(ASG), charOp_(0), var_(n) {
      assert(op==ASG);
      expr_ = rhs;
   }

   // This object takes over the ownership of f1 and f2, and will delete them
   // when it is destroyed.
   MatchCond(FacOp op, MatchCond* f1, MatchCond* f2): op_(op), charOp_(0) { 
      assert(op == OR || op == AND || op == SEQ_AND || op == SEQ_OR);
      ch_[0] = f1; ch_[1] = f2; 
   };

   MatchCond(const MatchCond& fc): op_(FALSE), charOp_(false) { deepCopy(fc); }

   ~MatchCond() { releaseResources(); };

   // Accessors, comparisons and assignment on MatchCond
   bool isCharop() const { return charOp_; };
   FacOp op() const { return op_; }
   string var() const { assert(op_ >= IN); return var_; };
   const Value& val() const { assert(op_ >= IN && op_ <= LT); return *v_[0]; };
   unsigned char cval() const 
     { assert(charOp_ && op_ >= EQ); return (char)v_[0]->lval(); };
   const ExprNode* expr() const { assert(op_ == ASG); return expr_; };

   unsigned size() const {
      switch (op_) {
      case AND: case SEQ_AND: case OR: case SEQ_OR:
         return ch_[0]->size() + ch_[1]->size();
      case TRUE: case FALSE: return 0;
      case EQREG: return val().reval()->size();
      default: return 1;
      }
   }

   bool hasRegexpOp() const {
      switch (op_) {
         case AND: case SEQ_AND: case OR: case SEQ_OR:
            return ch_[0]->hasRegexpOp() || ch_[1]->hasRegexpOp();
      case EQREG: return true;
      default: return false;
      }
   }

   // MatchCond contain any of the ops in opset on varName (ignore if empty)? 
   bool opPresent(unsigned opset, string varName="") const;

   // Return all of the variables referenced in this MatchCond
   void getVars(set<string>& s) const;

   bool operator==(const MatchCond& fc) const;
   bool operator!=(const MatchCond& fc) const { return !(*this == fc);};
   bool operator<(const MatchCond& fc) const;
   size_t hash(size_t seed=0) const;
   void print(ostream& os, int indent=0) const;

   // void shallowCopy(const MatchCond& fc);
   void deepCopy(const MatchCond& fc);
   const MatchCond& operator=(const MatchCond& fc) 
       {deepCopy(fc); return *this;};
   void releaseResources();

   bool negate();

   // Now, the main functions: residue, codeGen
   bool residue(const MatchCond&, unsigned char& nextc);

   static void codeGen(const vector<pair<MatchCond*, StmtNode*>>& rs, 
                       ostream& os, int indent, CGFlags& fl);

   // Helper functions used by codegen
   Merit<string, Value> getEqIneqReMerit(FacOp fop, float weight) const; 
   // Returns a dictionary that, for each variable v in this MatchCond,
   // specifies how useful it is to create a transition of type tt.
   Merit<MatchCond, bool> getIfMerit(float weight) const;
   // Returns a dictionary that, for each possible leaf-level MatchCond lmc in
   // this, specifies usefulness of an if-then-else transition on lmc.

private:
   bool charOpResidue(const MatchCond&, unsigned char& nextc);

};

inline 
ostream& operator<<(ostream& os, const MatchCond& mc) {mc.print(os); return os;}
__attribute__((unused))
static const char*  facOpNm[] = {
   "TRUE", "FALSE",
   "AND", "OR", "SEQ_AND", "SEQ_OR",
   "IN", "NOTIN",
   "EQBIT", "NEQBIT",
   "EQ", "NE",
   "EQREG", "NEQREG",
   "GT", "LT",
   "ASG",
   "IF",
   "INVALID"
};

namespace std {
template <> struct hash<MatchCond> {
 public:
   size_t operator()(const MatchCond& m) const { 
      return m.hash(); 
   }
};
};

void evCodeGen(const vector<RuleNode*>& rs, EventEntry* ee, 
               ostream& os, int indent, CGFlags& fl);

template <class KeyT, class ElemT>
class NewMerit {
   // A class that helps in deciding which variable to branch on. The idea is to
   // estimate the *SIZE* (not matching time) of the automaton that results when
   // a particular branch type is used. Then, we can pick the branch type that
   // minimizes this estimate. 
   //
   // Merit is computed for a MatchCond M and an operator op. It has two parts,
   // det and reduction (red). Determinacy is the probability that the entire
   // condition M is eliminated after the test. Specifically, det(M, op), also
   // denoted as det(M, v op x), captures the probability M/(v op x) = FALSE.
   // For instance,
   // 
   //      -- det(v=3, v=x) is (1 - 1/|domain(v)|), since v=3/v=x is false 
   //          for all values of v in its domain except the one specific value 3.
   //      -- det(v<k, v<x) is 0.5,
   //      -- det(C1&&C2, t) = 1 - (1-det(C1, t)) * (1-det(C2, t))
   //      -- det(C1||C2, t) = det(C1, t) * det(C2, t)
   //
   // When det=1, it does not mean that M will fail on all branches, just that
   // it will fail on all but one --- the "primary" branch for M. We design our
   // merit heuristic to capture the "excess" over this base.
   //
   // When M isn't eliminated by t, it is likely that performing t obviates the
   // need for some of tests in M. This contributes to a reduction (red) in the
   // number of remaining tests, captured using the following equations. (Note
   // that in this case, we are interested in the average reduction across the
   // branches where M is still viable.)
   //
   //      -- red(v=k, v > x) is 0 (M/v>x still has v=k if it is still viable)
   //      -- red(v>k, v=x) is 1 (we know v=x, so no need to check if v>k)
   //      -- red(C1&&C2, t) = red(C1||C2, t) = red(C1, t) + red(C2, t).
   //
   // Within a single rule, we compute det and red using the rules above. At the
   // level of a rule, we convert this into a quantity that will more directly
   // relate to the automaton size. Specifically, we try to compute the 
   // (expected) remaining size of the rule R after testing t as follows.
   // If det(R, t) = d, then, with probability d, we eliminate R entirely,
   // ending up with a size of zero. If it is not eliminated, and red(R, t) = r,
   // then, with a probability (1-d), we are left with a test of size |R|-r. So,
   // the expected remaining size is given by:
   //
   //      remsz = (1-d)*(|R|-r)
   //
   // When d = 1, remsz becomes zero, which seems inconsistent with the fact
   // that the rule will (almost always) remain on one of the branches. Thus,
   // remsz talks about the size left on the other ("non-primary") branches for
   // R. So, what remsz captures is the "excess" over the base case where R
   // remains viable in exactly one of the branches. In fact, it captures the
   // excess per branch on all except the primary branch. So, to compute the
   // excess, we need to multiply remsz by (#branches - 1).
   //
   // Once we compute remsz for a rule, we use Merit to keep track of remsz for
   // sets of rules. In this phase, remsz simply adds up for corresponding keys
   // (usually variables). If a key is present in ruleset A but not B, then
   // rules in B will succeed on any test on that key, and hence the excess
   // will be the size of the entire set B, i.e., we add remsz of A with sz of B.
   //
   // Once we are dealing with sets of rules, we don't need det and red as
   // defined above. We simply stop computing red. We also redefine det as the
   // probability of making a deterministic branch on a key. We do a
   // size-weighted average, i.e., 
   //         det of A + B is (A.det*A.sz + B.det*B.sz)/(A.sz+B.sz).

 public:
   struct TransInfo {
      float det;
      float red;
      float remsz;
      set<ElemT> elems;

      TransInfo() {det = 0.0; red = 0.0; remsz=0.0;};
   };

 private:
   bool intraRule;
   unsigned sz;
   unordered_map<KeyT, TransInfo> trans_;

   // Direct accessors for the dictionary, for internal use only.
   TransInfo& transInfo(string s) { return trans_[s]; }
   const TransInfo& transInfo(string s) const {
      return const_cast<NewMerit*>(this)->transInfo(s);
   }

 public:
   NewMerit() {sz=1; intraRule=true;};

   NewMerit(KeyT s, float w, float d, float u, const ElemT* v=0) {
      if (w != 0) { d = d/w; u = u/w; };
      auto& ts = trans_[s]; ts.det = d; ts.red = u; 
      if (v) ts.elems.insert(*v);
      sz=1; intraRule=true;
   }
   NewMerit(KeyT s, float w, float d, float u, ElemT v) {
      if (w != 0) { d = d/w; u = u/w; };
      auto& ts = trans_[s]; ts.det = d; ts.red = u; ts.elems.insert(v);
      sz=1; intraRule=true;
   }
   NewMerit(KeyT s, float w, float d, float u, const vector<ElemT>&vs) {
      if (w != 0) { d = d/w; u = u/w; };
      auto& ts = trans_[s]; ts.det = d; ts.red = u; 
      for (const auto& v: vs) ts.elems.insert(v);
      sz=1; intraRule=true;
   }

   bool present(KeyT s) const { return (trans_.find(s) != trans_.end()); }
   const unordered_map<KeyT, TransInfo>& trans() const { return trans_; };
   float det(KeyT s) const { return (present(s)? transInfo(s).det : 0.0); }
   float red(KeyT s) const { return (present(s)? transInfo(s).red : 0.0); }
   float remsz(KeyT s) const{ return (present(s)? transInfo(s).remsz : 0.0);}
   unsigned size() const {return sz; };

   void upd_remsz() {
      assert(intraRule);
      for (auto& ti: trans_) {
         float rs = (1-ti.second.det)*(sz - ti.second.red);
         ti.second.remsz = rs;
      }
      intraRule=false;
   }

   void or_trans(const NewMerit& o) {
      for (auto oti: o.trans_) {
         if (trans_.find(oti.first) == trans_.end())
            trans_[oti.first].remsz = sz;
         auto& ti1 = trans_[oti.first];
         if (intraRule) {
            ti1.det *= oti.second.det;
            ti1.red += oti.second.red;
         }
         else {
            ti1.remsz += oti.second.remsz;
            ti1.det = (ti1.det*sz + oti.second.det*o.sz)/(sz + o.sz);
         }
         for (auto ov: oti.second.elems)
            ti1.elems.insert(ov);
      }

      for (auto& ti: trans_) {
         auto oti = o.trans_.find(ti.first);
         if (oti == o.trans_.end()) {
            if (!intraRule) {
               ti.second.remsz += o.sz;
               ti.second.det *= ((float)sz)/(sz+o.sz); 
            }
            else ti.second.det = 0;
         }
      }

      sz += o.sz;
   }

   void and_trans(const NewMerit& o) {
      assert(intraRule);
      for (auto oti: o.trans_) {
         auto& ti1 = trans_[oti.first];
         ti1.red += oti.second.red;
         ti1.det = 1-(1-ti1.det)*(1-oti.second.det);
         for (auto ov: oti.second.elems)
            ti1.elems.insert(ov);
      }

      sz += o.sz;
   }

   void print(ostream& os) const {
      os << "Merit\n";
      for (auto kti: trans_) {
         os << kti.first << ": det=" << kti.second.det;
         if (!intraRule)
            os  << " remsz=" << kti.second.remsz << " ";
         else os << " red=" << kti.second.red << " ";
         for (const auto& v: kti.second.elems)
            os << v << " ";
         os << endl;
      }
   }
};
#endif
