#ifndef VALUE_H
#define VALUE_H

#include "all.h"
#include "Error.h"

class Type;
class Value;
class ProgramElem;
class SymTabEntry; 
class RegExpr;

/* 
   Types are meant to be architecture independent: short is 16-bits,
   int is 32-bits and long is 64-bits. Float is 32-bits and double
   is 64-bits. String is a length-delimited sequence of 8-bit characters.
   For now, though, we will cheat --- and the code will work only on x86_64.
*/

class Type {
 public:
  // Dont change the order below --- it is necessary for isSubType()
  // logic to work correctly.
  enum BaseTypeTag  {
    ERROR, UNKNOWN, VOID, 
    STRING, REGEXPR,
    BOOL, 
    BIT,
    UCHAR, CHAR, 
    USHORT, SHORT,
    ENUM, 
    UINT, INT,
    ULONG, LONG, 
    DOUBLE,
    STRUCT, CLASS, 
    VECTOR, SET, DICT, TUPLE, 
    MODULE, EVENT, FUNCTION,
    NUMERIC, /* BIT to DOUBLE */
    INTEGRAL, /* BIT to LONG */
    SIGNED, /* CHAR, SHORT, INT, LONG, DOUBLE */
    UNSIGNED, /* UCHAR, USHORT, UINT, ULONG */
    SCALAR, /* STRING to DOUBLE */
    PRIMITIVE, /* STRING to DOUBLE */
    DICTORSET, 
    CONTAINER, /* VECTOR, SET, DICT */
    CONTAINER_OR_STR, 
    NATIVE, /* Not STRUCT, CLASS or VOID */
    DATA, /* VOID to TUPLE */
    ANY   /* VOID to FUNCTION */
  };

  class TypeTag {
   private:
     BaseTypeTag tag_: 7;
     bool isConst_: 1;

   public:
     TypeTag() { tag_ = ERROR; isConst_ = 0; }
     TypeTag(BaseTypeTag t, bool isConst=false) 
        { tag_ = t; isConst_ = isConst; }

     BaseTypeTag tag() const { return tag_; };
     bool isConst() const { return isConst_; };
     void isConst(bool c) { isConst_ = c; };

     //operator BaseTypeTag() const { return tag_; };
     bool operator==(const TypeTag& t) const 
        { return (tag_ == t.tag_ && isConst_ == t.isConst_); }
     bool operator!=(const TypeTag& t) const { return !(*this == t);}

     bool operator==(BaseTypeTag t) const { return tag_ == t; }
     bool operator!=(BaseTypeTag t) const { return tag_ != t; }
     bool operator< (BaseTypeTag t) const { return tag_ <  t; }
     bool operator<=(BaseTypeTag t) const { return tag_ <= t; }
     bool operator> (BaseTypeTag t) const { return tag_ >  t; }
     bool operator>=(BaseTypeTag t) const { return tag_ >= t; }
  };

 public:
  static const int SIZE_UNDEF;

  static const string name(TypeTag t);
  static const string codeName(TypeTag t, bool inclConst=false);
  static int size(BaseTypeTag t);

  static const Type errorType, voidType, unkType;
  static const Type boolType;
  static const Type stringType, constStringType; 
  static const Type regExprType;
  static const Type bitType, ucharType, charType, ushortType, 
    shortType, uintType, constUintType, intType, 
    ulongType, longType,
    doubleType;
  
  static const Type* type[];
  static bool isConst(TypeTag t) { return (t.isConst()); }
  static bool isTagOnly(TypeTag t) { return (t <= DOUBLE); };
  static bool isString(TypeTag t) { return (t==STRING);}
  static bool isRegExpr(TypeTag t) { return (t==REGEXPR);}
  static bool isNumeric(TypeTag t) { return ((t >= BIT) && (t <= DOUBLE));}
  static bool isBool(TypeTag t) { return (t == BOOL); }
  static bool isInt(TypeTag t) {return (t >= BIT && t <= INT); }
  static bool isLong(TypeTag t) {	return (t >= ULONG && t <= LONG);}
  static bool isIntegral(TypeTag t) { return ((t >= BIT) && (t <= LONG));}
  static bool isSigned(TypeTag t) {
    return ((t==CHAR) || (t==SHORT) || (t==INT) || (t==LONG) || (t==DOUBLE)); }
  static bool isUnsigned(TypeTag t) { return (isIntegral(t) && !isSigned(t));}
  static bool isDouble(TypeTag t) {return (t==DOUBLE);}
  static bool isPrimitive(TypeTag t) { return ((t >= STRING) && (t <= DOUBLE));}
  static bool isContainer(TypeTag t) { return (t==VECTOR || t==SET || t==DICT);}
  static bool isContainerOrStr(TypeTag t) {return (t==STRING || isContainer(t));}
  static bool isNative(TypeTag t) {
    return ((t != CLASS) && (t != STRUCT) && (t > VOID)); }
  static bool isData(TypeTag t) { return ((t >= VOID) && (t <= TUPLE));}
  static bool isValid(TypeTag t) { return ((t >= VOID) && (t <= FUNCTION)); }
  static bool isScalar(TypeTag t) { return ((t >= STRING) && (t <= DOUBLE)); }
  static bool isTuple(TypeTag t) { return (t == TUPLE); }
  static bool isDict(TypeTag t) { return (t == DICT); }
  static bool isVector(TypeTag t) { return (t == VECTOR); }
  static bool isSet(TypeTag t) { return (t == SET); }
  static bool isVecOrSet(TypeTag t) { return (t == SET || t == VECTOR); }
  static bool isDictOrSet(TypeTag t) { return (t == SET || t == DICT); }

 private:
  TypeTag tag_;
  Type* keyType_;         // Used by dict
  union {
    SymTabEntry* typeDesc_; // enum, struct or class @@@@ incld for fun, module
    Type* elemType_;        // Used by list, set and dict
    struct {                // For function, module, event and tuples
      Type* retType_;       // Applicable only for functions
      vector<Type*>* argTypes_;  // Applicable for module, event, tuples
    };
  };

 public:
  Type(TypeTag tag=VOID);                    // All elementary types
  Type(SymTabEntry* td, TypeTag t);          // Must be enum/class/struct
  Type(Type* elemType, TypeTag t);           // For list and set types
  Type(Type* keyType, Type* elemType);       // For dict type         
  Type(vector<Type*>* tupleType, TypeTag t); // For tuples, modules, events
  Type(vector<Type*>* argt, Type* rt);       // For functions
  //Type(const Type& t) {operator=(t);};
  ~Type() {};
  
  const Type& operator=(const Type&);
  // Assignment does a deep copy for all fields except typeDesc_, which is
  // is a SymTabEntry* and has a reference semantics. (Note SymTabEntry
  // doesn't even have a copy constructor or assignment.)

  bool isConst() const { return tag_.isConst(); }; 

  string name() const {return name(tag_); }
  string fullName() const;
  string codeName(bool inclConst=false) const 
     {return codeName(tag_, inclConst); }
  string fullCodeName(bool inclConst=false) const;

  BaseTypeTag tag() const 
   { return tag_.tag(); }

  const SymTabEntry* typeDesc() const {
    if ((tag_ != CLASS) && (tag_ != STRUCT) && (tag_ != ENUM)) return nullptr;
    else return typeDesc_;
  };
  SymTabEntry* typeDesc() {
    if ((tag_ != CLASS) && (tag_ != STRUCT) && (tag_ != ENUM)) return nullptr;
    else return typeDesc_;
  };

  const Type* keyType() const { 
    if (tag_ == VECTOR || tag_ == SET) 
       return elemType_; 
    if (tag_ != DICT) return nullptr;
    else return keyType_; 
  };

  Type* keyType() { 
    if (tag_ == VECTOR || tag_ == SET) 
       return elemType_; 
    if (tag_ != DICT) return nullptr;
    else return keyType_; 
  };

  const Type* elemType() const { 
    if (tag_ != VECTOR && tag_ != SET && tag_ != DICT) return nullptr;
    else return elemType_; 
  };

  Type* elemType() { 
    if (tag_ != VECTOR && tag_ != SET && tag_ != DICT) return nullptr;
    else return elemType_; 
  };

  int arity() const {
    if (((tag_ != TUPLE) && (tag_ != MODULE) && 
		(tag_ != EVENT) && (tag_ != FUNCTION))||(argTypes_ == nullptr)) return 0;
    else return argTypes_->size(); 
  };  

  const vector<const Type*>* argTypes() const { 
    if ((tag_ != TUPLE) && (tag_ != MODULE) && 
	(tag_ != EVENT) && (tag_ != FUNCTION)) return nullptr;
    else return (vector<const Type*>*) argTypes_; 
  };
  
  vector<Type*>* argTypes() { 
    if ((tag_ != TUPLE) && (tag_ != MODULE) && 
	(tag_ != EVENT) && (tag_ != FUNCTION)) return nullptr;
    else return argTypes_; 
  };
  
  const Type* retType() const {
    if (tag_ != FUNCTION) return nullptr;
    else return retType_;
  };

  void constType(bool b=true) { tag_.isConst(b); };
  void tag(TypeTag t) { tag_ = t; }
  void typeDesc(SymTabEntry* ste) {
    if ((tag_ == CLASS) || (tag_ == STRUCT) || (tag_==ENUM)) typeDesc_ = ste; 
    else errMsg("Type::typeDesc(SymTabEntry*) called when type = " + name()); 
  };

  void elemType(Type *t) { 
    if (tag_ == VECTOR || tag_ == SET || tag_ == DICT) elemType_ = t;
    else errMsg("Type::elemType(Type*) called when type = " + name()); 
  };

  void argTypes(vector<Type*>* t) { 
    if ((tag_ == TUPLE) || (tag_ == MODULE) ||
	(tag_ == EVENT) || (tag_ == FUNCTION)) argTypes_ = t;
    else errMsg("Type::argTypes(vector<Type*> *) called when type = " + name()); 
  };
  void retType(Type* t) {
    if (tag_ == FUNCTION) retType_ = t;
    else errMsg("Type::retType(Type *) called when type = " + name());
  };

  void print(ostream &os, int indent=0) const;

  bool isValid() const { return isValid(tag_); };
  bool isBool() const { return isBool(tag()); };
  bool isString() const { return isString(tag()); };
  bool isRegExpr() const { return isRegExpr(tag()); };
  bool isIntegral() const { return isIntegral(tag()); };
  bool isNumeric() const { return isNumeric(tag()); };
  bool isScalar() const { return isScalar(tag()); };

  bool isAssignable(const Type& t) const ; /* initialization with literal*/
  bool isSubType(const Type& t, bool relaxed=0) const; // Is this a subtype of t?
  bool isSubType(TypeTag t) const;   // Is this a subtype of t?
  bool isEquiv(const Type& t) const      // Type equivalence
    { return (isSubType(t) && t.isSubType(*this)); };
   
  void unify(const Type* t, const ProgramElem* pe); // Updates this to the most 
  // general type that's a subtype of this and t. Doesn't handle type vars (yet).

  bool operator==(const Type& t) const;  // strict equality
  bool operator!=(const Type& t) const { return !operator==(t);};

  void printArgs(ostream& os, int indent=0)const;
};

inline ostream& operator<< (ostream& os, const Type& tt) {
  tt.print(os);
  return os;
};


/************************************************************************/
/* Value Class */
/************************************************************************/
class Value {
 private:
  const Type* type_;       // Shallow-copied from caller of constructor
  union {            // All components of this union shd be deep-copied
    string* sVal_;
    RegExpr* reVal_;
    long lVal_;
    double dVal_;
    vector<Value*>* listVal_;
  };

 public:
  Value() // void value
    { type_ = &Type::voidType;}
  Value(const char* s)
    { sVal_ = new string(s); type_ = &Type::stringType; }
  Value(const string& s)
    { sVal_ = new string(s); type_ = &Type::stringType; }
  Value(const RegExpr* re);
  Value(Type::BaseTypeTag t, unsigned long ul);
  Value(long l);
  Value(unsigned long ul);
  Value(unsigned u);
  Value(double d)                // Floats stored as double values
    { type_ = &Type::doubleType; dVal_ = d; }
  Value(int i, SymTabEntry* t);   // Only enum; Struct, class unsupported
  Value(const vector<Value*>* l, Type *t) ;  // Used for lists and tuples
  // Module, event and function values unsupported

  Value(Type::BaseTypeTag t) // @@@@ Works only for non-const primitive types
   { type_ = Type::type[t];  sVal_ = nullptr; reVal_ = nullptr;}
  Value(const Value& v): type_(v.type_) { sVal_ = nullptr; *this = v;}

  ~Value();

  const Type* type() const { return type_; } 
  //Type* type() { return type_; } 
  //void type(Type* t) { type_ = t;};
  
  bool isIntegral() const { return Type::isIntegral(type()->tag()); };
  bool isNumeric() const { return Type::isNumeric(type()->tag()); };
  bool isString() const { return Type::isString(type()->tag()); };
  bool isRegExpr() const { return Type::isRegExpr(type()->tag());}
  bool isDouble() const {return Type::isDouble(type()->tag());}

  string sval() const;
  const RegExpr* reval() const;
  RegExpr* reval();
  bool bval() const;
  int ival() const;
  long lval() const;
  unsigned long ulval() const;
  double dval() const;
  const vector<const Value*>* listVal() const;
  size_t hash() const;

  void sval(const string& s) ;
  void reval(const RegExpr& re);
  void bval(bool b) { lVal_ = b ;} 
  void ival(int i) ;
  void lval(long l) ;
  void dval(double d) ;
  void listVal(const vector<Value*>* vl) ; // 

  void print(ostream& os, int indent=0) const;
  void codeGen(ostream& os, const Type* coercedType, string prefix="", 
               int indent=0) const;

  bool operator==(const Value&) const;
  bool operator!=(const Value& v) const { return !operator==(v);};
  bool operator<(const Value& v) const;
  bool operator>(const Value& v) const { return v < *this; };
  bool operator<=(const Value& v) const
     { return (*this < v) || (*this == v); }
  bool operator>=(const Value& v) const {return v <= *this;};
  const Value& operator=(const Value&);

  static Value zero() {return Value(0l); };

 private:
  void releaseResources();
}
;

inline ostream& operator<< (ostream& os, const Value& tt) {
  tt.print(os);
  return os;
};

extern string newName(const string);

enum SymKind {
   // If this enum is changed, kindTag() needs to be updated correspondingly
   UNKNOWN_KIND, 
   GLOBAL_KIND, INTERFACE_KIND, STRUCT_KIND, CLASS_KIND, FUNCTION_KIND,
   MODULE_KIND,
   VARIABLE_KIND,
   ENUM_KIND, ENUM_ELEM_KIND,
   EVENT_KIND,
   LABEL_KIND, ON_KIND,
   BLOCK_KIND, RULE_BLOCK_KIND, EVENT_BLOCK_KIND, 
   ERROR_KIND
};

enum VarKind {FOREACH_VAR, FN_PARAM_VAR, LOCAL_VAR, GLOBAL_VAR, MODULE_PARAM_VAR,
              EVENT_PARAM_VAR, PATTERN_VAR /* for REWBR*/, RULE_VAR,
              COMPILER_GENERATED_VAR,
              UNDEFINED};

string kindTag(SymKind k);
string varKindTag(VarKind vk);

class VariableEntry;
class RuleNode;
class ExprNode;

struct CGFlags {
   bool stateMachine_;
   bool defn_;
   bool invocation_;
   bool learnMode_;
   bool first_;
   bool noneFlag_;
   string context_;
   string prefix_;
   map<const SymTabEntry*, ExprNode*> subst_;
   unordered_map<string, ExprNode*> evargsubst_;
   RuleNode* curRule_;

   CGFlags() { 
      prefix_=context_=""; curRule_=nullptr; learnMode_ = first_ = true;
      defn_=invocation_=stateMachine_ = noneFlag_ = false;
   };
   CGFlags(string p) {
      noneFlag_ = (p == "_NONE_");
      if (!noneFlag_) prefix_=p; 
      context_ = "";
      curRule_=nullptr; 
      learnMode_ = first_ = true;
      defn_=invocation_=stateMachine_ = false;
   };
   string prefix() const { return (prefix_+context_=="")? "" :
         prefix_+context_+ ".";}
   string context() const { return context_; };
   void addPrefix(string s) { 
      if (context_ == "")  
         context_ = s;
      else context_ = context_ + "." + s;
   }
   string fnPrefix() const {return prefix_;}
   string typePrefix() const {return prefix_;}
   string enumPrefix() const {return prefix_;}
   string enumElemPrefix() const {return prefix_;}
   bool stateMachine() const { return stateMachine_; };
   bool noneFlag() const { return noneFlag_; };
};

static_assert(sizeof(Value)==16, "fix Value size");

namespace std
{
template<> 
struct hash<Value> {
   size_t operator()(const Value& v) const { return v.hash();}
};
template<> 
struct less<Value> {
   size_t operator()(const Value& v1, const Value& v2) const { return v1 < v2;}
};
};
#endif
